/* -*- indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/***** BEGIN POSTBOX LICENSE BLOCK *****
 *
 * msgDisplayEmitter.js
 *
 * Copyright (c) 2007-2009 Postbox, Inc.
 *
 ***** END POSTBOX LICENSE BLOCK *****/

const gConversationPrefix = "<?xml version=\"1.0\" encoding=\"UTF-8\"?> \
<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \
  \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\" [ \
<!ENTITY % headersDTD SYSTEM 'chrome://messenger/locale/msgHdrViewOverlay.dtd' > \
%headersDTD; \
 \
]> \
<html xmlns=\"http://www.w3.org/1999/xhtml\"> \
  <head> \
    <link rel='important stylesheet' href='chrome://global/locale/intl.css'/> \
    <link rel='important stylesheet' href='chrome://messenger/skin/conversationHeaders.css'/> \
  </head> \
  <body>";

const gConversationSuffix = " \
  </body>                     \
</html>";

/** a message container can be in three states:
 *  ghost a ghost row (completely invisible such as a message in another folder when we are not showing all messages in conversation
 *  slim - a collapsed row where only the sender/date/folder line is visible
 *  open - the message body and headers are shown
 *
 *  additionally the message container has a inprimaryfolder attribute which is set to true when the message is in the same folder as the primary msg hdr
 */ 
const gConversationMsgTemplate = " \
<div class='alternate message-container' style='height:100%' inprimaryfolder='%MESSAGECONTAINERPRIMARYFOLDER%' mode='%MESSAGEMODE%'> \
  %MESSAGEHEADERS% \
  %CONVERSATIONIFRAME% \
</div>";

const gGhostConversationIframeTemplate = "";
const gConversationIframeTemplate = "<iframe id='flyingpigs%CONVERSATIONINDEX%' class='postbox-conversation-iframe' src='%MESSAGEBODYURLSRC%' style='height:100%; width:100%; border: none;' scrolling='no' ></iframe>";

const gConversationHeaderTemplate = "                           \
<div class='conversation-headers' index='%CONVERSATIONINDEX%' bodyurl='%MESSAGEBODYURL%' expandedheaders='false'>                                                                        \
  <table class='conversation-headers-table'> \
    <tr> \
      <td class='header-name mailheader from-header-name expanded-header-row'>&fromField.label;</td> \
      <td> \
        <span class='from-value mailheader-value'>%fromvalue%</span> \
        <span class='mailheader'>&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;</span>             \
        <span class='date-value mailheader-value'>%datevalue%</span> \
        <span class='mailheader'>&nbsp;&nbsp;&nbsp;|&nbsp;&nbsp;&nbsp;</span> \
        <span class='folder-value mailheader-value' messageKey='%messagekeyvalue%' folderURI='%folderurivalue%'>%foldervalue%</span> \
        <span id='conversationExpanders%CONVERSATIONINDEX%' class='conversationExpanders'> \
          <span class='mailheader'>&nbsp;&nbsp;(</span> \
          <span id='conversationRevealQuotes%CONVERSATIONINDEX%' index='%CONVERSATIONINDEX%' class='conversationRevealQuotesLink' collapsed='true'>&msgHdrViewShowQuotes.label;</span> \
          <span id='conversationHideQuotes%CONVERSATIONINDEX%' index='%CONVERSATIONINDEX%' class='conversationHideQuotesLink' collapsed='true'>&msgHdrViewHideQuotes.label;</span> \
          <span id='conversationRevealQuotes%CONVERSATIONINDEX%-separator' class='mailheader' collapsed='true'>&nbsp;|&nbsp;</span> \
          <span id='conversationRevealDetails%CONVERSATIONINDEX%' index='%CONVERSATIONINDEX%' class='conversationRevealDetailsLink'>&pbMsgHdrView.showDetails.label;</span> \
          <span id='conversationHideDetails%CONVERSATIONINDEX%' index='%CONVERSATIONINDEX%' class='conversationHideDetailsLink'>&pbMsgHdrView.hideDetails.label;</span> \
          <span class='mailheader'>)</span> \
        </span> \
      </td> \
    </tr> \
    <tr class='expanded-header-row' id='sender-box' collapsed='true'>                                                      \
      <td class='header-name mailheader'>&senderField.label;</td>             \
      <td class='header-value mailheader-value'>%sendervalue%</td> \
    </tr>                                                     \                                                   \
    <tr class='expanded-header-row' id='replyto-box' collapsed='true'>                                                      \
      <td class='header-name mailheader'>&replyToField.label;</td>             \
      <td class='header-value mailheader-value'>%repytovalue%</td> \
    </tr>                                                     \
    <tr class='expanded-header-row' id='to-box' collapsed='true'>                                                      \
      <td class='header-name mailheader'>&toField.label;</td>             \
      <td class='header-value mailheader-value'>%tovalue%</td> \
    </tr>                                                     \
    <tr class='expanded-header-row' id='cc-box' collapsed='true'>                                                      \
      <td class='header-name mailheader'>&ccField.label;</td>             \
      <td class='header-value mailheader-value'>%ccvalue%</td> \
    </tr>                                                     \
    <tr class='expanded-header-row' id='bcc-box' collapsed='true'>                                                      \
      <td class='header-name mailheader'>&bccField.label;</td>             \
      <td class='header-value mailheader-value'>%bccvalue%</td> \
    </tr>                                                     \
    <tr class='expanded-header-row' id='action-box'> \
      <td class='header-name mailheader'></td> \
      <td class='header-value mailheader-value' messageKey='%messagekeyvalue%' folderURI='%folderurivalue%'> \
        <span class='reply-action action-link'>&pbMsgHdrView.conversationMsgReply.label;</span> \
        <span>&nbsp;&nbsp;</span> \
        <span class='replyall-action action-link'>&pbMsgHdrView.conversationMsgReplyAll.label;</span> \
        <span>&nbsp;&nbsp;</span> \
        <span class='forward-action  action-link'>&pbMsgHdrView.conversationMsgForward.label;</span> \
        <span>&nbsp;&nbsp;</span> \
        <span class='open-action action-link'>&msgHdrViewOpenMessage.label;</span> \
      </td> \
    </tr> \
  </table>                                                   \
</div>";

/** 
  * The template used to display a message We currently show the headers in HTML followed by a separate
  * iframe which contains the actual message body. 
  */
const gMessageTemplate = "<?xml version=\"1.0\" encoding=\"UTF-8\"?> \
<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \
  \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\" [ \
<!ENTITY % headersDTD SYSTEM 'chrome://messenger/locale/msgHdrViewOverlay.dtd' > \
%headersDTD; \
 \
]> \
<html xmlns=\"http://www.w3.org/1999/xhtml\"> \
  <head> \
    <link rel='important stylesheet' href='chrome://global/locale/intl.css'/> \
    <link rel='important stylesheet' href='chrome://messenger/skin/msgDisplayHeaders.css'/> \
  </head> \
  <body> \
    <div style='height:100%'> \
      %MESSAGEHEADERS% \
      <iframe id='%FRAMEID%' src='%MESSAGEBODYURL%' style='height:100%; width:100%; border: none;' scrolling='no'></iframe> \
   </div> \
  </body> \
</html>";

/**
 * template for an e-mail address 
 */
const gEmailAddressTemplate = "<a class='email-link' href='mailto:%emailaddress%'>%displayname%</a>";

function msgDisplayEmitter() {}

msgDisplayEmitter.prototype = {
  _cachedAddresses: [],

  /**
   * If aMessageURIs contains more than one message, displays the selected URIs in converation mode
   * otherwise displays the message normally.
   * @param aMsgHdr the primary msg hdr for the conversation. this is never null andis always in aMessageURIs as well.
   * @param aMessageURIs nsISupportsArray of message URI strings
   * @param aCustomMessages true if the URIs refer to .eml files
   */
  display: function (aMessageURIs, aMsgHdr, aCustomMessages, aConsumer, aMsgWindow, aDisplayOutputFlags)
  {
    this._cachedAddresses = []; // reset our address cache
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                  getService(Components.interfaces.nsIPrefService).getBranch(null);
    var reverseOrder = prefs.getBoolPref('mail.conversation_display.reverse');
    var msgHdrsAndURLs = [];
    for (var index = 0; index < aMessageURIs.length; index++)
      msgHdrsAndURLs.push(getHeaderForURI(aMessageURIs.queryElementAt(index, Components.interfaces.nsISupportsCString).data,
                           aCustomMessages, aMsgWindow));

    // sort the conversation messages with the most recent at the top
    msgHdrsAndURLs.sort(reverseOrder ? compareMsgHdrByDateDescending : compareMsgHdrByDateAscending);

    var primaryMsgHdrUrl;
    var conversation;
    var numVisibleMessages = 0;
    var ioService = Components.classes["@mozilla.org/network/io-service;1"].
                      getService(Components.interfaces.nsIIOService);
    for (index = 0; index < msgHdrsAndURLs.length; index++)
    {
      var message = gConversationMsgTemplate;
      if (!index)
      {
        conversation = msgHdrsAndURLs[index].msgHdr ? this.generateHeadersFromHdr(msgHdrsAndURLs[index].msgHdr, gConversationPrefix) : gConversationPrefix;
        conversation = conversation.replace(/%nummessages%/, msgHdrsAndURLs.length);
      }

      message = message.replace(/%MESSAGEHEADERS%/, this.generateHeadersFromHdr(msgHdrsAndURLs[index].msgHdr, gConversationHeaderTemplate));
      var bodyUrl;
      // if we are offline and we do not have the message body for this message, don't try to load it..
      if (ioService.offline && msgHdrsAndURLs[index].url instanceof Components.interfaces.nsIMsgMailNewsUrl &&
          !msgHdrsAndURLs[index].url.msgIsInLocalCache)
      {
        // load a message informing the user that the body has not been downloaded for offline use
        var offlineMsgBody = '<html><head><meta http-equiv="Content-Type" content="text/html; charset=UTF-8"></head><body>';
        // get the string bundle
        var strBundleService = Components.classes["@mozilla.org/intl/stringbundle;1"].
                                getService(Components.interfaces.nsIStringBundleService);
        var messengerBundle = strBundleService.createBundle("chrome://messenger/locale/messenger.properties");
        var offlineString = messengerBundle.GetStringFromName("nocachedbodyforconversations");
        offlineMsgBody += offlineString + "</body></html>";
        // now encode it as a data url
        bodyUrl = "data:text/html;base64," + btoa(offlineMsgBody);
        message = message.replace(/%MESSAGEBODYURL%/,  bodyUrl);
      }
      else
      {
        bodyUrl = escape(msgHdrsAndURLs[index].url.spec);
        message = message.replace(/%MESSAGEBODYURL%/, escape(msgHdrsAndURLs[index].url.spec));
      }
      
      // make sure aMsgHdr is always open. And if not in aExpandAll mode, make sure every msgHdr that has
      // the same folder as aMsgHdr is open.
      if (aDisplayOutputFlags & Components.interfaces.nsIMsgDisplayEmitter.kExpandAll || (aMsgHdr && msgHdrsAndURLs[index].msgHdr.folder == aMsgHdr.folder))
      {
        message = message.replace(/%CONVERSATIONIFRAME%/, gConversationIframeTemplate);
        message = message.replace(/%MESSAGEBODYURLSRC%/, bodyUrl);
        message = message.replace(/%MESSAGEMODE%/, 'open');
      }
      else
      {
        message = message.replace(/%CONVERSATIONIFRAME%/, gGhostConversationIframeTemplate);
        message = message.replace(/%MESSAGEMODE%/, (aDisplayOutputFlags & Components.interfaces.nsIMsgDisplayEmitter.kShowAllMessages) ? 'slim' : 'ghost');
      }
      
      // is this message in the same folder as aMsgHdr.folder?
      message = message.replace(/%MESSAGECONTAINERPRIMARYFOLDER%/, aMsgHdr && msgHdrsAndURLs[index].msgHdr.folder == aMsgHdr.folder ? 'true' : 'false');
      
      // use numVisibleMessages instead of index here because we may have skipped some items when only showing unread
      // messages in the conversation
      if ((numVisibleMessages + 1) % 2 )
        message = message.replace(/class=\'alternate /, 'class=\'');
      message = message.replace(/%CONVERSATIONINDEX%/g, index);
      conversation += message;
      numVisibleMessages++;
      if (msgHdrsAndURLs[index].msgHdr == aMsgHdr)
        primaryMsgHdrUrl = msgHdrsAndURLs[index].url;
    }
    conversation += gConversationSuffix;
    this.load(primaryMsgHdrUrl, conversation, aConsumer, aMsgWindow);
  },

  /**
   * Helper method - given a string that represents the message document,
   */
  load: function(aUrl, aText, aConsumer, aMsgWindow)
  {
    var stream = Components.classes['@mozilla.org/io/string-input-stream;1'].
                   createInstance(Components.interfaces.nsIStringInputStream);
    stream.setData(aText, aText.length);
    var channel = Components.classes["@mozilla.org/network/input-stream-channel;1"].
                    createInstance(Components.interfaces.nsIInputStreamChannel);
    channel.setURI(aUrl);
    channel.contentStream = stream;
    var baseChannel = channel.QueryInterface(Components.interfaces.nsIChannel).contentType = "application/xhtml+xml";
    baseChannel.contentType = "application/xhtml+xml";
    baseChannel.contentCharset = "UTF-8";
    if (aUrl instanceof Components.interfaces.nsIMsgMailNewsUrl)
      aUrl.msgWindow = aMsgWindow;
    var request = channel.QueryInterface(Components.interfaces.nsIRequest);
    request.loadFlags |= Components.interfaces.nsIChannel.LOAD_DOCUMENT_URI;
    var uriLoader = Components.classes["@mozilla.org/uriloader;1"].
                      getService(Components.interfaces.nsIURILoader);
    uriLoader.openURI(channel, false, aConsumer);
  },

  updateHeader: function(aName, aValue, aFormattedHeader)
  {
    if (aValue)
    {
      // for bug #2680, gConversationHeaderTemplate now contains two headers (class='open-value mailheader' and class='folder-value mailheader')
      // that can open the message in a new tab.  they both contain %messagekeyvalue%, %folderurivalue% and %foldervalue%
      // so I need to call replace twice (or create a regex, but this is good enough.)
      // see bug #2699 for details
      aFormattedHeader = aFormattedHeader.replace('%' + aName + 'value%', aValue);
      aFormattedHeader = aFormattedHeader.replace('%' + aName + 'value%', aValue);
      aFormattedHeader = aFormattedHeader.replace('id=\'' + aName + '-box\' collapsed=\'true\'', 'id=\'' + aName + '-box\'');
    }
    return aFormattedHeader;
  },

  _converter : null,
  
  convertUnicodeToUTF8 : function(aStr)
  {
    if (!this._converter)
    {
      this._converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
      this._converter.charset = 'UTF-8';
    }
    
    var utf8Str = this._converter.ConvertFromUnicode(aStr);   
    utf8Str += this._converter.Finish();
    return utf8Str;
  },
  
  generateHeadersFromHdr: function(aMsgHdr, aHeaderTemplate)
  {
    if (!aMsgHdr)
      return '';
    var formattedHeaders = aHeaderTemplate;

    formattedHeaders = this.updateHeader('date', formatDateHeader(aMsgHdr.dateInSeconds * 1000), formattedHeaders);
    formattedHeaders = this.updateHeader('from', this.formatAddresses(aMsgHdr.mime2DecodedAuthor), formattedHeaders);
    formattedHeaders = this.updateHeader('to', this.formatAddresses(aMsgHdr.mime2DecodedRecipients), formattedHeaders);
    formattedHeaders = this.updateHeader('cc', this.formatAddresses(aMsgHdr.mime2DecodedCcList), formattedHeaders);
    formattedHeaders = this.updateHeader('bcc', this.formatAddresses(aMsgHdr.getStringProperty('bcc')), formattedHeaders);
    // we need to escape and convert the folder name and uri
    // see bug #3195 for details
    formattedHeaders = this.updateHeader('folder', escape(this.convertUnicodeToUTF8(aMsgHdr.folder.prettiestName)), formattedHeaders);
    formattedHeaders = this.updateHeader('folderuri', escape(aMsgHdr.folder.URI), formattedHeaders);
    formattedHeaders = this.updateHeader('messagekey', aMsgHdr.messageKey, formattedHeaders);
    return formattedHeaders;
  },

  /**
   * @param aFullAddresses comma delimited listed of addresses
   */ 
  formatAddresses: function(aFullAddresses)
  {
    let lowerCaseFullAddresses = aFullAddresses ? aFullAddresses.toLowerCase() : '';
    // we hash the e-mail address
    if (this._cachedAddresses[lowerCaseFullAddresses])
      return this._cachedAddresses[lowerCaseFullAddresses];

    // convert unicode to utf-8
    let convertedFullAddresses = this.convertUnicodeToUTF8(aFullAddresses);
    var content = "";
    var msgHeaderParser = Components.classes["@mozilla.org/messenger/headerparser;1"].
                            getService(Components.interfaces.nsIMsgHeaderParser);
    var addresses = {};
    var fullNames = {};
    var names = {};
    var numAddresses =  0;
    numAddresses = msgHeaderParser.parseHeadersWithArray(convertedFullAddresses, addresses, names, fullNames);
    var index = 0;
    while (index < numAddresses)
    {
      if (index)
        content += ", ";
      content += formatAddress(addresses.value[index], fullNames.value[index], names.value[index], this.convertUnicodeToUTF8);
      index++;
    }
    // store the fragment in our address cache
    this._cachedAddresses[lowerCaseFullAddresses] = content;
    return content;
  },

  QueryInterface : function (iid)
  {
    if (iid.equals(Components.interfaces.nsIMsgDisplayEmitter) || 
        iid.equals(Components.interfaces.nsISupports))
      return this;
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
};

/**
 * returns the msgHdr (can be null) and the nsIMsgMessageUrl
 */
function getHeaderForURI(aURI, aCustomMessage, aMsgWindow)
{
  var msgHdr;
  var url;
  if (aCustomMessage)
  {
    var ioService = Components.classes["@mozilla.org/network/io-service;1"].
                      getService(Components.interfaces.nsIIOService);
    url = ioService.newURI(aURI, null, null).QueryInterface(Components.interfaces.nsIMsgMessageUrl);
    msgHdr = null; // for now, we don't have access to any header data when opening a .eml file
  }
  else
  {
    var protocol = aURI.substring(0, aURI.indexOf(':'));
    var msgService = Components.classes["@mozilla.org/messenger/messageservice;1?type=" + protocol].
                       createInstance(Components.interfaces.nsIMsgMessageService);
    url = msgService.GetUrlForUri(aURI, aMsgWindow).QueryInterface(Components.interfaces.nsIMsgMessageUrl);
    msgHdr = url.messageHeader;
  }
  return {msgHdr: msgHdr, url: url};
}

function getDisplayNameFromAddressbook(aEmailAddress)
{  
  var RDF = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService().QueryInterface(Components.interfaces.nsIRDFService);
  var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(null);
  
  var parentDir = RDF.GetResource("moz-abdirectory://").QueryInterface(Components.interfaces.nsIAbDirectory);
  var enumerator = parentDir.childNodes;
  var card = null;
  
  while (!card && enumerator.hasMoreElements())
  {
    var addrbook = enumerator.getNext();
    var addrbookURI = addrbook.QueryInterface(Components.interfaces.nsIRDFResource).Value;
    // don't look in the collected address book when trying to get the display name
    var collectedAddressBookURI = prefs.getCharPref('mail.collect_addressbook');
    if ((addrbookURI != collectedAddressBookURI || collectedAddressBookURI == "moz-abmdbdirectory://abook.mab") &&
        addrbook instanceof Components.interfaces.nsIAbMDBDirectory)
    {
      card = addrbook.cardForEmailAddress(aEmailAddress);
    }
  }

  return (card ? card.displayName : "");
}  

function formatAddress(aEmailAddress, aFullName, aDisplayName, aConverterFunction)
{
  var formattedAddress = gEmailAddressTemplate;
  formattedAddress = formattedAddress.replace(/%emailaddress%/, escape(aEmailAddress));
  
  // use the display name from the addressbook, otherwise use the display from the email
  // if no display name from the email, use the email address.
  var displayName = getDisplayNameFromAddressbook(aEmailAddress);
  if (displayName)
    displayName = aConverterFunction(displayName);
  else
    displayName = aDisplayName || aEmailAddress;
  
  return formattedAddress.replace(/%displayname%/, escape(displayName));
}

function formatDateHeader(aDate)
{
  var dateService = Components.classes["@mozilla.org/intl/scriptabledateformat;1"].
                      getService(Components.interfaces.nsIScriptableDateFormat);
  var msgDate = new Date(aDate);
  var currentDate = new Date();
  var dateFormat = dateService.dateFormatShort;
  // same day?
  if (currentDate.getFullYear() == msgDate.getFullYear() &&
      currentDate.getMonth() == msgDate.getMonth() &&
      currentDate.getDate() == msgDate.getDate())
    dateFormat = dateService.dateFormatNone;
  return dateService.FormatDateTime("", dateFormat,
                             dateService.timeFormatNoSeconds,
                             msgDate.getFullYear(), msgDate.getMonth()+1, msgDate.getDate(),
                             msgDate.getHours(),msgDate.getMinutes(), msgDate.getSeconds());
}

/**
 * text nodes in XHTML cannot have &, > or < characters. We must escape those characters
 * with the appropriate entities
 * @param aValue the string to escape
 */
function escape(aValue)
{
  aValue = aValue.replace(/&/g, "&amp;");
  aValue = aValue.replace(/>/g, "&gt;");
  aValue = aValue.replace(/</g, "&lt;");
  aValue = aValue.replace(/"/g, "&quot;");
  return aValue.replace(/'/g, "&apos;");
}

/**
 * Helper method used to sort an array of msgHds by date with the most *recent* first
 */
function compareMsgHdrByDateDescending(aNode1, aNode2)
{
  return aNode2.msgHdr.dateInSeconds - aNode1.msgHdr.dateInSeconds;
}

/**
 * Helper method used to sort an array of msgHds by date with the most *recent* first
 */
function compareMsgHdrByDateAscending(aNode1, aNode2)
{
  return aNode1.msgHdr.dateInSeconds - aNode2.msgHdr.dateInSeconds;
}

/** 
 * XPCOM Registration
 */
 
const msgDisplayEmitter_contractID = "@mozilla.org/messenger/displayemitter;1";
const msgDisplayEmitter_CID = Components.ID("{6708B6F6-664F-41a7-94F9-B6C589E8D1D7}");

var msgDisplayEmitterFactory  = {
  createInstance : function (outer, iid) {
    if (outer != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return new msgDisplayEmitter().QueryInterface(iid);
  }
}

var Module = {
  /* nsISupports */
  QueryInterface : function (iid) {
    if (iid.equals(Components.interfaces.nsIModule) &&
        iid.equals(Components.interfaces.nsISupports))
      return this;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  /* nsIModule */
  getClassObject : function (compMgr, cid, iid) {
    if (cid.equals(msgDisplayEmitter_CID))
      return msgDisplayEmitterFactory;
    throw Components.results.NS_ERROR_FAILURE;
  },

  registerSelf: function (compMgr, fileSpec, location, type) {
    var compReg =
      compMgr.QueryInterface( Components.interfaces.nsIComponentRegistrar );

    compReg.registerFactoryLocation(msgDisplayEmitter_CID,
                                    "msgDisplayEmitter",
                                    msgDisplayEmitter_contractID,
                                    fileSpec,
                                    location,
                                    type);
  },

  unregisterSelf : function (compMgr, location, type) {
    var compReg = compMgr.QueryInterface(nsIComponentRegistrar);
    compReg.unregisterFactoryLocation(msgDisplayEmitter_CID, location);
  },

  canUnload: function(compMgr) {
    return true;
  }
}

function NSGetModule(compMgr, fileSpec) {
  return Module;
}
