/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/***** BEGIN POSTBOX LICENSE BLOCK *****
 *
 * nsAddressbookInitializer.js
 *
 * Copyright (c) 2007-2009 Postbox, Inc.
 *
 ***** END POSTBOX LICENSE BLOCK *****/

// Module object
var AddressbookInitializerMod = {
  firstTime: true, 
  cid: Components.ID("{376149b0-75de-4992-8c4d-350ffdd61ffb}"),
  progid: "@mozilla.org/addressbook-initializer;1",
  className: "Addressbook Initializer",

factory:  
  {
  createInstance: function (aOuter, aIID) 
    {
      if (aOuter != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
      if (!aIID.equals(Components.interfaces.nsISupports))
        throw Components.results.NS_ERROR_INVALID_ARG;
      
      // return the singleton
      var singletonAddressbookInitializer = new AddressbookInitializer;
      return singletonAddressbookInitializer.QueryInterface(aIID);
    }       
  }, // factory

  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    
      if (aCID.equals(this.cid))
        return this.factory;
    
    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {        
    if (this.firstTime) 
    {
      this.firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(this.cid, this.className, this.progid, aFileSpec, aLocation, aType);  
  },

  unregisterSelf: function(aCompMgr, aFileSpec, aLocation)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(this.cid, aFileSpec);  
  },

  canUnload: function(aCompMgr)
  {
    return true;
  }
};

function NSGetModule(aCompMgr, aFileSpec)
{
  return AddressbookInitializerMod;
}

/* keep in sync with nsMsgFolderFlags.h */
var MSG_FOLDER_FLAG_INBOX = 0x1000;
var MSG_FOLDER_FLAG_SENTMAIL = 0x0200;
/* keep in sync with nsMsgMessageFlags.h */
var MSG_FLAG_HAS_RE = 0x000010;
  
const kSecondsInADay = 86400;

var gMaxAgeInSeconds = 0;
  
function GetMaxAgeInSeconds()
{
  if (!gMaxAgeInSeconds)
  {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(null);
    var maxAgeInDays = prefs.getIntPref("mail.auto_collected_addresses.max_age_in_days");
    gMaxAgeInSeconds = kSecondsInADay * maxAgeInDays;
  }
  return gMaxAgeInSeconds;
}

function CollectAddresses(aMsgHdr)
{
  // sic:  yes, should be collector
  var abAddressCollecter = Components.classes["@mozilla.org/addressbook/services/addressCollecter;1"].getService(Components.interfaces.nsIAbAddressCollecter);
  
  var now = Math.floor(Date.now() / 1000);
  if (now - aMsgHdr.dateInSeconds > GetMaxAgeInSeconds())
  {
    // this header is too old, so ignore the addresses in it
    return;
  }
  
  if (aMsgHdr.folder.flags & MSG_FOLDER_FLAG_INBOX)
  {
    var junkScore = aMsgHdr.getStringProperty("junkscore");
    var isJunk = (junkScore != "") && (junkScore != "0");
    if (!(aMsgHdr.flags & MSG_FLAG_HAS_RE) || isJunk)
    {
      // only autocollect the authors of messages in the inbox if they are replies (and not spam).
      // we can't use MSG_FLAG_REPLIED, as none of our messages will have that (unless they are migrating from tbird).
      // so we use MSG_FLAG_HAS_RE as an best guess.
      return;
    }
  }
        
  try
  { 
    abAddressCollecter.collectUnicodeAddress(aMsgHdr.mime2DecodedAuthor, true /* aCreateCard */, Components.interfaces.nsIAbPreferMailFormat.unknown /* aSendFormat */);
  }
  catch (ex)
  {
    dump("XXX failed to collect author, ex = " + ex + "\n");
  }

  try
  {
    abAddressCollecter.collectUnicodeAddress(aMsgHdr.mime2DecodedRecipients, true /* aCreateCard */, Components.interfaces.nsIAbPreferMailFormat.unknown /* aSendFormat */);
  }
  catch (ex)
  {
    dump("XXX failed to collect recipients, ex = " + ex + "\n");
  }

  try
  {
    abAddressCollecter.collectUnicodeAddress(aMsgHdr.mime2DecodedCcList, true /* aCreateCard */, Components.interfaces.nsIAbPreferMailFormat.unknown /* aSendFormat */);
  }
  catch (ex)
  {
    dump("XXX failed to collect cc list, ex = " + ex + "\n");
  }
}

const kAutoCollectedPrefPrefix = "mail.auto_collected_addresses.";

function NeverCollectFromFolderAgain(aURI)
{
  var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(null);
  prefs.setBoolPref(kAutoCollectedPrefPrefix + aURI, true);
}

function AlreadyCollectedFromFolder(aURI)
{
  try {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(null);
    if (prefs.getBoolPref(kAutoCollectedPrefPrefix + aURI))
      return true;
  }
  catch (ex)
  {
    // pref may not exist
  }
  return false;
}


// constructor for AddressbookInitializer
function AddressbookInitializer()
{
  this.observerService =
    Components.classes["@mozilla.org/observer-service;1"]
              .getService(Components.interfaces.nsIObserverService);

  this.observerService.addObserver(this, "profile-before-change", false);
  this.observerService.addObserver(this, "quit-application", false);
  
  this.notificationService = Components.classes["@mozilla.org/messenger/msgnotificationservice;1"].getService(Components.interfaces.nsIMsgFolderNotificationService);
  this.notificationService.addListener(this);
}

AddressbookInitializer.prototype = {
  QueryInterface: function(aIID)
  {
    if (aIID.equals(Components.interfaces.nsISupports))
      return this;
    else if (aIID.equals(Components.interfaces.nsIMsgFolderListener))
      return this;
    
    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
  },
  _currentFolderURI: null,
  _ignoreCurrentFolder: false,
  _currentFolderIsSentOrInbox: false,
  shutdown: function()
  {
    // on shutdown, make sure to remember to stop collecting from the current uri (if it was a sent or inbox)
    // see bug #1624 for details
    if (this._currentFolderURI && this._currentFolderIsSentOrInbox)
      NeverCollectFromFolderAgain(this._currentFolderURI);
      
    this.notificationService.removeListener(this);
  },
  observe: function (aSubject, aTopic, aData) 
  {
    switch (aTopic) 
    {
      case "profile-before-change":
        this.shutdown();
        break;
      case "quit-application":
        this.shutdown();
        this.observerService.removeObserver(this, "quit-application");
        this.observerService.removeObserver(this, "profile-before-change");
        break;
    }
  },
  itemAdded: function(aItem)
  {
    var msgHdr;
    try
    {
      msgHdr = aItem.QueryInterface(Components.interfaces.nsIMsgDBHdr);
    }
    catch (ex) {}

    if (msgHdr)
    {
      if (this._currentFolderURI && this._currentFolderURI == msgHdr.folder.URI)
      {
        if (this._ignoreCurrentFolder)
        {
          // this is currently the folder we are ignoring, so bail out
          return;
        }

        // if we aren't ignoring this folder, collect the addresses
        CollectAddresses(msgHdr);
      }
      else if (msgHdr.folder.flags & MSG_FOLDER_FLAG_INBOX || msgHdr.folder.flags & MSG_FOLDER_FLAG_SENTMAIL)
      {
        // if we were not ignoring the current folder, and it is an inbox or sent folder, then we need to remember not to collect from that folder ever again.
        if (!this._ignoreCurrentFolder && this._currentFolderURI && this._currentFolderIsSentOrInbox)
          NeverCollectFromFolderAgain(this._currentFolderURI);

        this._currentFolderURI = msgHdr.folder.URI;
        this._currentFolderIsSentOrInbox = true;

        if (AlreadyCollectedFromFolder(this._currentFolderURI))
        {
          // if we've already collected from the new "current folder", don't collect the address and start ignoring this folder.
          this._ignoreCurrentFolder = true;
        }
        else
        {
          // this is an inbox or a sent folder we've never collected from, so don't ignore it (so that we'll collect more headers)
          // and collect this header.
          this._ignoreCurrentFolder = false;
          CollectAddresses(msgHdr);
        }
      }
      else
      {
        // if we were not ignoring the current folder, and it is an inbox or sent folder, then we need to remember not to collect from that folder ever again.
        if (this._currentFolderURI && this._currentFolderIsSentOrInbox)
          NeverCollectFromFolderAgain(this._currentFolderURI);

        this._currentFolderURI = msgHdr.folder.URI;
        this._currentFolderIsSentOrInbox = false;
        // we only care about sent or inbox, so ignore this folder
        this._ignoreCurrentFolder = true;
      }
    }
  },
  itemDeleted: function(aItem)
  {
  },
  itemMoveCopyCompleted: function(aMove, aSrcItems, aDestFolder)
  {
  },
  folderRenamed: function(aOrigFolder, aNewFolder)
  {
  },
  // extensibility hook
  itemEvent: function(aItem, aEvent, aData)
  {
  }, 
};
