/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/***** BEGIN POSTBOX LICENSE BLOCK *****
 *
 * nsMsgFullTextIndexer.js
 *
 * Copyright (c) 2007-2009 Postbox, Inc.
 *
 ***** END POSTBOX LICENSE BLOCK *****/

const Cc = Components.classes;
const Ci = Components.interfaces;
const MSG_DB_LARGE_COMMIT = 1;
const kCurDBVersion = 8;

var gMessenger;
var gMozStorage;

var gMsgTextTablePartialStorageStmt;
var gMsgTextTableStorageStmt;
var gMsgTextTableUpdateStmt;
var gMsgTableStorageStmt;
var gFolderTableStorageStmt;
var gFolderURILookupStmt;
var gMsgTxtLookupStmt;  
var gFrecencyLookupStmt;
var gMessageLookupStmt;
var gFrecencyUpdateStmt;
var gMsgMovedStmt;
var gUrlInfoStorageStmt;
var gAttachmentsStorageStmt;
var gAttachmentsLookupStmt;
var gBodyStorageStmt;
var gBodyLookupStmt;
var gEventLookupStmt;
var gEventStorageStmt;
var gTotalMessageCountForFolder;
var gTotalCountForFolder;
var gTotalCount;
var gTotalPendingDeleteCount
var gFolderLookupStmt;
var gDistinctMessageIdAndFolderLookupStmt;
var gAllMessageIdsInFolder;

var gRefsStorageStmt;
var gRefsLookupStmt;

var gTagsStorageStmt;
var gTagsLookupStmt;
var gTagsRemoveStmt;
var gTagLabelLookupStmt;
var gTagLabelStorageStmt;

var gMimeTypesStorageStmt;
var gMimeTypesLookupStmt;
var gMimeTypesRemoveStmt;
var gMimeTypeNameLookupStmt;
var gMimeTypeNameStorageStmt;

var gPigLogToDisk = false;

var gMsgTextIndexer;

var gDBConn;

var gCurrentFolderToIndex = null;
var gSortedHeadersToIndex;

var gLastFolderIndexedUri = ""; // this is stored in the pref "mail.fulltextsearch.lastFolderIndexedUri"
const kLastFolderIndexPrefName = "mail.fulltextsearch.lastFolderIndexedUri"
var gPrefBranch;
var gIndexMsgs;
var gAlarm;
var gCyclingThroughFoldersState = "start";
var gMessenger;
var gDelayedTransactionAlarm;
var gIndexingOneFolder = false;

const kUrlIndexingDelimiter = '\t';
const kDelimiterReplacement = ' ';
const kUrlReplacerRegEx = /\t/g;

var gUrlDetector = Cc["@mozilla.org/txttohtmlconv;1"].createInstance(Ci.mozITXTToHTMLConv);
var gMessengerStringBundle = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService).createBundle("chrome://messenger/locale/messenger.properties");
var gAccountManager = Cc["@mozilla.org/messenger/account-manager;1"].getService(Ci.nsIMsgAccountManager);
var gIOService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
var gFileHandler = gIOService.getProtocolHandler("file").QueryInterface(Ci.nsIFileProtocolHandler);
var gConsoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);

var gUnicodeConverter = Cc["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Ci.nsIScriptableUnicodeConverter);
gUnicodeConverter.charset = "UTF-8";
  
var gIndexingFeedback = null;
var gIgnoreIdle = false;
var gDontIdleIndex = false;
var gSuggestIndexing = true;
var gIndexing = false;
var gUniqueId = 0;

var gMsgHdrsToIndex = Cc["@mozilla.org/supports-array;1"].createInstance(Components.interfaces.nsISupportsArray);
var gBatchedHeadersToAdd = [];
var gBatchedHeadersToAddFolderId = -1;

var gBatchMaxWithBodies;
var gBatchMaxWithoutBodies;  
var gBatchDelay;

function CloseFolderDatabaseIfNotInUse(aMsgFolder)
{
  if (!(aMsgFolder.flags & MSG_FOLDER_FLAG_INBOX) &&  
      !Cc["@mozilla.org/messenger/services/session;1"].getService(Ci.nsIMsgMailSession).IsFolderOpenInWindow(aMsgFolder))
  {
    PigLog("set db to null for " + aMsgFolder.URI);
    aMsgFolder.setMsgDatabase(null);
  }
}

function SetCurrentFolderToIndex(aFolder)
{ 
  if (gCurrentFolderToIndex)
  {
    CloseFolderDatabaseIfNotInUse(gCurrentFolderToIndex);
  }
  gCurrentFolderToIndex = aFolder;
}

function GetNextUniqueStringForSortingInternal(aNum, aUniqueString)
{
  return PadWithZeros(aNum) + "." + aUniqueString;
}

function GetNextUniqueStringForSorting(aNum)
{
  gUniqueId++;
  return GetNextUniqueStringForSortingInternal(aNum, PadWithZeros(gUniqueId));
}

var gIdleIndexProgressFeedback = {
  _getLast3Pane : function ()
  {
    try {
      var mailWindow = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService()
                       .QueryInterface(Components.interfaces.nsIWindowMediator)
                       .getMostRecentWindow("mail:3pane");

      if (mailWindow)
        return mailWindow;
    }
    catch (ex)
    {
    }
    return null;
  },
 
  forceGC : function ()
  {
    // this will force the cycle collector to collect
    try 
    {
      var mailWindow = this._getLast3Pane();
      if (mailWindow)
        mailWindow.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIDOMWindowUtils).garbageCollect();
    }
    catch (ex)
    {
      PigLog("XXX forceGC failed, ex = " + ex);
    }
  },
  
  onStart : function ()
  {
    this.onStatus(gMessengerStringBundle.GetStringFromName("startIndexing"));

    /* the meteors are starting, but sometimes not stopping.  until I figure that out, turning them off
       see bug #1380 for details
    var mailWindow = this._getLast3Pane();
    if (mailWindow)
      mailWindow.statusFeedback.startMeteors();
    */
  },
  onStatus : function (aText)
  {
    var mailWindow = this._getLast3Pane();
    if (mailWindow)
      mailWindow.statusFeedback.showStatusString(aText);
  },
  onStop : function ()
  {
    /* the meteors are starting, but sometimes not stopping.  until I figure that out, turning them off
       see bug #1380 for details
    var mailWindow = this._getLast3Pane();
    if (mailWindow)
      mailWindow.statusFeedback.stopMeteors();
    */
    
    // fix for bug #2963
    // force GC when done idle indexing that when the user comes back from idle, we don't GC and block the UI
    this.forceGC();
    
    // on stop, clear status.
    this.onStatus("");
  },
  onSearchingFolder : function (aMsgFolder)
  {    
    // when idle indexing, only show "Indexing..."
    // see bug #1220 for details.
    this.onStatus(gMessengerStringBundle.GetStringFromName("indexingMessageOnIdle"));
    
    // fix for bug #2963
    // force GC after indexing every folder so that when the user comes back from idle, we don't GC and block the UI
    this.forceGC();
  },
  onPreparingToIndex : function (aMsgFolder)
  {
    // when idle indexing, only show "Indexing..."
    // see bug #1220 for details.
    this.onStatus(gMessengerStringBundle.GetStringFromName("indexingMessageOnIdle"));
  },
  onDownloadingMessage : function (aMsgHdr)
  {
    // when idle indexing, only show "Indexing..."
    // see bug #1220 for details.
    this.onStatus(gMessengerStringBundle.GetStringFromName("indexingMessageOnIdle"));
  }, 
  onIndexingMessage : function (aMsgHdr)
  {
    // when idle indexing, only show "Indexing..."
    // see bug #1220 for details.
    this.onStatus(gMessengerStringBundle.GetStringFromName("indexingMessageOnIdle"));
  }
};

function convertUrlsAndTitlesToIndexableString(aURLs, aTitles)
{
  var urlIndexString = '';
  var addedURLs = [];
  for (var index = 0; index < aURLs.length; index++)
  {
    var url = aURLs[index];
    if (/^http(s)?:/i.test(url)) // only index http and https urls
    {
      var linkText = aTitles[index];
      var linkTextForComparison = linkText;
      linkTextForComparison = linkTextForComparison.replace(/ /g, "");
      if (linkTextForComparison && (/^http(s)?:/i.test(linkTextForComparison)))
      {
        var ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        var linkTextURI = ioService.newURI(linkTextForComparison, null, null);
        var linkURI = ioService.newURI(url, null, null);
        // compare hosts, but ignore possible www. prefix
        if (linkURI.host.replace(/^www\./, "") == linkTextURI.host.replace(/^www\./, ""))
          linkText = '';
      }

      var key = linkText.replace(kUrlReplacerRegEx, kDelimiterReplacement) + kUrlIndexingDelimiter + url.replace(kUrlReplacerRegEx, kDelimiterReplacement);
      // make sure we haven't already added this url
      if (addedURLs[key])
        continue;
      if (urlIndexString)
        urlIndexString += kUrlIndexingDelimiter;
      // to do - avoid inserting the link text if linkText == link.href. But what do we insert to replace it?
      // for each url, insert the linktext / url pair delimited by kUrlIndexingDelimiter
      // we convert the occurrences of kUrlIndexingDelimiter to a space
      addedURLs[key] = true;
      urlIndexString += key;
    }
  }

  // convert urlIndexString to UTF-8
  var urlIndexStringAsUtf8 = gUnicodeConverter.ConvertFromUnicode(urlIndexString);
  urlIndexStringAsUtf8 += gUnicodeConverter.Finish();
  return urlIndexStringAsUtf8;
}

// Module object
var GlobalIndexerMod = {
  firstTime:  true,
  cid: Components.ID("{4c268f4a-6755-4678-88f3-541e5ce017df}"),
  progid: "@mozilla.org/msg-global-index;1",
  className: "Msg Global Indexer",

factory:  
  {
  createInstance: function (aOuter, aIID) 
    {
      if (aOuter != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
      
      if (!aIID.equals(Ci.nsISupports) &&
          !aIID.equals(Ci.nsIMsgGlobalIndex))
        throw Components.results.NS_ERROR_INVALID_ARG;
      // return the singleton
      var singletonGlobalIndexer = new msgGlobalIndexer;
      return singletonGlobalIndexer.QueryInterface(aIID);
    }       
  }, // factory
  

  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Ci.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    
      if (aCID.equals(this.cid))
        return this.factory;
    
    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  { 
    if (this.firstTime) 
    {
      this.firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    
    aCompMgr = aCompMgr.QueryInterface(Ci.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(this.cid, this.className, this.progid, aFileSpec, aLocation, aType);  
    
    var catMan = Components.classes["@mozilla.org/categorymanager;1"]
                           .getService(Components.interfaces.nsICategoryManager);
                            
    catMan.addCategoryEntry("postbox-mail-url-listeners",
                            this.progid, this.progid, true, true);
  },

  unregisterSelf: function(aCompMgr, aFileSpec, aLocation)
  {
    aCompMgr = aCompMgr.QueryInterface(Ci.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(this.cid, aFileSpec);  
    
    var catMan = Components.classes["@mozilla.org/categorymanager;1"]
                           .getService(Components.interfaces.nsICategoryManager);
                           
    catMan.deleteCategoryEntry("postbox-mail-url-listeners",
                               this.progid, true);
  },

  canUnload: function(aCompMgr)
  {
    return true;
  }
};
                            
function NSGetModule(aCompMgr, aFileSpec)
{
  return GlobalIndexerMod;
}

/* keep in sync with nsMsgMessageFlags.h */
const MSG_FLAG_ATTACHMENT = 0x10000000;
const MSG_FLAG_MARKED = 0x000004;
const MSG_FLAG_OFFLINE = 0x0080;

// how long (in seconds) the system is idle before we'll do background indexing and other work
// set in constructor for msgGlobalIndexer
var gIdleTime;

// how long (in seconds) to wait before lazily updating a message frecency, urlinfo, attachmentinfo, or bodyinfo
// this is done to allow for batching and to prevent delay while reading conversations.
// set in constructor for msgGlobalIndexer
var gLazyUpdateTime;

function WontPromptForPassword(aServer)
{
  return (!aServer.serverBusy && (!aServer.serverRequiresPasswordForBiff || !aServer.passwordPromptRequired));
}

// constructor for msgGlobalIndexer
function msgGlobalIndexer()
{
  this._idleService = Cc["@mozilla.org/widget/idleservice;1"].getService(Ci.nsIIdleService);
  this.observerService =
    Components.classes["@mozilla.org/observer-service;1"]
              .getService(Components.interfaces.nsIObserverService);

  this.observerService.addObserver(this, "profile-before-change", false);
  this.observerService.addObserver(this, "quit-application", false);
  InitGlobalIndexingIntegration();

  var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(null);
  gIdleTime = prefs.getIntPref("mail.fulltextsearch.idle_time");
  gLazyUpdateTime = prefs.getIntPref("mail.fulltextsearch.lazy_update_time");
  gPigLogToDisk = prefs.getBoolPref("mail.fulltextsearch.log_to_disk");

  gBatchMaxWithBodies = prefs.getIntPref("mail.fulltextsearch.batch_max.with_bodies");
  gBatchMaxWithoutBodies = prefs.getIntPref("mail.fulltextsearch.batch_max.without_bodies");
  gBatchDelay = prefs.getIntPref("mail.fulltextsearch.batch_max.delay");

  // set up the idle timer after we InitGlobalIndexingIntegration() which will initialize gMsgTextIndexer
  // see bug #1038 for details
  this._idleService.addIdleObserver(this, gIdleTime);

  this.log("start up");
  this.log("idle time = " + gIdleTime);
  this.log("lazy update time = " + gLazyUpdateTime);
  this.logTags();

  // see bug #2339 for details
  gDoneCreatingIndexerService = true;
}

msgGlobalIndexer.prototype = {
  QueryInterface: function(aIID)
  {
    if (aIID.equals(Ci.nsISupports))
      return this;
    else if (aIID.equals(Ci.nsIMsgGlobalIndex))
      return this;
    else if (aIID.equals(Ci.nsIUrlListener))
      return this;
    else if (aIID.equals(Ci.nsIObserver))
      return this;
    
    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
  },

  observe: function (aSubject, aTopic, aData) 
  {
    switch (aTopic) 
    {
      case "idle":
        PigLog("observing idle, gIndexing = " + gIndexing + " number urls running = " + this._runningUrls.length + " offline = " + gIOService.offline);
        
        // if we are ignoring idle (gIgnoreIdle), we are manually indexing, so don't start indexing on idle.
        // if we are not ignoring idle, and we get the "idle" notification, we should only start indexing if:
        //   we aren't currently indexing (gIndexing)
        //   we aren't currently running a url (this._runningUrls.length)
        //   we aren't in the middle of a chunking search or loading msgfti.sqlite into the os fs cache (gDontIdleIndex)
        // don't index if we are offline (we're online)
        if (!gIgnoreIdle && !gIndexing && !this._runningUrls.length && !gDontIdleIndex && !gIOService.offline)
        {
          PigLog("start indexing due to idle");
          // proces folders waiting for update before indexing.
          this.processFoldersWaitingForUpdate();

          this.setFeedback(gIdleIndexProgressFeedback);
          this.startIndexing();
        }
        break;
      case "profile-before-change":
        this.shutdown();
        break;
      case "quit-application":
        this.shutdown();
        this.log("quit-application");
        this.observerService.removeObserver(this, "quit-application");
        this.observerService.removeObserver(this, "profile-before-change");
        break;
    }
  },

  shutdown : function ()
  {    
    this.stopIndexing();
    this._idleService.removeIdleObserver(this, gIdleTime);

    if (gDelayedTransactionAlarm)
    {
      gDelayedTransactionAlarm.cancel();
    }
  },
  
  hitTypeAllowed : function (aPostProcessor, aHitType)
  {
    return (!aPostProcessor || aPostProcessor.hitTypeAllowed(aHitType));
  },

  postProcessAndNotifyListener : function (aListener, aPostProcessor, aMsgHdr, aHitType, aMatchData, aSortValue, aBodyText)
  {
    if (aHitType == Ci.nsIMsgGlobalSearchNotify.urlHit)
    {
      var returnValue = true;
      var urls = aMsgHdr.getStringProperty("urls");
      if (urls == "")
        return returnValue;

      var titlesAndUrls = urls.split("\t");
      for (var i = 0; i < titlesAndUrls.length; i+=2)
      {
        if (!aListener.onSearchHit(aMsgHdr, aHitType, titlesAndUrls[i] + "\t" + titlesAndUrls[i+1], aSortValue, "0.0", null, aBodyText))
          returnValue = false;
      }
      // for url hits, don't stop reporting if we are in the middle of a message
      // wait until we are done with the current message.  
      // see bug #952 for details
      return returnValue;
    }
    else if (aHitType == Ci.nsIMsgGlobalSearchNotify.eventHit)
    {
      if (!aListener.onSearchHit(aMsgHdr, aHitType, "", aSortValue, aMatchData.date, aMatchData.text, aBodyText))
        return false;
    }
    else
    {
      if (!aListener.onSearchHit(aMsgHdr, aHitType, aMatchData, aSortValue, "0.0", null, aBodyText))
        return false;
    }
 
    return true;
  },
  determineMaxPossibleResults : function(aQueryString, aLimit)
  {
    var start = Date.now();
    // any code that does an executeStep must end any current transaction
    EndTransaction(); 

    // bind is not working with fts
    // so we have to manually escape single quotes
    // note, we have to worry about other stuff, like unsafe binary values?
    // http://www.phptutorial.info/?sqlite-escape-string
    // see bug #1178 for details
    var countSql = aQueryString.replace(/\?3/g, aLimit);
    
    PigLog("determineMaxPossibleResults: " + countSql);

    var hasMoreStmt = getDBConn().createStatement(countSql);
    var maxPossibleResults = 0;

    if (hasMoreStmt.executeStep())
      maxPossibleResults = hasMoreStmt.getInt32(0);
    hasMoreStmt.reset();
    PigLog("determineMaxPossibleResults took " + (Date.now() - start) + "ms");
    return maxPossibleResults;
  },

  get dontIdleIndex()
  {
    return gDontIdleIndex;
  },

  set dontIdleIndex(aVal)
  {
    PigLog("setting dontIdleIndex to " + aVal);
    gDontIdleIndex = aVal;
  },

  get suggestIndexing()
  {
    return gSuggestIndexing;
  },

  set suggestIndexing(aVal)
  {
    PigLog("setting suggestIndexing to " + aVal);
    gSuggestIndexing = aVal;
  },

  executeFullTextQuery : function(aSQLSafeSearchString, aQueryString, aChunkSize, aSortColumn, aPostProcessor, aListener)
  {
    // process any batched headers so that if we just discovered those headers, they'll show up on search
    // and process any lazy items so that they show up when we search
    // see bug #1040 for details
    ProcessBatchedHeadersAndLazyTables();

    // XXX sspitzer todo:
    // this check can take about 150 - 200 ms!
    // make it faster
    var start = Date.now();
    // check if there are messages in this chunk.  
    // if not, we can stop, as we're out of messages.
    var hasMoreStmt = getDBConn().createStatement(aQueryString);
    var atLeastOneMessageInThisChunk = hasMoreStmt.executeStep();
    hasMoreStmt.reset();
    PigLog("executeFullTextQuery: [" + aQueryString + "] took " + (Date.now() - start) + "ms\n");
    if (!atLeastOneMessageInThisChunk)
      return false;

    var limitToEvents = (aPostProcessor && aPostProcessor.limitToEvents);
    var limitToContacts = (aPostProcessor && aPostProcessor.limitToContacts);

    var boundQueryString;
    var hasMatchCondition = (aPostProcessor && aPostProcessor.matchCondition != "?2");
    
    PigLog("aQueryString = [" + aQueryString + "]");
    
    // IMPORTANT: 
    // for performance, use message.rowid for contact searches and use
    // message_text.docid for other searches, to avoid full table scans.
    // see bug #3122 and #3400 for details
    var messageIdInSQL = " WHERE " + (limitToContacts ? "message.rowid" : "message_text.docid") + " IN (" + aQueryString + ")";
    
    // do some query rewriting for performance
    // see bug #3030, for details
    if (aChunkSize || limitToContacts)
    {
      // we are unable to re-write contact queries or if we have a chunkSize, so keep those queries as is.
    }
    else if (aQueryString.indexOf("SELECT message.rowid FROM message WHERE message.folder_id > -1") == 0)
    {
      // convert [SELECT message.rowid FROM message WHERE message.folder_id > -1]
      //   into  [ message.folder_id > -1]
      messageIdInSQL = " WHERE " + aQueryString.substr(aQueryString.indexOf(" message.folder_id"));
    }
    else if (aQueryString.indexOf("SELECT message.rowid FROM message JOIN folder ON message.folder_id = folder.rowid WHERE") == 0)
    {
      // convert [SELECT message.rowid FROM message JOIN folder ON message.folder_id = folder.rowid WHERE (folder.uri = 'imap://bugzillaw54%40gmail.com@imap.gmail.com/INBOX')" ]
      //   into  [ (folder.uri = 'imap://bugzillaw54%40gmail.com@imap.gmail.com/INBOX') ]
      messageIdInSQL = " WHERE " + aQueryString.substr(aQueryString.indexOf(" (folder.uri"));
    }
    else if (aQueryString.indexOf("SELECT message.rowid FROM message LEFT OUTER JOIN tags ON message.rowid = tags.fk") == 0)
    {
      // convert [SELECT message.rowid FROM message LEFT OUTER JOIN tags ON message.rowid = tags.fk WHERE message.folder_id > -1 AND tag_id is NULL]
      //  into   [ LEFT OUTER JOIN tags ON message.rowid = tags.fk WHERE message.folder_id > -1 AND tag_id is NULL]
      messageIdInSQL = aQueryString.substr(aQueryString.indexOf(" LEFT OUTER JOIN tags"));
      // remove this JOIN for folder and tag searches
      messageIdInSQL = messageIdInSQL.replace(/JOIN folder ON message.folder_id = folder.rowid/, "");
    }
    else if (aQueryString.indexOf("SELECT message.rowid FROM message JOIN tags ON message.rowid = tags.fk") == 0)
    {
      // convert [SELECT message.rowid FROM message JOIN tags ON message.rowid = tags.fk WHERE message.folder_id > -1 AND tags.tag_id IN (4)]
      //  into   [ JOIN tags ON message.rowid = tags.fk WHERE message.folder_id > -1 AND tags.tag_id IN (4)]
      messageIdInSQL = aQueryString.substr(aQueryString.indexOf(" JOIN tags"));
      // remove this JOIN for folder and tag searches
      messageIdInSQL = messageIdInSQL.replace(/JOIN folder ON message.folder_id = folder.rowid/, "");
    }
    
    if (aSQLSafeSearchString == "" && !hasMatchCondition)
    {
      // see RunFocusSearch()
      boundQueryString = "SELECT message_id, folder.uri, \"-1 -1\", message_text.body, " + (limitToContacts ? "COUNT(*) contact_count" : aSortColumn) + ", ";

      if (limitToEvents) 
        boundQueryString += "event.date, event.text ";
      else
        boundQueryString += "\"0.0\", NULL ";
 
      boundQueryString += "FROM message ";

      if (limitToEvents) 
        boundQueryString += "JOIN event ON message.rowid = event.fk ";

      boundQueryString += "JOIN folder ON message.folder_id = folder.rowid JOIN message_text ON message.rowid = message_text.docid " + messageIdInSQL;
    }
    else
    {
      // see RunFullTextSearch()
      // "?2" will cause us to look for matches in any column of the message_text table
      var matchCondition = aPostProcessor ? aPostProcessor.matchCondition : "?2";
      
      // XXX todo, fix me, remove that NULL and fix all the other places we do searchStmt.getString(x)
      var sql = "SELECT message_id, folder.uri, NULL, message_text.body, " + (limitToContacts ? "COUNT(*) contact_count" : aSortColumn) + ", ";
      if (limitToEvents) 
        sql += "event.date, event.text ";
      else
        sql += "\"0.0\", NULL ";

      sql += "FROM message ";

      if (limitToEvents)
        sql += "JOIN event ON message.rowid = event.fk ";

      sql += "JOIN folder ON message.folder_id = folder.rowid JOIN message_text ON message.rowid = message_text.docid " + messageIdInSQL + " AND message_text match '" + matchCondition + "'";

      // bind is not working with fts
      // so we have to manually escape single quotes
      // (and we've done that for aSQLSafeSearchString)
      // note, we have to worry about other stuff, like unsafe binary values?
      // http://www.phptutorial.info/?sqlite-escape-string
      // see bug #1178 for details
      boundQueryString = sql.replace(/\?2/g, aSQLSafeSearchString);
    }

    if (aPostProcessor && aPostProcessor.limitToUrls)
    {
      boundQueryString += " AND urls NOT NULL";
    }

    if (limitToEvents)
      boundQueryString += " ORDER BY " + aSortColumn + " DESC";  // one day:  "ORDER BY event.date ASC, ..." for agenda view?
    else if (limitToContacts)
      boundQueryString += " GROUP BY message_text.sender, message_text.recipients ORDER BY contact_count DESC";
    else
      boundQueryString += " ORDER BY " + aSortColumn + " DESC";
 
    PigLog("boundQueryString = " + boundQueryString);
    var searchStmt = getDBConn().createStatement(boundQueryString);
     
    var msgDB = null;
    var lastFolderURI = null;

    var imageHitsAllowed = !aPostProcessor || this.hitTypeAllowed(aPostProcessor, Ci.nsIMsgGlobalSearchNotify.imageHit);
    var attachmentHitsAllowed = !aPostProcessor || this.hitTypeAllowed(aPostProcessor, Ci.nsIMsgGlobalSearchNotify.attachmentHit);
    var contactHitsAllowed = !aPostProcessor || this.hitTypeAllowed(aPostProcessor, Ci.nsIMsgGlobalSearchNotify.contactHit);
    var urlHitsAllowed = !aPostProcessor || this.hitTypeAllowed(aPostProcessor, Ci.nsIMsgGlobalSearchNotify.urlHit);
    var messageHitsAllowed = !aPostProcessor || this.hitTypeAllowed(aPostProcessor, Ci.nsIMsgGlobalSearchNotify.messageHit);
    var eventHitsAllowed = !aPostProcessor || this.hitTypeAllowed(aPostProcessor, Ci.nsIMsgGlobalSearchNotify.eventHit);
    
    // call this to make sure ShouldIgnoreFile() is in scope
    EnsureSubScriptsLoaded();
    
    while (searchStmt.executeStep())
    {
      try 
      {
        var currentFolderURI = searchStmt.getString(1);

        var start = Date.now();
        if (currentFolderURI != lastFolderURI)
        {
          msgDB = GetMsgFolderFromUri(currentFolderURI, false).getMsgDatabase(null);
          lastFolderURI = currentFolderURI;
        }

        var messageId = decodeURIComponent(searchStmt.getString(0));
        var msgHdr = msgDB.getMsgHdrForMessageID(messageId);
        PigLog("3 getMsgHdrForMessageID took = " + (Date.now() - start) + "ms for " + messageId);

        var sortValue = searchStmt.getString(4);
        
        var eventDate = 0;
        var eventText = null;
                
        if (eventHitsAllowed)
        {
          if (!this.postProcessAndNotifyListener(aListener, aPostProcessor, msgHdr, Ci.nsIMsgGlobalSearchNotify.eventHit, {date: searchStmt.getString(5), text: searchStmt.getString(6)}, sortValue, searchStmt.getUTF8String(3)))
          {
            searchStmt.reset();
            return false;
          }
        }

        if (aListener && msgHdr)
        {
          // now that tags and mimetypes are in SQL, postProcess() will just return true
          // unless we've got a search session (which only happens if our folder is a virtual folder)
          // or we've got matchText.
          var bodyText = searchStmt.getUTF8String(3);
          
          if (!aPostProcessor || aPostProcessor.postProcess(msgHdr))
          {
            if (contactHitsAllowed)
            {
              // all messages have contacts, so we always notify and let searchresults.xml
              if (!this.postProcessAndNotifyListener(aListener, aPostProcessor, msgHdr, Ci.nsIMsgGlobalSearchNotify.contactHit, "", sortValue, bodyText))
              {
                searchStmt.reset();
                return false;
              }
            }

            if (messageHitsAllowed)
            {
              if (!this.postProcessAndNotifyListener(aListener, aPostProcessor, msgHdr, Ci.nsIMsgGlobalSearchNotify.messageHit, "", sortValue, bodyText))
              {
                searchStmt.reset();
                return false;
              }
            }

            if (urlHitsAllowed)
            {
              if (!this.postProcessAndNotifyListener(aListener, aPostProcessor, msgHdr, Ci.nsIMsgGlobalSearchNotify.urlHit, "", sortValue, bodyText))
              {
                searchStmt.reset();
                return false;
              }
            }

            if (imageHitsAllowed || attachmentHitsAllowed)
            {
              // for image, events, and attachments hits, don't stop reporting if we are in the middle of a message
              // wait until we are done with the current message.  
              // see bug #952 for details
              var returnValue = true;
              if (msgHdr.flags & MSG_FLAG_ATTACHMENT)
              {
                var attachmentIndex = 0;
                do
                {             
                  attachmentIndex++;
                  var url = msgHdr.getAttachmentProperty(attachmentIndex, "url");
                  if (url == "")
                    break;
                 
                  if (url.toLowerCase().indexOf("file:") == 0)
                  {
                    var fileExists = false;
                    try 
                    {
                      fileExists = gFileHandler.getFileFromURLSpec(url).exists();
                    }
                    catch (ex) { dump("XXX getFileFromURLSpec failed, ex = " + ex + "\n");}
                    
                    if (!fileExists)
                    {
                      FTIDump("skipping detached image or attachment because it doesn't exist on disk\n");
                      continue;
                    } 
                  }
                  
                  var contentType = msgHdr.getAttachmentProperty(attachmentIndex, "contentType").toLowerCase();
                  if (contentType == "text/x-moz-deleted")
                  {
                    FTIDump("skipping deleted image or attachment\n");
                    continue;
                  }

                  var displayName = msgHdr.getAttachmentProperty(attachmentIndex, "displayName");
                  try 
                  {                                     
                    displayName = gUnicodeConverter.ConvertToUnicode(displayName);
                  }
                  catch (ex) 
                  {
                    // if we fail to convert it, use it as is.  see bug #1506
                    //dump("XXX failed to convert displayName to unicode: " + displayName + "\n");
                  }
                  var filesuffix = displayName.substr(displayName.lastIndexOf(".")).toLowerCase();
                
                  // XXX todo stop duplicating this image/* logic
                  if (ShouldIgnoreFile(displayName))
                  { 
                    PigLog("ignoring " + displayName);
                  }
                  else if (contentType == 'image/jpeg' || contentType == 'image/jpg' || contentType == 'image/png' || contentType == 'image/pjpeg' || contentType == 'image/gif' || filesuffix == '.jpeg' || filesuffix == '.jpg' || filesuffix == '.png' || filesuffix == '.pjpeg' || filesuffix == '.gif')
                  {
                    if (imageHitsAllowed && !this.postProcessAndNotifyListener(aListener, aPostProcessor, msgHdr, Ci.nsIMsgGlobalSearchNotify.imageHit, attachmentIndex, sortValue, bodyText))
                      returnValue = false;
                  }
                  else if (contentType == "text/calendar" || contentType == "application/ics") 
                  {
                    // XXX todo stop indexing text/calendar and application/ics, like we do for text/x-vcard?
                    // for now, we ignore it.  
                    PigLog("ignoring " + contentType);
                  }
                  else
                  {
                    // handle the case where we don't have the right mime type
                    // XXX todo stop duplicating this text/calendar, application/ics, *.ics logic
                    if (filesuffix == ".ics")
                    {
                      // XXX todo stop .ics, like we do for .vcf
                      // for now, we ignore it.  
                      PigLog("ignoring .ics");
                    }
                    else if (attachmentHitsAllowed)
                    {
                      if (!this.postProcessAndNotifyListener(aListener, aPostProcessor, msgHdr, Ci.nsIMsgGlobalSearchNotify.attachmentHit, attachmentIndex, sortValue, bodyText))
                        returnValue = false;
                    }
                  }
                } while (true);
              }
              
              // if we hit our max results while processing the images, events, or attachments in that message we can stop
              if (!returnValue)
              {
                searchStmt.reset();
                return returnValue;
              } 
            }
          }
        }
      }
      catch (ex) 
      {
        FTIDump("executeFullTextQuery failed: (" + currentFolderURI + "): " + ex + "\n");
      }
    }

    searchStmt.reset();
    return true;
  },

  addUrlInfo : function(aMsgHdr, aURLs, aURLsCount, aTitles, aTitlesCount)
  {
    var folderID = GetFolderIdForURI(aMsgHdr.folder.URI);
    var urlInfo = convertUrlsAndTitlesToIndexableString(aURLs, aTitles);
    PigLog("addUrlInfo = [" + urlInfo + "]");
    LazyUpdate(aMsgHdr, folderID, urlInfo, gLazyUrlInfoUpdateTable);
  },

  addEventInfo : function(aMsgHdr, aEventTimes, aEventsTimesCount, aEventTexts, aEventTextsCount)
  {
    var folderID = GetFolderIdForURI(aMsgHdr.folder.URI);
    var events = [];
    for (var i = 0; i < aEventsTimesCount; i++)
    {
      events.push({date: aEventTimes[i], text: aEventTexts[i]});
    }
    FTIDump("addEventInfo = [" + events + "]\n");
    LazyUpdate(aMsgHdr, folderID, events, gLazyEventUpdateTable);
  },

  addBodyInfo : function (aMsgHdr, aBodyText)
  {
    DoubleCheckBody(aMsgHdr, aBodyText);

    var folderID = GetFolderIdForURI(aMsgHdr.folder.URI);
    FTIDump("addBodyInfo = [" + aBodyText + "]\n");
    LazyUpdate(aMsgHdr, folderID, aBodyText, gLazyBodyInfoUpdateTable);
  },
  
  addAttachmentInfo : function (aMsgHdr, aAttachmentText)
  {
    var folderID = GetFolderIdForURI(aMsgHdr.folder.URI);
    PigLog("addAttachmentInfo = [" + aAttachmentText + "]\n");
    LazyUpdate(aMsgHdr, folderID, aAttachmentText, gLazyAttachmentInfoUpdateTable);
  },
  
  debugInfoForFolder : function(aMsgFolder)
  {
    if (aMsgFolder)
    {
      var indexedCountForFolder = GetCountForFolder(aMsgFolder);
      return indexedCountForFolder + " messages in this folder (sqlite)";
    }
    else
    {
      var totalCount = 0;
      if (gTotalCount.executeStep())
        totalCount = gTotalCount.getInt64(0);
      gTotalCount.reset();
      
      var totalPendingDeleteCount;
      if (gTotalPendingDeleteCount.executeStep())
        totalPendingDeleteCount = gTotalPendingDeleteCount.getInt64(0);
      gTotalPendingDeleteCount.reset();
      return totalCount + " messages indexed, " + gMsgHdrsToIndex.Count() + " pending to be indexed, " + totalPendingDeleteCount  + " pending to be deleted";
    }
  },

  debugInfoForHeader : function(aMsgHdr)
  { 
    if (MessageHeaderIsInvalid(aMsgHdr))
      return "this message is not valid, zoinks!";

    var folderID = GetFolderIdForURI(aMsgHdr.folder.URI);
    var frecency = "0.0";
    var oid = -1;

    gFrecencyLookupStmt.bindUTF8StringParameter(0, encodeURIComponent(aMsgHdr.messageId));
    gFrecencyLookupStmt.bindInt64Parameter(1, folderID);
    if (gFrecencyLookupStmt.executeStep())
    {
      frecency = gFrecencyLookupStmt.getString(0);
      oid = gFrecencyLookupStmt.getInt64(1);
    }
    gFrecencyLookupStmt.reset();

    if (oid == -1)
      return "this message is not indexed";
   
    var count = 0;
    gTotalMessageCountForFolder.bindUTF8StringParameter(0, encodeURIComponent(aMsgHdr.messageId));
    gTotalMessageCountForFolder.bindInt64Parameter(1, folderID);
    if (gTotalMessageCountForFolder.executeStep())
      count = gTotalMessageCountForFolder.getInt64(0);
    gTotalMessageCountForFolder.reset();

    var distinctHdrs = Cc["@mozilla.org/supports-array;1"].createInstance(Components.interfaces.nsISupportsArray);
    this.getMsgHdrsForConversation(aMsgHdr, true /* aDistinctMessageIds */, false /* aIgnoreDeleted */, false /* aIncludeThread */, distinctHdrs);
    var allHdrs = Cc["@mozilla.org/supports-array;1"].createInstance(Components.interfaces.nsISupportsArray);
    this.getMsgHdrsForConversation(aMsgHdr, false /* aDistinctMessageIds */, false /* aIgnoreDeleted */, false /* aIncludeThread */, allHdrs);
    
    var messageInfo = "messageId = " + aMsgHdr.messageId + " threadId = " + aMsgHdr.threadId + " messageKey = " + aMsgHdr.messageKey + " ";
    for (var i = 0; i < aMsgHdr.numReferences; i++)
      messageInfo += aMsgHdr.getStringReference(i) + " ";
    return messageInfo + "count=" + count + " frecency=" + frecency + " [" + DecodeFrecency(aMsgHdr, aMsgHdr.folder.flags, frecency) + "] distinct conversation count=" + distinctHdrs.Count() + " all conversation count=" + allHdrs.Count();
  },

  setFeedback : function (aFeedback)
  {
    gIndexingFeedback = aFeedback;
  },
  
  startIndexing : function ()
  {
    gIndexing = true;
        
    if (gIndexingFeedback)
    {
      gIndexingFeedback.onStart();
    }
      
    onTimer();
  },
  
  stopIndexing : function()
  {
    gIndexing = false;
        
    // on shutdown, we call stopIndexing().  If we weren't indexing, gIndexingFeedback would be null
    if (gIndexingFeedback)
    {
      gIndexingFeedback.onStop();
      gIndexingFeedback = null;
    }
    
    gIgnoreIdle = false;

    if (gAlarm)
      gAlarm.cancel();

    // during idle, we may have stopped indexing (because we cycled through all the folders)
    // set gCyclingThroughFoldersState to stopped.
    // see bug #1164 amd #2456 for details
    gCyclingThroughFoldersState = "stopped";
    
    gIndexingOneFolder = false;

    // on stopIndexing, process all batched headers and lazy tables.
    // note, we call stopIndexing() if the user stops a manual index (see indexerProgress.js)
    // or if the user comes back from idle,
    // or if on idle indexing there are no more message to inde, 
    // or on shutdown.
    ProcessBatchedHeadersAndLazyTables();
  },
  
  resumeIndexing : function (aFolderURI)
  {
    gIgnoreIdle = true;

    if (aFolderURI == "")
    {
      SetCurrentFolderToIndex(null);
      gIndexingOneFolder = false
    }
    else
    {
      // remove any messages pending to be indexed, as we want indexing to go in folder order
      // or to be just for the specific folder we are indexing (for "Index Message In Folder")
      while (gMsgHdrsToIndex.Count())
        gMsgHdrsToIndex.DeleteElementAt(0);

      SetCurrentFolderToIndex(GetMsgFolderFromUri(aFolderURI, true));
      gIndexingOneFolder = true;
    }

    gCyclingThroughFoldersState = "start";
    
    gLastFolderIndexedUri = "";
    gSortedHeadersToIndex = null;
    this.startIndexing();
  },

  resetThenIndex : function (aFolderURI)
  {
    gIgnoreIdle = true;
    var dbConn = getDBConn();
    ResetGlobalIndex(dbConn, aFolderURI);
  },
 
  resetForFolder : function (aMsgFolder)
  {  
    try {
      FTIRemoveFolder(aMsgFolder.URI);
    }
    catch (ex)
    {
      FTIAlert("resetForFolder failed: " + ex + "\n");
    }
  },

  messageIdNotHash : function (aMessageId, aMsgHdrHash)
  {
    if (aMsgHdrHash[aMessageId])
      return false;
    else
      return true;
  },

  messageIdAndFolderURINotInArray : function (aMessageId, aFolderURI, aMsgHdrArray)
  {
    for (var i = 0; i < aMsgHdrArray.Count(); i++)
    {
      var msgHdr = aMsgHdrArray.GetElementAt(i).QueryInterface(Components.interfaces.nsIMsgDBHdr);
      if (msgHdr.messageId == aMessageId && (!aFolderURI || msgHdr.folder.URI == aFolderURI))
        return false;
    }
    return true;
  },

  hasFirstReference : function (aMsgHdr)
  {
    var hasMessage = false;
      
    if (aMsgHdr.numReferences)
    {
      gHasMessage.bindUTF8StringParameter(0, encodeURIComponent(aMsgHdr.getStringReference(0)));
      if (gHasMessage.executeStep())
        hasMessage = (gHasMessage.getInt64(0) == 1);
      gHasMessage.reset();
    }
    
    return hasMessage;
  },
  
  getKeywordsForConversation : function (aMsgHdr)
  {
    var start = Date.now();
    // when getting the getKeywordsForConversation, we don't need all the headers, we just need one
    var result = this.getKeywordsForConversationInternal(aMsgHdr, false /* aAllHeaders */);
    PigLog("BEACHBALL: getKeywordsForConversationInternal took = " + (Date.now() - start) + "ms for " + aMsgHdr.messageId + " " + aMsgHdr.folder.URI + " result = [" + result + "]");
    var total = Date.now() - start;
    if (total > 1000)
      PigLog("BEACHBALL: getKeywordsForConversation TOOK TOO LONG");
    return result;
  },

  getKeywordsForConversationInternal : function (aMsgHdr)
  {
    var hdrs = Cc["@mozilla.org/supports-array;1"].createInstance(Components.interfaces.nsISupportsArray);

    // when determining all the tags for a conversation, we can just use the first hdr we find (in the conversation)
    // see bug #2235 for details
    this.getFirstMsgHdrForConversation(aMsgHdr, hdrs);
    
    // tagKeysForMsgHdr() comes from chrome://messenger/content/mailCore.js
    // EnsureSubScriptsLoaded() ensures that mailCore.js is loaded into scope.
    EnsureSubScriptsLoaded();
    var allKeywords = "";
    for (var i = 0; i < hdrs.Count(); i++)
    {
      var hdr = hdrs.GetElementAt(i).QueryInterface(Components.interfaces.nsIMsgDBHdr);
      PigLog("getKeywordsForConversationInternal, header #" + i);
      PigLog("\tthreadId = " + hdr.threadId);
      for (var j = 0; j < hdr.numReferences; j++)
      {
        PigLog("\tref[" + j + "] = " + hdr.getStringReference(j));
      }

      var keywords = tagKeysForMsgHdr(hdr);
      PigLog("\tkeywords = [" + keywords + "] for [" + hdr.subject + "]");
      if (keywords)
      {
        if (allKeywords)
          allKeywords += " ";
        allKeywords += keywords;
      }
    }

    if (!allKeywords)
      return allKeywords;
    
    var keywordsArray = allKeywords.split(" ");
    var uniqueKeywordsArray = [];
    for (var i = 0; i < keywordsArray.length; i++)
    {
      if (uniqueKeywordsArray.indexOf(keywordsArray[i]) == -1)
        uniqueKeywordsArray.push(keywordsArray[i]);
    }
    return uniqueKeywordsArray.join(" ");
  },

  getFirstMsgHdrForConversation : function (aMsgHdr, aMsgHdrArray)
  {
    var start = Date.now();

    if (aMsgHdrArray.Count())
      FTIAlert("we expect aMsgHdrArray to be empty!");

    var refs = [];
    var headerHash = new Object();

    this.getMsgHdrsForConversationInternal(aMsgHdr, headerHash, refs, false /* aAllHeaders */);
    PigLog("getMsgHdrsForConversationInternal took = " + (Date.now() - start) + "ms for " + aMsgHdr.messageId + " " + aMsgHdr.folder.URI);

    for each (header in headerHash)
    {
      aMsgHdrArray.AppendElement(header);
    }

    PigLog("getFirstMsgHdrForConversation (distinct) took = " + (Date.now() - start) + "ms for " + aMsgHdr.messageId + " " + aMsgHdr.folder.URI);
    return;
  },

  getMsgHdrsForConversation : function (aMsgHdr, aDistinctMessageIds, aIgnoreDeleted, aIncludeThread, aMsgHdrArray)
  {
    var start = Date.now();

    if (aMsgHdrArray.Count())
      FTIAlert("we expect aMsgHdrArray to be empty!");

    var refs = [];
    var headerHash = new Object();

    // include msg hdrs from the thread that contains aMsgHdr
    // they might not be in the conversation, due to subject threading
    // see bug #3328 for details
    if (aIncludeThread)
    {
      var database = aMsgHdr.folder.getMsgDatabase(null);
      if (database)
      {
        var thread = database.GetThreadContainingMsgHdr(aMsgHdr);
        for (var index = 0; index < thread.numChildren; index++)
        {
          var header = thread.getChildAt(index);
          headerHash[header.messageId] = header.QueryInterface(Ci.nsISupports);
        }
      }
    }
    
    this.getMsgHdrsForConversationInternal(aMsgHdr, headerHash, refs, true /* aAllHeaders */);
    PigLog("getMsgHdrsForConversationInternal took = " + (Date.now() - start) + "ms for " + aMsgHdr.messageId + " " + aMsgHdr.folder.URI);

    for each (header in headerHash)
    {
      aMsgHdrArray.AppendElement(header);
    }

    if (aDistinctMessageIds)
    {      
      // if we want distinct message ids, for each message id in our conversation
      // ...
      for (var i = 0; i < aMsgHdrArray.Count(); i++)
      {
        var msgHdr = aMsgHdrArray.GetElementAt(i).QueryInterface(Components.interfaces.nsIMsgDBHdr);
        var bestFolder = msgHdr.folder;
        gFolderLookupStmt.bindUTF8StringParameter(0, encodeURIComponent(msgHdr.messageId));
        while (gFolderLookupStmt.executeStep())
        {
          try 
          {
            var currentFolderURI = gFolderLookupStmt.getUTF8String(0);
            if (currentFolderURI == bestFolder.URI)
              continue;
              
            var currentFolder = GetMsgFolderFromUri(currentFolderURI, false);
            if (isFolderBetter(bestFolder, currentFolder, aMsgHdr.folder))
            {
              // currentFolder is better than the current best folder, use that one instead
              bestFolder = currentFolder;
            }
          }
          catch (ex)
          {
            FTIAlert("getMsgHdrsForConversation() failed, ex " + ex);
          }
        }
        gFolderLookupStmt.reset();
        
        try
        {
          // if bestFolder == the original msg hdr, just use the original msg hdr
          if (bestFolder != msgHdr.folder)
          {
            // get the msg hdr from the best folder and original message id.
            var bestHeader = bestFolder.getMsgDatabase(null).getMsgHdrForMessageID(msgHdr.messageId);
            if (bestHeader)
            {
              // replace the current msg hdr with a better one
              aMsgHdrArray.SetElementAt(i, bestHeader.QueryInterface(Ci.nsISupports));
            }
          }
        }
        catch (ex)
        {
          // this should not happen
          FTIAlert("getMsgHdrsForConversation() failed, ex " + ex);
        }
      }
    }
    else
    {
      var nonDistinctHeaders = Cc["@mozilla.org/supports-array;1"].createInstance(Components.interfaces.nsISupportsArray);

      // if we don't want distinct message ids, for each message id in our conversation
      // go get all the duplicates.  note, this will not get duplicates within the same folder.
      for (var i = 0; i < aMsgHdrArray.Count(); i++)
      {
        var msgHdr = aMsgHdrArray.GetElementAt(i).QueryInterface(Components.interfaces.nsIMsgDBHdr);

        gFolderLookupStmt.bindUTF8StringParameter(0, encodeURIComponent(msgHdr.messageId));
        while (gFolderLookupStmt.executeStep())
        {
          try 
          {
            var folderURI = gFolderLookupStmt.getUTF8String(0);

            // get msg hdr from folderURI and message id.
            var start2 = Date.now();
            var header = GetMsgFolderFromUri(folderURI, false).getMsgDatabase(null).getMsgHdrForMessageID(msgHdr.messageId);
            PigLog("2 getMsgHdrForMessageID took = " + (Date.now() - start2) + "ms for " + msgHdr.messageId + " " + msgHdr.folder.URI);

            if (header)
            {
              if (this.messageIdAndFolderURINotInArray(header.messageId, header.folder.URI, aMsgHdrArray) &&
                  this.messageIdAndFolderURINotInArray(header.messageId, header.folder.URI, nonDistinctHeaders))
              {
                // add this header to the nonDistinctHeaders array
                // when we are done, we'll append these to aMsgHdrArray
                nonDistinctHeaders.AppendElement(header.QueryInterface(Ci.nsISupports));
              }
            }
            else
            {
              // todo:  figure out why this can happen
              FTIDump("unable to find header 3: " + msgHdr.messageId + " " + folderURI);
            }
          }
          catch (ex)
          {
            FTIAlert("getMsgHdrsForConversation() failed, ex " + ex);
          }
        }
        gFolderLookupStmt.reset();
      }

      for (var i = 0; i < nonDistinctHeaders.Count(); i++)
      {
        aMsgHdrArray.AppendElement(nonDistinctHeaders.GetElementAt(i));
      }
    }

    // if we are ignoring delete messages, walk the array and remove any that are in trash folders (or are marked as deleted)
    // unless aMsgHdr is in the trash folder (or is marked as deleted), in which case don't ignore them
    if (aIgnoreDeleted && !(aMsgHdr.folder.flags & MSG_FOLDER_FLAG_TRASH) && !(aMsgHdr.flags & MSG_FLAG_IMAP_DELETED))
    {
      for (var i = aMsgHdrArray.Count() - 1; i >= 0; i--)
      {
        var msgHdr = aMsgHdrArray.GetElementAt(i).QueryInterface(Components.interfaces.nsIMsgDBHdr);
        if ((msgHdr.folder.flags & MSG_FOLDER_FLAG_TRASH) || (msgHdr.flags & MSG_FLAG_IMAP_DELETED))
          aMsgHdrArray.DeleteElementAt(i);
      }
    }
    
    PigLog("getMsgHdrsForConversation (aDistinctMessageIds=" + aDistinctMessageIds + " aIgnoreDeleted=" + aIgnoreDeleted + " aIncludeThread=" + aIncludeThread + ") took = " + (Date.now() - start) + "ms for " + aMsgHdr.messageId + " " + aMsgHdr.folder.URI);  
  },

  countHeadersInHash : function (aMsgHdrHash)
  {
    var count = 0;
    for each (header in aMsgHdrHash)
    {
      count++;
    }
    return count;
  },

/*
  given a msg hdr, get all the messages in the conversation that contains that hdr, including itsself.

  1)  append the header
  
  2)  get the references.
    a)  for each reference, if not already in our array, find the folder uri and use that and the reference to get the msg hdr from the appropriate msg db.  
        then call getMsgHdrsForConversationInternal() recursively with that msg hdr.
        
    b)  for each reference, if not already in our array, find all the message that refer to those references (see bug #1514) 
   
  3)  get the message id.
        find all message ids (and the corresponding folder uri) that refer to it.  
        for each message id, if not already in our array, use the folder uri and the mesasge id to get the msg hdr from the appropriate msg db.  
        then call getMsgHdrsForConversationInternal() recursively with that msg hdr.
*/
  getMsgHdrsForConversationInternal : function (aMsgHdr, aMsgHdrHash, aRefs, aAllHeaders)
  {
    var start = Date.now();
    if (!aMsgHdr)
    {
      FTIAlert("unexpected:  calling getMsgHdrsForConversationInternal() with null");
      return;
    }
    
    // we want distinct by message id
    if (this.messageIdNotHash(aMsgHdr.messageId, aMsgHdrHash))
    {
      // if we don't want all the headers, we just want one additional header.
      // so we stop (and return) when our header hash (aMsgHdrHash) has two headers.
      if (!aAllHeaders && this.countHeadersInHash(aMsgHdrHash) == 2)
        return;

      // add this header to the conversation
      aMsgHdrHash[aMsgHdr.messageId] = aMsgHdr.QueryInterface(Ci.nsISupports);
    }

    // if we don't want all the headers, we just want one additional header.
    // so we stop (and return) when our header hash (aMsgHdrHash) has two headers.
    if (!aAllHeaders && this.countHeadersInHash(aMsgHdrHash) == 2)
      return;

    this.getMsgHdrsThatRefersToMessage(aMsgHdr.messageId, aMsgHdrHash, aRefs, aAllHeaders);
    PigLog("BEACHBALL: getMsgHdrsThatRefersToMessage took = " + (Date.now() - start) + "ms for " + aMsgHdr.messageId + " " + aMsgHdr.folder.URI);

    // if we don't want all the headers, we just want one additional header.
    // so we stop (and return) when our header hash (aMsgHdrHash) has two headers.
    if (!aAllHeaders && this.countHeadersInHash(aMsgHdrHash) == 2)
      return;

    this.getMsgHdrsThatMessageRefersTo(aMsgHdr, aMsgHdrHash, aRefs, aAllHeaders);
    PigLog("BEACHBALL: getMsgHdrsThatMessageRefersTo took = " + (Date.now() - start) + "ms for " + aMsgHdr.messageId + " " + aMsgHdr.folder.URI);
  },
  
  getMsgHdrsThatMessageRefersTo : function (aMsgHdr, aMsgHdrHash, aRefs, aAllHeaders)
  {
    // find all msg headers (recursively) that this message refers to
    for (var i = 0; i < aMsgHdr.numReferences; i++)
    {
      var ref = aMsgHdr.getStringReference(i);
      if (ref && this.messageIdNotHash(ref, aMsgHdrHash))
      {
        // use the first message we find
        gFolderLookupStmt.bindUTF8StringParameter(0, encodeURIComponent(ref));
        var msgHdr = null;
        while (!msgHdr && gFolderLookupStmt.executeStep())
        {
          try {
            var folderURI = gFolderLookupStmt.getUTF8String(0);
            
            // get msg hdr from folderURI and message id.
            var start = Date.now();
            msgHdr = GetMsgFolderFromUri(folderURI, false).getMsgDatabase(null).getMsgHdrForMessageID(ref);
            PigLog("1 getMsgHdrForMessageID took = " + (Date.now() - start) + "ms for " + ref + " " + folderURI);

            // todo:  figure out why this can happen
            if (!msgHdr)
              FTIDump("unable to find header 1: " + ref + " " + folderURI);
          }
          catch (ex)
          {
            FTIAlert("getMsgHdrsThatMessageRefersTo() failed 1, ex " + ex);
          }
        }
        gFolderLookupStmt.reset();
        
        if (msgHdr)
        {
          // we found a db header for this reference, so recurse
          // getMsgHdrsForConversationInternal() will append msgHdr to our array
          this.getMsgHdrsForConversationInternal(msgHdr, aMsgHdrHash, aRefs, aAllHeaders);
        }
        else
        {
          // we did not find a db header for this reference,
          // but we want to find other messages that may refer to this reference
          // so we call getMsgHdrsThatRefersToMessage() directly
          this.getMsgHdrsThatRefersToMessage(ref, aMsgHdrHash, aRefs, aAllHeaders);
        }
      }
    }
  },
  
  getMsgHdrsThatRefersToMessage : function (aMessageId, aMsgHdrHash, aRefs, aAllHeaders)
  {
    // we've already looked for headers that refer to this message id, so don't do it again.
    if (aRefs.indexOf(aMessageId) != -1)
      return;
    aRefs.push(aMessageId);

    // find all messages (recursively) that refer to this one
    gDistinctMessageIdAndFolderLookupStmt.bindUTF8StringParameter(0, encodeURIComponent(aMessageId));
    
    // we need to build up this array before we recurse, so that we can
    // reset the statement.
    var headersToCheck = [];
    while (gDistinctMessageIdAndFolderLookupStmt.executeStep())
    {
      try {
        var messageId = decodeURIComponent(gDistinctMessageIdAndFolderLookupStmt.getUTF8String(0));
        var folderURI = gDistinctMessageIdAndFolderLookupStmt.getUTF8String(1);

        if (this.messageIdNotHash(messageId, aMsgHdrHash))
        {
          headersToCheck.push({messageId: messageId, folderURI: folderURI});
        }
      }
      catch (ex)
      {
        FTIAlert("getMsgHdrsThatMessageRefersTo() failed 2, ex " + ex);
      }
    }
    gDistinctMessageIdAndFolderLookupStmt.reset();
    
    for (var i = 0; i < headersToCheck.length; i++)
    {
      try {
        var start = Date.now();
        var msgHdr = GetMsgFolderFromUri(headersToCheck[i].folderURI, false).getMsgDatabase(null).getMsgHdrForMessageID(headersToCheck[i].messageId);
        PigLog("4 getMsgHdrForMessageID took = " + (Date.now() - start) + "ms for " + headersToCheck[i].messageId + " " + headersToCheck[i].folderURI);

        if (msgHdr)
        {
	        // recurse, getMsgHdrsForConversationInternal() will append msgHdr to our array
          this.getMsgHdrsForConversationInternal(msgHdr, aMsgHdrHash, aRefs, aAllHeaders);
        }
        else
        {
          // todo:  figure out why this can happen
          FTIDump("unable to find header 2: " + headersToCheck[i].messageId + " " + headersToCheck[i].folderURI);
        }
      }
      catch (ex)
      {
        FTIAlert("getMsgHdrsThatMessageRefersTo() failed 3, ex " + ex);
      }
    }
  },
  
  _foldersToUpdate : [],

  processFoldersWaitingForUpdate : function ()
  {
    for each (var folderURI in this._foldersToUpdate) 
    {
      PigLog("processFoldersWaitingForUpdate with " + folderURI);
      try
      {
        var folder = GetMsgFolderFromUri(folderURI, true);
        UpdateFolderIfWontPrompt(folder);
      }
      catch (ex)
      {
        FTIAlert("processFoldersWaitingForUpdate() failed:  " + ex + "\n");
      }
    }
    this._foldersToUpdate = [];
  },

  updateFolderOnIdle : function (aFolderURI)
  {
    PigLog("updateFolderOnIdle with " + aFolderURI);
    if (this._foldersToUpdate.indexOf(aFolderURI) == -1) 
      this._foldersToUpdate.push(aFolderURI);
  },

  getTagIdForLabel : function (aLabel)
  {
    return GetTagIdForLabel(aLabel);
  },

  logHdr : function (aMsgHdr)
  {
    this.log("messageId = " + aMsgHdr.messageId);
    this.log("key = " + aMsgHdr.messageKey);
    this.log("subject = " + aMsgHdr.subject);
    this.log("folderURI = " + aMsgHdr.folder.URI);
    this.log("keywords as string property = " + aMsgHdr.getStringProperty("keywords"));
    // tagKeysForMsgHdr() comes from chrome://messenger/content/mailCore.js
    // EnsureSubScriptsLoaded() ensures that mailCore.js is loaded into scope.
    EnsureSubScriptsLoaded();
    this.log("tagKeysForMsgHdr = " + tagKeysForMsgHdr(aMsgHdr));
  },
  
  _logFile : null,
  _logStream : null,
  _lastLogTime : null,

  log : function (aText)
  {
    if (!gPigLogToDisk)
      return;

    if (!this._logFile)
    {
      // get profile directory
      this._logFile = Components.classes["@mozilla.org/file/directory_service;1"]
                             .getService(Components.interfaces.nsIProperties)
                             .get("ProfD", Components.interfaces.nsIFile);
      this._logFile.append("postbox.log");
    
      if (!this._logFile.exists())
        this._logFile.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0666);
    }

    if (!this._logStream)
    {
      this._logStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                                  .createInstance(Components.interfaces.nsIFileOutputStream);

      this._logStream.init(this._logFile, 0x02 | 0x10, 0666, 0);
    }

    var d = new Date();
    var dateString = d.toString().substr(0, d.toString().indexOf(" GMT"));
    var timeInMilliseconds = d.getTime();
    var text = timeInMilliseconds + " [" + dateString + "]:\t" + aText + "\r\n";
    this._logStream.write(text, text.length);

    // close (at most) once every three seconds, and make sure to close on quit
    if (((timeInMilliseconds - this._lastLogTime) > 3000) || aText == "quit-application")
    {
      this._logStream.close();
      this._logStream = null;
      this._lastLogTime = timeInMilliseconds;
    }
  },

  logTags : function ()
  {
    this.log("from prefs:");
    var prefB = Components.classes["@mozilla.org/preferences-service;1"].
                    getService(Components.interfaces.nsIPrefBranch);
    var tagPrefs = prefB.getChildList("mailnews.tags.", {});
    for each (prefName in tagPrefs) {
      if ((prefName.length - ".tag".length) == prefName.lastIndexOf(".tag"))
        this.log(prefName + " = " + prefB.getCharPref(prefName));
      else
        this.log(prefName);
    }

    this.log("from the tag service:");
    var tagService = Components.classes["@mozilla.org/messenger/tagservice;1"]
                               .getService(Components.interfaces.nsIMsgTagService);
    var tagArray = tagService.getAllTags({});
    for (var i = 0; i < tagArray.length; i++)
    {
      this.log("tag #" + i);
      this.log("key = " + tagArray[i].key);
      this.log("tag = " + tagArray[i].tag);
      this.log("color = " + tagArray[i].color);
      this.log("ordinal = " + tagArray[i].ordinal);
      this.log("active = " + tagArray[i].active);
    }
  },
  _runningUrls: [],
  /* nsIUrlListener */
  OnStartRunningUrl : function (aURL)
  {
    PigLog("OnStartRunningUrl = " + aURL.spec);
    var urlIndex = this._runningUrls.indexOf(aURL);
    if (urlIndex != -1)
      PigLog("XXX OnStartRunningUrl, unexpected url was found, was the url already started?");
    this._runningUrls.push(aURL);
  },
  OnStopRunningUrl : function (aURL, aExitCode)
  {
    PigLog("OnStopRunningUrl = " + aURL.spec);
    var urlIndex = this._runningUrls.indexOf(aURL);
    if (urlIndex != -1)
      this._runningUrls.splice(urlIndex, 1);
    else
      PigLog("XXX OnStopRunningUrl, unexpected url not found, was the url already stopped?");
      
    // if there are no urls running and we've been idle for more than our max, re-notify that we are idle
    // check to make sure we aren't stopped because we have gone through all our folders once already
    // on idle.  see bug #2456 for details
    PigLog("number urls running = " + this._runningUrls.length + " idle time = " + Cc["@mozilla.org/widget/idleservice;1"].getService(Ci.nsIIdleService).idleTime + " gCyclingThroughFoldersState = " + gCyclingThroughFoldersState);
    if (!this._runningUrls.length && 
       (Cc["@mozilla.org/widget/idleservice;1"].getService(Ci.nsIIdleService).idleTime >= (gIdleTime * 1000)) &&
       gCyclingThroughFoldersState == "start")
    {
      var isSmtpUrl = false;
      try {
        aURL.QueryInterface(Components.interfaces.nsISmtpUrl);
        isSmtpUrl = true;
      }
      catch (ex)
      { 
      }

      // if we hit zero running urls and our last url was an SMTP url and we've got folders to update,
      // don't manually notify that we are idle.
      // nsMsgComposeAndSend::DoFcc() tells the indexer that the fcc folder needs update on the next idle
      // but we don't want to do that because we're about to copy a message to the fcc folder
      // see bug #2499 for details
      if (isSmtpUrl && this._foldersToUpdate.length)
      {
        PigLog("don't manually notify idle because we hit zero with an SMTP url and we've got folders to update");
      }
      else 
      {
        PigLog("manually notify idle");
        this.observe("", "idle", "");
      }
    }
    
    // if we've stopped cycling through folders, and there are no more urls running, set the state to "start"
    // so that on if we come out of idle, but go back into it, we'll start cycling again.
    if (gCyclingThroughFoldersState == "stopped" && !this._runningUrls.length)
    {
      gCyclingThroughFoldersState = "start";
    }
  }
};

function InitGlobalIndexingIntegration()
{
  FTIDump("initializing global indexing\n");
  gMessenger = Cc["@mozilla.org/messenger;1"].createInstance().QueryInterface(Ci.nsIMessenger);
  gMozStorage = Cc["@mozilla.org/storage/service;1"].getService(Ci.mozIStorageService);
  gPrefBranch = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(null);

  try 
  {
    gIndexMsgs = gPrefBranch.getBoolPref("mail.fulltextsearch.enable");
    gLastFolderIndexedUri = gPrefBranch.getCharPref(kLastFolderIndexPrefName);
  } catch (ex) {}
  
  if (!gIndexMsgs)
    return;
    
  var nsIMsgFolderListener = Ci.nsIMsgFolderListener;
  var notificationService = Cc["@mozilla.org/messenger/msgnotificationservice;1"].getService(Ci.nsIMsgFolderNotificationService);
  notificationService.addListener(gMsgFolderListener);
  
  var mailSession = Cc["@mozilla.org/messenger/services/session;1"].getService(Ci.nsIMsgMailSession);
  mailSession.AddFolderListener(gFolderListener, Ci.nsIFolderListener.event);
    
  // XXX todo, figure out where to call notificationService.removeListener(gMsgFolderListener);
  var ObserverService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  ObserverService.addObserver(CreateMsgDisplayedObserver, "MsgMsgDisplayed", false);
  gMsgTextIndexer = new nsMsgFullTextIndexer;
  InitSQLStatements();
}

// The following set of functions handle background indexing of messages.
// They iterate over accounts, folders, and messages to find un-indexed
// messages, and indexes those messages.

function compareServers(a, b)
{
  return a.sortKey - b.sortKey;
}

function compareFolders(a, b)
{
  return a.compareSortKeys(b);
}

function compareMessages(a, b)
{
  // newest first
  return a.dateInSeconds -  b.dateInSeconds;
}

function getFolderRank(aFlags)
{
  if (aFlags & MSG_FOLDER_FLAG_INBOX)
    return 5;
  else if (aFlags & MSG_FOLDER_FLAG_ARCHIVE)
    return 4;
  else if (aFlags & MSG_FOLDER_FLAG_SENTMAIL)
    return 3;
  else if (aFlags & MSG_FOLDER_FLAG_TRASH)
    return 1;
  else if (aFlags & MSG_FOLDER_FLAG_JUNK)
    return 0;
  else
    return 2; // a regular folder
}

// is b better than a, using c as the best possible folder?
function isFolderBetter(a, b, c)
{
  // if a is already the best, or if they are the same, b is not better
  if (a == c || a == b)
  {
    return false;
  }
  
  // if a and b are on two different servers, and one is on the best server, it is better
  if (!a.server.equals(b.server))
  {
    if (a.server.equals(c.server))
      return false;
    else if (b.server.equals(c.server))
      return true;
  }
  
  // a and b are on the same server, compare folder flags.
  // if flags are the same, use the sort order
  if (getFolderRank(a.flags) > getFolderRank(b.flags))
    return false;
  else if (getFolderRank(a.flags) < getFolderRank(b.flags))
    return true;
  else
    return (a.compareSortKeys(b) < 0);
}

function UpdateFolderIfWontPrompt(aFolder)
{
  // before we check to see if there are an headers in the folder that aren't index, we should update the folder
  // note, this is async, so it doesn't completely do what we want.  see bug #1176 for details.
  //
  // so if we need to be authenticated to biff, check that we are
  // (since we don't want to prompt the user for password UI)
  // and make sure the server isn't already in the middle of downloading new messages
  // this code is ported from nsMsgBiffManager::PerformBiff()
  var server = aFolder.server;
  if (WontPromptForPassword(server))
  {
    PigLog("calling updateFolder() for " + aFolder.URI);
    aFolder.updateFolder(null);
  }
}

function FindNextFolderToIndex()
{
  var servers = gAccountManager.allServers;
  var foundFolder = false;
  var useNextFolder = false;

  var sortedServers = [];
  for (var i = 0; i < servers.Count(); i++)
  {
    var server = servers.QueryElementAt(i, Components.interfaces.nsIMsgIncomingServer);
    if (server.indexMsgs) // skip servers we're not indexing
      sortedServers.push(server);
  }
  sortedServers.sort(compareServers);
    
  for (var i = 0; i < sortedServers.length && !foundFolder; i++)
  {
    var rootFolder = sortedServers[i].rootFolder;
    var allFolders = Cc["@mozilla.org/supports-array;1"].createInstance(Components.interfaces.nsISupportsArray);
    rootFolder.ListDescendents(allFolders);
    var numFolders = allFolders.Count();
    
    var sortedFolders = [];
    for (var folderIndex = 0; folderIndex < numFolders; folderIndex++)
    {
      var folder = allFolders.GetElementAt(folderIndex).QueryInterface(Components.interfaces.nsIMsgFolder);
      sortedFolders.push(folder);
    }
    sortedFolders.sort(compareFolders);
    
    for (var folderIndex = 0; folderIndex < numFolders && !foundFolder; folderIndex++)
    {
      var folder = sortedFolders[folderIndex];
      
      // if no folder was indexed (or the pref's not set), just use the first folder
      if (!gLastFolderIndexedUri.length || useNextFolder)
      {
        try
        {
          UpdateFolderIfWontPrompt(folder);
        }
        catch (ex)
        {
          FTIAlert("FindNextFolderToIndex, UpdateFolderIfWontPrompt() failed:  " + ex + "\n");
          folder = null;
        }
   
        if (!folder)
          continue;
        
        SetCurrentFolderToIndex(folder);
        gSortedHeadersToIndex = null;
        
        if (gIndexingFeedback)
          gIndexingFeedback.onSearchingFolder(gCurrentFolderToIndex);
        foundFolder = true;
      }
      else if (gLastFolderIndexedUri == folder.URI)
        useNextFolder = true;
    }
  }
  
  if (!foundFolder)
    gLastFolderIndexedUri = "";
}

function MessageHeaderIsInvalid(msgHdr)
{
  try {
    if (msgHdr && msgHdr.messageId && msgHdr.messageId.length)
    {
      // if we have a non-empty messageId, we are in good shape, so the message header is not invalid.
      return false;
    }
  }
  catch (ex)
  {
    // we might have a message header for a message in a db that has been closed or is invalid
    // if so, accessing messageId will throw NS_ERROR_INVALID_POINTER
    // see bug #1136 for details
    if (ex.result == Components.results.NS_ERROR_INVALID_POINTER)
    {
      FTIAlert("hitting bug #1136, fail gracefully.  the proper fix is bug #1135");
    }
    else
    {
      FTIAlert("MessageHeaderIsInvalid() unexpected exception = " + ex);
    }
  }
  return true;
}

function GetCountForFolder(aMsgFolder)
{
  try
  {
    var folderID = GetFolderIdForURI(aMsgFolder.URI);
    var totalCountForFolder = 0;
    gTotalCountForFolder.bindInt64Parameter(0, folderID);
    if (gTotalCountForFolder.executeStep())
      totalCountForFolder = gTotalCountForFolder.getInt64(0);
    gTotalCountForFolder.reset();
    return totalCountForFolder;
  }
  catch (ex)
  {
    FTIAlert("GetCountForFolder() for [" + (aMsgFolder ? aMsgFolder.URI : "null") +  "] unexpected exception = " + ex);
  }
  return -1;
}
  
function DoubleCheckHeaders(aFolder)
{
  var start = Date.now();
  
  var headerHash = new Object();
  var folderID = GetFolderIdForURI(aFolder.URI);
  gAllMessageIdsInFolder.bindInt64Parameter(0, folderID);
  while (gAllMessageIdsInFolder.executeStep())
  {
    headerHash[gAllMessageIdsInFolder.getUTF8String(0)] = true;
  }
  gAllMessageIdsInFolder.reset();

  PigLog("getting all sqlite message ids for " + aFolder.URI + " took " + (Date.now() - start) + "ms");
  
  try
  {
    // this can throw if you've got a local folder without an msf
    // see bug #3170 for details
    var msgDB = aFolder.getMsgDatabase(null);
    var headerEnumerator = msgDB.EnumerateIndexedMessages();
  } 
  catch (ex) 
  {
    FTIAlert("failed while getting indexed messages enumerator for " + aFolder.URI + ":" + ex);
    return;
  }
  
  var numMissing = 0;
  try 
  {
    // iterate over the folder finding the next message to index...
    while (headerEnumerator.hasMoreElements())
    {
      var msgHdr = headerEnumerator.getNext().QueryInterface(Components.interfaces.nsIMsgDBHdr);
      if (!headerHash[encodeURIComponent(msgHdr.messageId)])
      {
        PigLog("header missing (" + msgHdr.messageId + "), mark as unindexed");
        msgHdr.markIndexed(false);
        numMissing++;
      }
    }
  }
  catch (ex)
  {
    FTIAlert("failed while enumerating indexed messages for " + aFolder.URI + ":" + ex);
    return;
  }
  
  if (numMissing)
    msgDB.Commit(MSG_DB_LARGE_COMMIT);
    
  PigLog("confirming all mork message ids for took " + aFolder.URI + " took " + (Date.now() - start) + "ms.  numMissing = " + numMissing);
}

function FindNextHdrToIndex()
{
  if (!gSortedHeadersToIndex)
  {
    // see bug #2789 for details
    DoubleCheckHeaders(gCurrentFolderToIndex);
    
    var start = Date.now();
    gSortedHeadersToIndex = [];

    try {
      // XXX this is not cheap for a large folder with lots of un-indexed messages
      // what about a reverse enumerator, aren't messages in the db oldest-to-newest?
      // or instead of sorting, just reverse the list?
      //
      // this can throw if you've got a local folder without an msf
      // see bug #3170 for details
      var msgDB = gCurrentFolderToIndex.getMsgDatabase(null);
      var enumerator = msgDB.EnumerateUnindexedMessages();

      while (enumerator.hasMoreElements())
      {
        var msgHdr = enumerator.getNext().QueryInterface(Components.interfaces.nsIMsgDBHdr);
        gSortedHeadersToIndex.push(msgHdr);
      }
    }
    catch (ex)
    {
      FTIAlert("failed while enumerating unindexed messages for " + gCurrentFolderToIndex.URI + ":" + ex);
    }

    PigLog("enumeration of " + gSortedHeadersToIndex.length + " headers took = " + (Date.now() - start) + "ms");
    start = Date.now();
    gSortedHeadersToIndex.sort(compareMessages);
    PigLog("sort of " + gSortedHeadersToIndex.length + " headers took = " + (Date.now() - start) + "ms");
  }

  var headerToIndex = gSortedHeadersToIndex.pop();
  while (headerToIndex)
  {
    if (!MessageHeaderIsInvalid(headerToIndex))
      return headerToIndex;
    PigLog("how does this happen?  does the sorted array of hdrs (from the enumerator) out live the message?");
    headerToIndex = gSortedHeadersToIndex.pop();
  }

  CloseFolderDatabaseIfNotInUse(gCurrentFolderToIndex);
  gSortedHeadersToIndex = null;
  return null;  
}

function determineFoldersToClear()
{
  var folders = [];
  var servers = gAccountManager.allServers;
  
  for (var i = 0; i < servers.Count(); i++)
  {
    var server = servers.QueryElementAt(i, Components.interfaces.nsIMsgIncomingServer);
    if (server.indexMsgs)
    {
        var rootFolder = server.rootFolder;
        var allFolders = Cc["@mozilla.org/supports-array;1"].createInstance(Ci.nsISupportsArray);
        rootFolder.ListDescendents(allFolders);
        var numFolders = allFolders.Count();
        for (var folderIndex = 0; folderIndex < numFolders; folderIndex++)
        {
          folders.push(allFolders.GetElementAt(folderIndex).QueryInterface(Ci.nsIMsgFolder));    
        }
    }
  }
  return folders;
}
   
// this routine will clear the textIndexed property on every msg
// hdr in the db for the passed in folder. This is for when we blow away the global index
// and need to re-index messages.
function clearIndexedProperty(folder)
{
  try
  {
    var msgDB = folder.getMsgDatabase(null);
    msgDB.dBFolderInfo.NumIndexedMessages = 0;
    folder.updateSummaryTotals(true); // force update from db.
    msgDB.Commit(MSG_DB_LARGE_COMMIT);
    var headerEnumerator = msgDB.EnumerateIndexedMessages();
  } 
  catch (ex) 
  {
    FTIAlert("failed while getting indexed messages enumerator for " + folder.URI + ":" + ex);
    return;
  }
  
  if (gIndexingFeedback)
    gIndexingFeedback.onPreparingToIndex(folder);
 
  try {
    // iterate over the folder finding the next message to index...
    while (headerEnumerator.hasMoreElements())
    {
      var msgHdr = headerEnumerator.getNext().QueryInterface(Components.interfaces.nsIMsgDBHdr);
      msgHdr.markIndexed(false);
    }
  }
  catch (ex)
  {
    FTIAlert("failed while enumerating indexed messages for " + folder.URI + ":" + ex);
  }

  msgDB.Commit(MSG_DB_LARGE_COMMIT);
  folder.setMsgDatabase(null);
}

function onTimer()
{
  try 
  {
    PigLog("BUG 2238: onTimer");
    if (gMsgTextIndexer.msgHdr /* we're currently indexing a msg */)
    {
      PigLog("BUG 2238: currently indexing a msg");
      return;
    }

    // check if we are back from idle manually.
    // we can't wait for the "back" notification, as it can come up to 5 seconds later
    // we want to stop on a dime.  see bug #1194 for details
    // but only do this if we aren't ignoring idle
    // gIgnoreIdle will be true when we are manually indexing
    if (!gIgnoreIdle && Cc["@mozilla.org/widget/idleservice;1"].getService(Ci.nsIIdleService).idleTime < (gIdleTime * 1000))
    {
      Cc['@mozilla.org/msg-global-index;1'].getService(Ci.nsIMsgGlobalIndex).stopIndexing();
      PigLog("BUG 2238: back from idle");
      return;
    }

    // if we have any messages queued up, index the first one
    if (gMsgHdrsToIndex.GetElementAt(0))
    {
      var msgHdr = gMsgHdrsToIndex.QueryElementAt(0, Components.interfaces.nsIMsgDBHdr);
      if (!msgHdr)
        FTIAlert("why is msgHdr null? " + gMsgHdrsToIndex.GetElementAt(0));
      gMsgTextIndexer.IndexMessage(msgHdr);
      PigLog("BUG 2238: index the first one");
      return;
    }

    var msgHdrToIndex = null;

    // if there are no more folders to check, stop indexing.  we'll try again the next time we go back into idle.
    if (gCyclingThroughFoldersState != "start")
    {
      // remove items from message table with folder_id of -1,
      // and any associated rows (in message_text, etc).
      // note, this could happen when offline, 
      // but let's not to keep thing simple.
      // we do this when stopping, and not starting, 
      // as I figure it is more likely that the user might be around 
      // at the start of indexing, but probably not at the end.
      //
      // if we are ignoring idle (gIgnoreIdle), we are manually indexing, 
      // so don't sweep if the user is manually indexing a folder or
      // all folder
      if (!gIgnoreIdle)
        FTIRemoveDeletedItems();
          
      Cc['@mozilla.org/msg-global-index;1'].getService(Ci.nsIMsgGlobalIndex).stopIndexing();
      PigLog("BUG 2238: no more folders to check");
      return;
    }

    // find the current folder we're working on
    if (!gCurrentFolderToIndex)
      FindNextFolderToIndex();
    
    if (gCurrentFolderToIndex)
    {
      var msgHdrToIndex = FindNextHdrToIndex();
    }
    else
    {
      // we've cycled through all the folders, we should take a break
      // from indexing of existing messages
      gCyclingThroughFoldersState = "stopping";
      gPrefBranch.setCharPref(kLastFolderIndexPrefName, "");
    }
      
    if (!msgHdrToIndex)
    {
      PigLog("BUG 2238: reached end of folder");
      gLastFolderIndexedUri = gCurrentFolderToIndex ? gCurrentFolderToIndex.URI : "";
      gPrefBranch.setCharPref(kLastFolderIndexPrefName, gLastFolderIndexedUri);
      SetCurrentFolderToIndex(null);
      gSortedHeadersToIndex = null;

      if (gIndexingOneFolder)
      {
        gCyclingThroughFoldersState = "stopping";
        gIndexingOneFolder = false;
      }
    }
    else
    {
      QueueMessageToGetIndexed(msgHdrToIndex);
    }
    gAlarm = restartTimer(gAlarm, onTimer, 0);
    PigLog("BUG 2238: restart timer");
  }
  catch (ex)
  {
    PigLog("BUG 2238: ex = " + ex);
  }
}

function restartTimer(aAlarm, aOnTimer, aSeconds)
{
  if (aAlarm)
    aAlarm.cancel();

  var gJSlib = Cc["@mozilla.org/url-classifier/jslib;1"].getService().wrappedJSObject;
  return new gJSlib.G_Alarm(aOnTimer, aSeconds*1000);
}

var CreateMsgDisplayedObserver = 
{
  // Components.interfaces.nsIObserver
  observe: function(aHeaderSink, aTopic, aData)
  {
    // as of bug #1509, we index the body on display (see onMessageParsed() in dataPane.js)
    // but we still want to continue to (lazily) update the frecency.
    var msgHdr = gMessenger.msgHdrFromURI(aData);

    if (MessageHeaderIsInvalid(msgHdr))
      return;

    var folderID = GetFolderIdForURI(msgHdr.folder.URI);
    LazyUpdate(msgHdr, folderID, null, gLazyFrecencyUpdateTable);
  }
};

/* keep in sync with nsMsgFolderFlags.h */
var MSG_FOLDER_FLAG_INBOX = 0x1000;
var MSG_FOLDER_FLAG_ARCHIVE = 0x0080;
var MSG_FOLDER_FLAG_SENTMAIL = 0x0200;
var MSG_FOLDER_FLAG_TRASH = 0x0100;
var MSG_FOLDER_FLAG_JUNK = 0x40000000;

/* keep in sync with nsMsgMessageFlags.h */
const MSG_FLAG_IMAP_DELETED = 0x200000;

function DecodeFrecency(msgHdr, folderFlags, oldFrecency)
{
  return CalculateFrecencyInternal(msgHdr, folderFlags, oldFrecency, false /* incrementVisitCount */, true /* explain */, false /* preserveUniqueId */);
}

function CalculateFrecency(msgHdr, folderFlags, oldFrecency, incrementVisitCount, preserveUniqueId)
{
  return CalculateFrecencyInternal(msgHdr, folderFlags, oldFrecency, incrementVisitCount, false /* explain */, preserveUniqueId);
}

function CalculateFrecencyInternal(msgHdr, folderFlags, oldFrecencyString, incrementVisitCount, explain, preserveUniqueId)
{
  // use Math.floor to drop the unique decimal stuff, and convert to a number;
  var oldFrecency = Math.floor(oldFrecencyString);

  // I'm just taking a stab at this. I'd like to use 64 bits for frecency
  // but JS uses floating point numbers so there are only 53 bits
  // of precision. I'm going 
  // to use the top two bits for location (0 for Junk or Trash, 
  // 1 for normal, 2 for Archive or Sent, 3 for Inbox, see bug #1174 for details)
  // I'm going to keep track of the visit count in the low 5 bits
  // bits 6-32 will be some sort of calculation combining most recent read
  // time, visit count, number of tags, number of attachments, etc.
  // bits 33-34 will be the folder bits.
  var folderBits = 1; // normal
  if (folderFlags & MSG_FOLDER_FLAG_INBOX)
    folderBits = 3;
  else if (folderFlags & MSG_FOLDER_FLAG_ARCHIVE || folderFlags & MSG_FOLDER_FLAG_SENTMAIL)
    folderBits = 2;
  else if (folderFlags & MSG_FOLDER_FLAG_JUNK || folderFlags & MSG_FOLDER_FLAG_TRASH)
    folderBits = 0;

  // if the message is junk, and it's not in the junk or trash folders (folderBits == 0)
  // act like it was in the junk folder and use 0 for the folderBits.
  if (folderBits != 0)
  {
    var junkScore = msgHdr.getStringProperty("junkscore");
    var isJunk = ((junkScore != "") && (junkScore != "0"));
    if (isJunk)
      folderBits = 0;
  }

  var visitCount = (oldFrecency & 0x001F);
  
  // max out visit count at 31 (0x001F)
  // see bug #1179 for details
  if (incrementVisitCount && visitCount < 0x001F)
    visitCount++;

  var newTime;
  if (oldFrecency)
  {
    // if there's an old frecency, we're probably reading the
    // message and updating the frecency because of that.
    newTime = Math.floor(Date.now() / 1000);
  }
  else
  {
    // otherwise, we're probably indexing because the message
    // was just added, or we're doing background indexing.
    newTime = msgHdr.dateInSeconds;
  }

  // use the low five bits for visit count (0 - 31, max)
  var lowDWord = (newTime & 0xFFFFFFE0) | visitCount;

  var attachFactor = (msgHdr.flags & MSG_FLAG_ATTACHMENT) ? 1 : 0;
  var starFactor = (msgHdr.flags & MSG_FLAG_MARKED) ? 1 : 0;

  // extract the tag keys from the msgHdr and count them.
  // tagKeysForMsgHdr() comes from chrome://messenger/content/mailCore.js
  // EnsureSubScriptsLoaded() ensures that mailCore.js is loaded into scope.
  EnsureSubScriptsLoaded();
  var tagKeys = tagKeysForMsgHdr(msgHdr);
  var tagFactor = tagKeys ? tagKeys.split(" ").length : 0;

  // do we need to prevent lowDWord from spilling into highDWord?
  // not yet, but if we change our frecency algorithm, we might.  
  // the max lowDWord is 0xFFFFFFFFFF since the folder mask is 0x10000000000
  lowDWord += ((8 * visitCount) + (10 * attachFactor) + (8 * tagFactor) + (12 * starFactor)) << 5;
  highDWord = folderBits * 0x10000000000;
  if (explain)
    return "folder=" + folderBits + " attach=" + attachFactor + " tag=" + tagFactor + " todo=" + starFactor + " time=" + (newTime & 0xFFFFFFE0) + " visit=" + visitCount;
  else
  {
    if (preserveUniqueId)
      return GetNextUniqueStringForSortingInternal(highDWord + lowDWord, oldFrecencyString.split(".")[1]);
    else
      return GetNextUniqueStringForSorting(highDWord + lowDWord);
  }
}

// the nsIFolderListener object
var gFolderListener = {
  OnItemAdded: function(parentItem, item) {},
  OnItemRemoved: function(parentItem, item) {},
  OnItemPropertyChanged: function(item, property, oldValue, newValue) {},
  OnItemIntPropertyChanged: function(item, property, oldValue, newValue) {},
  OnItemBoolPropertyChanged: function(item, property, oldValue, newValue) {},
  OnItemUnicharPropertyChanged: function(item, property, oldValue, newValue){},
  OnItemPropertyFlagChanged: function(item, property, oldFlag, newFlag) {},
  OnItemEvent: function(folder, event) {
    var eventType = event.toString();
    if (eventType == "FolderLoaded") {
      if (folder) {
        var msgFolder = folder.QueryInterface(Components.interfaces.nsIMsgFolder);
        if (msgFolder) 
        {
          PigLog("folder loaded = " + msgFolder.URI);
          CloseFolderDatabaseIfNotInUse(msgFolder);
        }
      }
    }
  }
}

// the nsIMsgFolderListener object
var gMsgFolderListener = {
  itemAdded: function(aItem) 
  {    
    var msgHdr = null;
    try
    {
      msgHdr = aItem.QueryInterface(Components.interfaces.nsIMsgDBHdr);
    }
    catch (ex) {}

    if (msgHdr)
    {
      QueueMessageToGetIndexed(msgHdr);
    }
  },
  
  // folder or msg deleted (no trash)
  itemDeleted: function(aItem)
  {
    var msgHdr;
    var msgHdrArray;

    try {
      var folder = aItem.QueryInterface(Components.interfaces.nsIMsgFolder);
      if (folder)
        FTIRemoveFolder(folder.URI);
      return;
    } catch (ex) {}

    try {
      msgHdr = aItem.QueryInterface(Ci.nsIMsgDBHdr);
    } catch (ex) {msgHdrArray = aItem.QueryInterface(Ci.nsISupportsArray)}
  
    if (msgHdr || msgHdrArray)
    {
      if (msgHdr)
        FTIRemoveMsg(msgHdr);
      else
        FTIRemoveMsgs(msgHdrArray);
    }
  },
  
  itemMoveCopyCompleted: function(aMove, aSrcItems, aDestFolder)
  {
    // see bug #1035 for details
    if (aSrcItems.Count() == 0)
      return;

    var folder;
    try {
      folder = aSrcItems.QueryElementAt(0, Components.interfaces.nsIMsgFolder);
    } catch (ex) {}

    if (folder)
    {
      if (aMove)
        FTIFolderMoved(folder, aDestFolder);
      else
        FTIFolderCopied(folder, aDestFolder);
    }
    else
    {
      var msg = aSrcItems.QueryElementAt(0, Ci.nsIMsgDBHdr);
      if (msg)
      {
        if (MessageHeaderIsInvalid(msg))
          return;

        var numMsgs = aSrcItems.Count();

        var srcFolderID = GetFolderIdForURI(msg.folder.URI);
        var dstFolderID = GetFolderIdForURI(aDestFolder.URI);

        if (srcFolderID == -1 || dstFolderID == -1)
        {
          FTIAlert("itemMoveCopyCompleted failed, either source or dest folder id was -1");
          throw Components.results.NS_ERROR_UNEXPECTED;
        }

        var changesToCommit = [];
        
        for (var msgIndex = 0; msgIndex < numMsgs; msgIndex++)
        {
          msg = aSrcItems.QueryElementAt(msgIndex, Ci.nsIMsgDBHdr);

          if (MessageHeaderIsInvalid(msg))
            continue;

          if (aMove)
          {
            // if we move the message before it's indexed, remove it
            // from list of messages to index.
            gMsgHdrsToIndex.RemoveElement(msg);

            // process any batched headers
            ProcessBatchedHeaders();

            // update the folderID of any matching items from the lazy update tables
            UpdateLazyItems(msg, srcFolderID, dstFolderID);
            
            var moveChangeToCommit = FTIMsgMoved(srcFolderID, dstFolderID, aDestFolder.flags, msg);

            if (moveChangeToCommit) 
              changesToCommit.push(moveChangeToCommit);
          }
          else
          {
            // todo: batch these like we batch FTIMsgMoved()
            FTIMsgCopied(srcFolderID, dstFolderID, aDestFolder.flags, msg);
          }            
        }

        // process commits in batch
        if (changesToCommit.length > 1)
          BeginTransaction();
        
        for (var i = 0; i < changesToCommit.length; i++)
        {
          gMsgMovedStmt.bindInt64Parameter(0, changesToCommit[i].oid);
          gMsgMovedStmt.bindInt64Parameter(1, changesToCommit[i].folderId);
          gMsgMovedStmt.bindStringParameter(2, changesToCommit[i].frecency);
          gMsgMovedStmt.execute();
        }

       if (changesToCommit.length > 1)
         EndTransaction(); 
      }
    }
  },
  
  folderRenamed: function(aOrigFolder, aNewFolder)
  {
    FTIFolderRenamed(aOrigFolder, aNewFolder);
  },

  // extensibility hook
  itemEvent: function(aItem, aEvent, aData)
  {
    if (aEvent == "keywords-change")
    {
      var folderID = GetFolderIdForURI(aItem.folder.URI);
      LazyUpdate(aItem, folderID, null, gLazyTagUpdateTable);
    }
  }, 
};

function QueueMessageToGetIndexed(msgHdr)
{   
  if (msgHdr)
    PigLog("QueueMessageToGetIndexed [" + msgHdr.messageId + "] [" + msgHdr.folder.URI + "] [" + msgHdr.subject + "]");

  // add the message to the message table
  // This allows us to get cross folder conversation even if the message isn't fully indexed yet.
  // And one day, would allow us to move away from .msf files.
  BatchFindOrAddHdrToMsgTable(msgHdr, null /* aBodyText */);
    
  // if we are indexing messages on this server, queue up the message for full indexing
  if (msgHdr.folder.server.indexMsgs)
  {
    if (gMsgHdrsToIndex.GetIndexOf(msgHdr) == -1)
    {
      var isupportsHdr = msgHdr.QueryInterface(Ci.nsISupports);
      gMsgHdrsToIndex.AppendElement(isupportsHdr);  
    }
  }
}

// we'd like to share this code with the message display code, so that when a message
// is displayed, we'd full text index the information at that time as well. Maybe just a utility function
// that we can share somehow? 

// This module will also detect the attachment info while it's at it.

function nsMsgFullTextIndexer() {};

nsMsgFullTextIndexer.prototype =
{
  msgHdr : null,
  numAttachments : 0,
  message: null,

  IndexMessage: function(msgHdr)
  {
    try {  
      if (this.msgHdr)
      {
        FTIAlert("already indexing message");
        return;
      }

      if (MessageHeaderIsInvalid(msgHdr))
      {
        FTIAlert("msg hdr w/o message id");
        this.onDoneStreamingCurMessage(false);
        return;
      }

      FTIDump("indexing message [" + msgHdr.subject + "]\n");
  
      PigLog("BUG 2238: set this.msgHdr [" + msgHdr.subject + "] [" + msgHdr.messageId + "]");
      this.msgHdr = msgHdr;
      this.numAttachments = 0;
      
      var folder = msgHdr.folder;
      if (folder)
      {
        // only stream the message if:
        //   we have it offline already OR
        //   we are online and we won't prompt for password to get the message
        //
        // if we are offline and we attempt to stream a message that is not availabe offline,
        // streamMessage() will throw NS_ERROR_INVALID_POINTER, because there is no msg window
        // see bug #2872 for details
        if ((msgHdr.flags & MSG_FLAG_OFFLINE) || (!gIOService.offline && WontPromptForPassword(folder.server)))
        {
          var uri = folder.getUriForMsg(msgHdr);
          var msgService = gMessenger.messageServiceFromURI(uri);
          PigLog("start streaming message [" + this.msgHdr.messageId + "] [" + this.msgHdr.subject + "]");
          if (gIndexingFeedback)
            gIndexingFeedback.onDownloadingMessage(this.msgHdr);
         
          msgService.streamMessage(uri, this, null, this, true, "stream");
        }
        else
        {
          PigLog("don't start streaming message [" + this.msgHdr.messageId + "] [" + this.msgHdr.subject + "]");
          this.onDoneStreamingCurMessage(false);
          return;
        }
      }
    }
    catch (ex)
    {
      // see bug #1163 for details
      FTIAlert("unexpected failure while indexing message: " + ex);
      this.onDoneStreamingCurMessage(false);
      return;
    }
  },
    
  onStartRequest: function(request, context)
  {      
    // hook up the full text indexer as the msg header sink on the url
    var channel = request.QueryInterface(Ci.nsIChannel);
    var uri = channel.URI;
    var mailnewsUrl = uri.QueryInterface(Ci.nsIMsgMailNewsUrl);
    mailnewsUrl.msgHeaderSink = this.QueryInterface(Ci.nsIMsgHeaderSink);
    this.message = "";
  },

  onStopRequest: function(request, context, status, errorMsg) 
  {
    try
    {
      if (!this.msgHdr)
        FTIAlert("on stop request w/ null msg hdr");
  
      if (status == 0 && this.msgHdr)
      {     
        if (gIndexingFeedback)
          gIndexingFeedback.onIndexingMessage(this.msgHdr);
          
        AddHdrToFullTextIndex(this.msgHdr, this.message);
      }
      else
      {
        FTIAlert("message not successfully indexed! status = " + status);
        this.onDoneStreamingCurMessage(false);    
        return;
      }
    }
    catch (ex)
    {
      FTIAlert("message not successfully indexed! ex = " + ex);
      this.onDoneStreamingCurMessage(false);    
      return;
    }

    this.onDoneStreamingCurMessage(true);
  },

  onDataAvailable: function(request, context, inputStream, offset, count) 
  {
    try 
    {
      var inStream = Cc["@mozilla.org/scriptableinputstream;1"].
             createInstance(Ci.nsIScriptableInputStream);
            
      inStream.init(inputStream);
      
      this.message += inStream.read(count);
    }
    catch (ex)
    {
      FTIAlert("error in ODA" + ex);
    }
  },

  processHeaders: function(aHeaderNames, aHeaderValues, dontCollectAddress)
  {
  },

  isImage: function(aFileName, aContentType) {
    var filesuffix = aFileName.substr(aFileName.lastIndexOf(".")).toLowerCase();
    // XXX todo stop duplicating this image/* logic
    return (aContentType == 'image/jpeg' || aContentType == 'image/jpg' || aContentType == 'image/png' || aContentType == 'image/pjpeg' || aContentType == 'image/gif' ||
            filesuffix == '.jpeg' || filesuffix == '.jpg' || filesuffix == '.png' || filesuffix == '.pjpeg' || filesuffix == '.gif');
  },

  /**
   * handleAttachment is similar to msgHdrViewOverlay.js::updateAttachmentProperties()
   * and should be kept in sync.
   */
  handleAttachment: function(contentType, url, displayName, uri, aNotDownloaded) 
  {
    // don't ignore inline images, see bug #2005 for details
    if ((displayName.match(/^Part 1\.([0-9]|\.)*/) && !this.isImage(displayName, contentType)) ||
        contentType == "text/x-vcard")
    {
      FTIDump("ignoring " + displayName + "\n");
      return;      
    }
    FTIDump("ct = " + contentType + " url = " + url + " displayName = " + displayName + "\n");
    // do the onDataAvailable calls correspond to the handleAttachment calls w.r.t.
    // message/stream offsets?
    this.numAttachments++;
    var attachPropertyRoot = "attach" + this.numAttachments + ".";
    var displayNameUtf8 = gUnicodeConverter.ConvertFromUnicode(displayName);
    displayNameUtf8 += gUnicodeConverter.Finish();
    this.msgHdr.setStringProperty(attachPropertyRoot + "displayName", displayNameUtf8); 
    this.msgHdr.setStringProperty(attachPropertyRoot + "url", url);
    FTIDump("setting " + attachPropertyRoot + "url to " + url + "\n");
    this.msgHdr.setStringProperty(attachPropertyRoot + "contentType", contentType);
    // extract the part number out of the url
    var partArray = url.match(/[&?]part=([0-9]|\.)*/g);
    // detached files don't seem to have a partnum in the url
    if (partArray)
    {
      var part = partArray[0].substring(6);
      FTIDump("handle attachment part = " + part + "\n");
      this.msgHdr.setStringProperty(attachPropertyRoot + "partNum", part);
    }
    
    // mark as having attachment, like we do on display in msgHdrViewOverlay.js::handleAttachment() 
    // see bug #2795 for details.
    this.msgHdr.markHasAttachments(true);
    
    // XXX don't we need to commit these changes to the msf file?     
  },
  
  onEndAllAttachments: function() {this.msgHdr.numAttachments = this.numAttachments;},
    
  // onEndMsgHeaders is called after libmime is done processing a message. At this point it is safe for
  // elements like the UI to update junk status, process return receipts, etc.
  onEndMsgHeaders: function(url) {},


  // onEndMsgDownload is triggered when layout says it is actually done rendering 
  // the message body in the UI. 
  onEndMsgDownload: function(url) {},

  securityInfo : null,

  onMsgHasRemoteContent: function(aMsgHdr) {},
  get dummyMsgHeader() {},
  
  // used as a hook for extension mime content handlers to store data that can later
  // be accessed by other parts of the code, e.g., UI code.
  // TODO - Should replace securityInfo
  get properties() {},

  onDoneStreamingCurMessage: function(successful)
  {
    if (this.msgHdr)
      PigLog("done streaming message, (" + successful + ") [" + this.msgHdr.messageId + "] [" + this.msgHdr.subject + "]");
    else
      PigLog("done streaming message, (" + successful + ") but no header");

    this.message = "";
    if (this.msgHdr != gMsgHdrsToIndex.GetElementAt(0))
    {
      FTIAlert("problem in onDoneStreamingCurMessage, please add details to bug #1584: this.msgHdr=" + this.msgHdr + " gMsgHdrsToIndex.GetElementAt(0)=" + gMsgHdrsToIndex.GetElementAt(0) + " successful=" + successful);
      var zerothHdr = gMsgHdrsToIndex.GetElementAt(0);
      if (!this.msgHdr)
        FTIAlert("null this.msgHdr in onDoneStreamingCurMessage");
      else
      {
        try
        {
          FTIAlert("removing unexpected element - this.msgHdr = " + this.msgHdr.subject + " - " + formatDateHeader(this.msgHdr.date) + "gMsgHdrsToIndex[0] = " + zerothHdr.subject + " - " + formatDateHeader(zerothHdr.date));
        }
        catch (ex)
        {
          // wallpaper over bug #1584, where formatDateHeader() was throwing an exception
          FTIAlert("removing unexpected element, ex = " + ex + "\n");
        }
      }
    }
      
    if (gMsgHdrsToIndex.Count())
      gMsgHdrsToIndex.DeleteElementAt(0);
    else
      FTIAlert("done streaming message, but gMsgHdrsToIndex is empty!\n");

    PigLog("BUG 2238: set this.msgHdr to null");
    this.msgHdr = null;
    onTimer();
  },

  /* nsIUrlListener */
  OnStartRunningUrl : function (aURL)
  {
    PigLog("nsMsgFullTextIndexer OnStartRunningUrl = " + aURL.spec);
  },
  OnStopRunningUrl : function (aURL, aExitCode)
  {
    // XXX todo, see bug #2874 
    // only call onDoneStreamingCurMessage() if the url is one we are streaming
    // other urls can stop with exit codes.
    PigLog("nsMsgFullTextIndexer OnStopRunningUrl = " + aURL.spec + " aExitCode = " + aExitCode);
    if (aExitCode)
    {
      // our streaming failed, possibly due to losing internet access.  
      // we need to call onDoneStreamingCurMessage() to clean things up
      // see bug #2616 for details
      this.onDoneStreamingCurMessage(false);
    }
  },

  QueryInterface: function(aIID)
  {
    if (aIID.equals(Ci.nsIMsgHeaderSink) ||
        aIID.equals(Ci.nsIStreamListener) ||
        aIID.equals(Ci.nsIUrlListener) ||
        aIID.equals(Ci.nsISupports))
      return this;

    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
  }
};

var gFoldersToClear = [];
var gClearAllIndexedAlarm = null;

function ResetGlobalIndex(aDbConn, aFolderURI)
{
  // remove any messages pending to be indexed, as we want indexing to go in folder order
  // or to be just for the specific folder we are indexing (for "Index Message In Folder")
  while (gMsgHdrsToIndex.Count())
    gMsgHdrsToIndex.DeleteElementAt(0);
    
  if (aFolderURI != "")
  {
    FTIRemoveFolder(aFolderURI);
    // clear the "textIndexed" property on every msg hdr in this folders .msf file
    var msgFolder = GetMsgFolderFromUri(aFolderURI, true);
    gLastFolderIndexedUri = "";
    SetCurrentFolderToIndex(msgFolder);
    gSortedHeadersToIndex = null;
    
    gIndexingOneFolder = true;
    gFoldersToClear = [msgFolder];
    gClearAllIndexedAlarm = restartTimer(gClearAllIndexedAlarm, onClearAllIndexedAlarm, 0);
    return;
  }
  
  // reset the last folder we indexed, so that we start at the "top"
  // see bug #1300 for details
  gLastFolderIndexedUri = "";
  SetCurrentFolderToIndex(null);
  gSortedHeadersToIndex = null;
  
  // reset the unique id
  gUniqueId = 0;
  
  // drop the folder id cache
  gFolderIdCache = {};

  // drop the tag id cache
  gTagIdCache = {};

  // drop the mime type id caches
  gMimeTypeIdCache = {};
  gMimeTypeIdIsImageCache = {};

  // drop all pending lazy updates
  gLazyFrecencyUpdateTable = [];
  gLazyUrlInfoUpdateTable = [];
  gLazyBodyInfoUpdateTable = [];
  gLazyAttachmentInfoUpdateTable = [];
  gLazyEventUpdateTable = [];
  gLazyTagUpdateTable = [];
  
  // drop all tables
  // NOTE: when you drop the tables, the indexes are dropped
  aDbConn.beginTransaction();
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS message_text");
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS message");
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS event");
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS refs");
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS tags");
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS tag_ids");
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS mimetypes");
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS mimetype_ids");
  aDbConn.executeSimpleSQL("DROP TABLE IF EXISTS folder");
  aDbConn.commitTransaction();
  
  // recreate all the tables and indexes
  CreateTables();
      
  // clear the "textIndexed" property on every msg hdr in every .msf file
  gFoldersToClear = determineFoldersToClear();
  gClearAllIndexedAlarm = restartTimer(gClearAllIndexedAlarm, onClearAllIndexedAlarm, 0);
}
  
function onClearAllIndexedAlarm()
{
  var folder = gFoldersToClear.pop();
  if (folder)
  {
    clearIndexedProperty(folder);
    gClearAllIndexedAlarm = restartTimer(gClearAllIndexedAlarm, onClearAllIndexedAlarm, 0);
  }
  else
  {
    // if we are ignoring idle, we intend to start indexing once we are done clearing all the folders
    if (gIgnoreIdle)
      Cc['@mozilla.org/msg-global-index;1'].getService(Ci.nsIMsgGlobalIndex).startIndexing();
  }
}

const PR_RDONLY = 0x1;  // see prio.h

var gCatLog = false;

var gCatStart;
var gCatAlarm;
var gCatStream;

var catWorker = function(aFile) {
  try
  {
    if (gCatLog)
      gConsoleService.logStringMessage("catWorker, starting...");
      
    gCatStart = Date.now();
    var fileInputStream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
    fileInputStream.init(aFile, PR_RDONLY, 0, 0);

    gCatStream = Cc["@mozilla.org/scriptableinputstream;1"].createInstance(Ci.nsIScriptableInputStream);
    gCatStream.init(fileInputStream);
  }
  catch (ex)
  {
    Components.utils.reportError(ex);
  }
};

function onCatAlarm()
{
  try
  {
    // while loading msgfti.sqlite into the os fs cache, it is not a good idea to idle index.
    var globalIndex = Components.classes['@mozilla.org/msg-global-index;1'].
                      getService(Components.interfaces.nsIMsgGlobalIndex);
    globalIndex.dontIdleIndex = true;
  }
  catch (ex)
  {
    Components.utils.reportError(ex);
  }
  
  try 
  {  
    var available = gCatStream.available();
    
    if (gCatLog)
      gConsoleService.logStringMessage("catWorker, data available [" + available + "]");
    
    if (available) {
      var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(null);
      var chunkSize = prefs.getIntPref("mail.search.prime_os_fs_cache.chunk_size");
      
      if (gCatLog)
        gConsoleService.logStringMessage("catWorker, reading " + Math.min(chunkSize, available) + " bytes...");
      
      gCatStream.read(Math.min(chunkSize, available));  /* discard all data, we are simulating "cat msgfti.sqlite > /dev/null" */
      gCatAlarm = restartTimer(gCatAlarm, onCatAlarm, 0);
      return;
    }
    
    if (gCatLog)
      gConsoleService.logStringMessage("catWorker, done, took " + (Date.now() - gCatStart) + "ms");
  }
  catch (ex)
  {
    Components.utils.reportError(ex);
  }
  
  try
  {
    // on done, or error, we'll fall through to here.  in either case, it is safe to index now.
    var globalIndex = Components.classes['@mozilla.org/msg-global-index;1'].
                      getService(Components.interfaces.nsIMsgGlobalIndex);
    globalIndex.dontIdleIndex = false;
  }
  catch (ex)
  {
    Components.utils.reportError(ex);
  }
}

catWorker.prototype = {

  run: function() {
    try {
      gCatAlarm = restartTimer(gCatAlarm, onCatAlarm, 0);
    } catch(ex) {
      Components.utils.reportError(ex);
    }
  },
  
  QueryInterface: function(iid) {
    if (iid.equals(Components.interfaces.nsIRunnable) ||
        iid.equals(Components.interfaces.nsISupports)) {
            return this;
    }
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
};

function getDBConn()
{
  if (!gDBConn)
  {
    var dbFile = getFTIDB();
    var primeOsFsCache = false;
    var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(null);
    
    if (gCatLog)
      gConsoleService.logStringMessage("primeOsFsCache? exists: " + dbFile.exists() + " enabled: " + prefs.getBoolPref("mail.search.prime_os_fs_cache.enabled"));
    
    if (dbFile.exists() && prefs.getBoolPref("mail.search.prime_os_fs_cache.enabled"))
      primeOsFsCache = true;
      
    gDBConn = Cc["@mozilla.org/storage/service;1"].getService(Ci.mozIStorageService).openUnsharedDatabase(dbFile);

    if (primeOsFsCache)
    {
      var tm = Cc["@mozilla.org/thread-manager;1"].getService(Ci.nsIThreadManager);
      var target = tm.currentThread;
      target.dispatch(new catWorker(dbFile), target.DISPATCH_NORMAL);
    }
    
    CreateTables(); // creates tables if needed.
  }
  return gDBConn;
}

var gLazyFrecencyUpdateTable = [];
var gLazyUrlInfoUpdateTable = [];
var gLazyEventUpdateTable = [];
var gLazyTagUpdateTable = [];
var gLazyBodyInfoUpdateTable = [];
var gLazyAttachmentInfoUpdateTable = [];

function ProcessLazyTables()
{  
  gLazyFrecencyUpdateTable.reverse();
  var item = gLazyFrecencyUpdateTable.pop();
  while (item)
  {
    UpdateFrecencyForItem(item.msgHdr, item.folderID);
    item = gLazyFrecencyUpdateTable.pop();
  }
  
  gLazyUrlInfoUpdateTable.reverse();
  var item = gLazyUrlInfoUpdateTable.pop();
  while (item)
  {
    UpdateUrlInfoForItem(item.msgHdr, item.folderID, item.data);
    item = gLazyUrlInfoUpdateTable.pop();
  }

  gLazyBodyInfoUpdateTable.reverse();
  var item = gLazyBodyInfoUpdateTable.pop();
  while (item)
  {
    UpdateBodyInfoForItem(item.msgHdr, item.folderID, item.data);
    item = gLazyBodyInfoUpdateTable.pop();
  }

  gLazyAttachmentInfoUpdateTable.reverse();
  var item = gLazyAttachmentInfoUpdateTable.pop();
  while (item)
  {
    UpdateAttachmentInfoForItem(item.msgHdr, item.folderID, item.data);
    item = gLazyAttachmentInfoUpdateTable.pop();
  }
  
  gLazyEventUpdateTable.reverse();
  var item = gLazyEventUpdateTable.pop();
  while (item)
  {
    UpdateEventForItem(item.msgHdr, item.data);
    item = gLazyEventUpdateTable.pop();
  }
  
  gLazyTagUpdateTable.reverse();
  var item = gLazyTagUpdateTable.pop();
  while (item)
  {
    UpdateTagsForItem(item.msgHdr);
    item = gLazyTagUpdateTable.pop();
  }
}

function UpdateLazyItemFromTable(lazyTable, msgHdr, srcFolderID, dstFolderID)
{
  for (var i = 0; i < lazyTable.length; i++)
  {
    if (MessageHeaderIsInvalid(lazyTable[i].msgHdr))
      continue;

    if (lazyTable[i].msgHdr.messageId == msgHdr.messageId && lazyTable[i].folderID == srcFolderID)
    {
      // update the folderID, but keep the item in the lazy table
      FTIDump("change " + srcFolderID + " to " + dstFolderID + " for " + msgHdr.messageId + " in the lazy table\n");
      lazyTable[i].folderID = dstFolderID;

      // we can return here, as there should only be one match, as we called RemoveLazyItemFromTable() from LazyUpdate()
      return;
    }
  }
}

// NOTE:  if msgHdr is null, remove all items for a given folderID
function RemoveLazyItemFromTable(lazyTable, msgHdr, folderID, update)
{
  if (!lazyTable.length)
    return;

  // go from end to start, as we may be splicing out items
  for (var i = lazyTable.length - 1; i >= 0; i--)
  {
    if (MessageHeaderIsInvalid(lazyTable[i].msgHdr))
    {
      lazyTable.splice(i, 1);
    }
    else if ((!msgHdr || lazyTable[i].msgHdr.messageId == msgHdr.messageId) && lazyTable[i].folderID == folderID)
    {
      if (update)
      {
        if (lazyTable == gLazyFrecencyUpdateTable)
          UpdateFrecencyForItem(lazyTable[i].msgHdr, lazyTable[i].folderID);
        else if (lazyTable == gLazyEventUpdateTable)
          UpdateEventForItem(lazyTable[i].msgHdr, lazyTable[i].data);
        else if (lazyTable == gLazyUrlInfoUpdateTable)
          UpdateUrlInfoForItem(lazyTable[i].msgHdr, lazyTable[i].folderID, lazyTable[i].data);
        else if (lazyTable == gLazyBodyInfoUpdateTable)
          UpdateBodyInfoForItem(lazyTable[i].msgHdr, lazyTable[i].folderID, lazyTable[i].data);
        else if (lazyTable == gLazyAttachmentInfoUpdateTable)
          UpdateAttachmentInfoForItem(lazyTable[i].msgHdr, lazyTable[i].folderID, lazyTable[i].data);
        else if (lazyTable == gLazyTagUpdateTable)
          UpdateTagsForItem(lazyTable[i].msgHdr);  
      }

      // remove the matching item from our lazy table 
      FTIDump("remove (" + lazyTable[i].msgHdr + "," + folderID + ") from the lazy table\n");
      lazyTable.splice(i, 1);

      if (msgHdr)
      {
        // we can return here, as there should only be one match, as we called RemoveLazyItemFromTable() from LazyUpdate()
        return;
      }
    }
  }
}

function UpdateLazyItems(msgHdr, srcFolderID, dstFolderID)
{
  UpdateLazyItemFromTable(gLazyFrecencyUpdateTable, msgHdr, srcFolderID, dstFolderID);
  UpdateLazyItemFromTable(gLazyUrlInfoUpdateTable, msgHdr, srcFolderID, dstFolderID);
  UpdateLazyItemFromTable(gLazyBodyInfoUpdateTable, msgHdr, srcFolderID, dstFolderID);
  UpdateLazyItemFromTable(gLazyAttachmentInfoUpdateTable, msgHdr, srcFolderID, dstFolderID);
  UpdateLazyItemFromTable(gLazyEventUpdateTable, msgHdr, srcFolderID, dstFolderID);
  UpdateLazyItemFromTable(gLazyTagUpdateTable, msgHdr, srcFolderID, dstFolderID);
}

// NOTE:  if msgHdr is null, remove all items for a given folderID
function RemoveLazyItemsAndProcessBatchedHeaders(msgHdr, folderID, update)
{
  // XXX todo sspitzer
  // if (update), use BeginTransaction() / EndTransaction()?
  // so that we can batch Update*ForItem() work?
  // see onDelayedTransaction() for an example
  
  // first, process any batched headers.
  ProcessBatchedHeaders();

  // update the lazy items for this message and then remove it from our tables
  // we do this before we copy or move a message so that the lazy data will be correct
  // when deleting a message, we want to remove it, but not update the frecency or urlinfo
  RemoveLazyItemFromTable(gLazyFrecencyUpdateTable, msgHdr, folderID, update);
  RemoveLazyItemFromTable(gLazyUrlInfoUpdateTable, msgHdr, folderID, update);
  RemoveLazyItemFromTable(gLazyBodyInfoUpdateTable, msgHdr, folderID, update);
  RemoveLazyItemFromTable(gLazyAttachmentInfoUpdateTable, msgHdr, folderID, update);
  RemoveLazyItemFromTable(gLazyEventUpdateTable, msgHdr, folderID, update);
  RemoveLazyItemFromTable(gLazyTagUpdateTable, msgHdr, folderID, update);
}

function LazyUpdate(msgHdr, folderID, data, lazyTable)
{
  // if you read the same message quickly, before we got a chance to update the data, it might still be in our tables.
  // so look for a similar lazy item in the appropriate table, and remove it.
  RemoveLazyItemFromTable(lazyTable, msgHdr, folderID, false /* update */);
  lazyTable.push({msgHdr: msgHdr, folderID: folderID, data: data});

  BeginDelayedTransaction();
}

function UpdateMimeTypesForItem(aOid, currentMimeTypesIds, existingMimeTypesIds, aBatch)
{
  var mimeTypesToAdd = [];
  var mimeTypesToRemove = [];

  // we have two sorted arrays of integers, walk the arrays to find out what need to remove or add
  var currentPos = 0;
  var existingPos = 0;
  while (existingPos < existingMimeTypesIds.length && currentPos < currentMimeTypesIds.length)
  {
    if (existingMimeTypesIds[existingPos] < currentMimeTypesIds[currentPos])
    {
      mimeTypesToRemove.push(existingMimeTypesIds[existingPos]);
      existingPos++;
    }
    else if (existingMimeTypesIds[existingPos] > currentMimeTypesIds[currentPos])
    {
      mimeTypesToAdd.push(currentMimeTypesIds[currentPos]);
      currentPos++;
    }
    else
    {
      // in both arrays, nothing to do but advance
      existingPos++;
      currentPos++;
    }
  }

  // if there are more existing, we need to remove them
  while (existingPos < existingMimeTypesIds.length)
  {
    mimeTypesToRemove.push(existingMimeTypesIds[existingPos]);
    existingPos++;
  }

  // if there are more current, we need to add them
  while (currentPos < currentMimeTypesIds.length)
  {
    mimeTypesToAdd.push(currentMimeTypesIds[currentPos]);
    currentPos++;
  }

  FTIDump("add " + aOid + " " + mimeTypesToAdd);
  FTIDump("remove " + aOid + " " + mimeTypesToRemove);

  // XXX TODO
  // we will want to batch up all the lazy commits into one, see bug #2160
  FTIAddMimeTypesForOid(aOid, mimeTypesToAdd, aBatch);
  FTIRemoveMimeTypesForOid(aOid, mimeTypesToRemove, aBatch);
}

function UpdateTagsForItem(msgHdr)
{
  FTIDump("UpdateTagsForItem: " + msgHdr.messageId + "\n");

  // find the oid for this message
  var oid = FTIGetOidFromMessageHdr(msgHdr);
  if (oid == -1)
  {
    FTIDump("UpdateTagsForItem: source message " + msgHdr.messageId + " not indexed, so queue it up\n");
    msgHdr.markIndexed(false);
    // XXX don't we need to commit this change to the msf file?
    QueueMessageToGetIndexed(msgHdr);
    return;
  }

  var existingTagIds = FTIGetTagsForOid(oid);

  // tagKeysForMsgHdr() comes from chrome://messenger/content/mailCore.js
  // EnsureSubScriptsLoaded() ensures that mailCore.js is loaded into scope.
  EnsureSubScriptsLoaded();
  var labels = tagKeysForMsgHdr(msgHdr);
  var currentTagIds = labels ? GetTagIdsForLabels(labels.split(" ")) : [];

  // find the missing tags
  var tagsToAdd = [];
  for (var i = 0; i < currentTagIds.length; i++)
  {
    var tagId = currentTagIds[i];
    if (existingTagIds.indexOf(tagId) == -1)
    {
      tagsToAdd.push(tagId);
    }
  }

  // find the extra tags
  var tagsToRemove = [];
  for (var i = 0; i < existingTagIds.length; i++)
  {
    var tagId = existingTagIds[i];
    if (currentTagIds.indexOf(tagId) == -1)
    {
      tagsToRemove.push(tagId);
    }
  }

  FTIDump("add " + oid + " " + tagsToAdd);
  FTIDump("remove " + oid + " " + tagsToRemove);

  // XXX TODO
  // we will want to batch up all the lazy commits into one, see bug #2160
  BeginTransaction();
  FTIAddTagsForOid(oid, tagsToAdd, false /* aBatch */);
  FTIRemoveTagsForOid(oid, tagsToRemove, false /* aBatch */);
  EndTransaction();
}

function UpdateEventForItem(msgHdr, events)
{
  if (MessageHeaderIsInvalid(msgHdr))
    return;

  // first, get the oid for the message
  // if the message is not in the index, GetOidFromMessageHdr() will return -1;
  var oid = FTIGetOidFromMessageHdr(msgHdr);
  if (oid == -1)
  {
    FTIDump("UpdateEventForItem: source message " + msgHdr.messageId + " not indexed, so queue it up\n");
    msgHdr.markIndexed(false);
    // XXX don't we need to commit this change to the msf file?
    QueueMessageToGetIndexed(msgHdr);
    return;
  }

  // look up any existing events
  var existingEvents = FTIGetEventsForOid(oid);
  
  // don't re-add any existing events
  var eventsToAdd = [];
  for (var i = 0; i < events.length; i++)
  {
    // avoid duplicate dates.  first one we find wins, so we'll follow the order of the detectors in _handlers
    var foundExistingDate = false;
    var currentEventTime = events[i].date;  // date in seconds
    for (var j = 0; j < existingEvents.length; j++)
    {
      foundExistingDate = (currentEventTime == existingEvents[j].date);  // date in seconds
      if (foundExistingDate)
        break;
    }
    if (!foundExistingDate)
      eventsToAdd.push(events[i]);
  }

  // XXX TODO
  // we will want to batch up all the lazy commits into one, see bug #2160
  FTIAddEventsForOid(oid, eventsToAdd, true /* batch */);
}

function UpdateUrlInfoForItem(msgHdr, folderID, urlInfo)
{
  if (MessageHeaderIsInvalid(msgHdr))
    return;

  PigLog("UpdateUrlInfoForItem: set urlinfo for (" + msgHdr.messageId + "," + folderID + ") : " + urlInfo);

  // save urls in .msf file, in case the user is reading the 
  // message before it's indexed. We'll use it when we index later.
  // We also use it to confirm if a message has urls when focusing.
  msgHdr.setStringProperty("urls", urlInfo);
  // XXX don't we need to commit this change to the msf file?  

  // first, get the oid for the message
  // if the message is not in the index, GetOidFromMessageHdr() will return -1;
  var oid = FTIGetOidFromMessageHdr(msgHdr);
  if (oid == -1)
  {
    PigLog("UpdateUrlInfoForItem: source message " + msgHdr.messageId + " not indexed, so queue it up");
    msgHdr.markIndexed(false);
    // XXX don't we need to commit this change to the msf file?
    QueueMessageToGetIndexed(msgHdr);
    return;
  }
 
  var shouldCommitUrlInfo = true;
  // check if the urls have changed, and if not, don't commit
  gUrlInfoLookupStmt.bindInt64Parameter(0, oid);
  if (gUrlInfoLookupStmt.executeStep())
  {
    // check if the body has changed, and if not, don't commit
    shouldCommitUrlInfo = (gUrlInfoLookupStmt.getUTF8String(0) != urlInfo);
  }
  gUrlInfoLookupStmt.reset();
  
  if (shouldCommitUrlInfo)
  {    
    // make sure to preserve null for urls
    gUrlInfoStorageStmt.bindInt64Parameter(0, oid);
    if (urlInfo)
      gUrlInfoStorageStmt.bindUTF8StringParameter(1, urlInfo);
    else
      gUrlInfoStorageStmt.bindNullParameter(1);

    // XXX TODO
    // we will want to batch up all the lazy commits into one, see bug #2160
    gUrlInfoStorageStmt.execute();
  }
}

function UpdateBodyInfoForItem(msgHdr, folderID, bodyInfo)
{
  if (MessageHeaderIsInvalid(msgHdr))
    return;
    
  // first, get the oid for the message
  // if the message is not in the index, GetOidFromMessageHdr() will return -1;
  var oid = FTIGetOidFromMessageHdr(msgHdr);
  if (oid == -1)
  {
    FTIDump("UpdateBodyInfoForItem: source message " + msgHdr.messageId + " not indexed, so queue it up\n");
    msgHdr.markIndexed(false);
    // XXX don't we need to commit this change to the msf file?    
    QueueMessageToGetIndexed(msgHdr);
    return;
  }

  var shouldCommitBodyInfo = true;
  // check if the body has changed, and if not, don't commit
  gBodyLookupStmt.bindInt64Parameter(0, oid);
  if (gBodyLookupStmt.executeStep())
  {
    shouldCommitBodyInfo = (gBodyLookupStmt.getUTF8String(0) != bodyInfo);
  }
  gBodyLookupStmt.reset();

  if (shouldCommitBodyInfo)
  {
    gBodyStorageStmt.bindInt64Parameter(0, oid);
    gBodyStorageStmt.bindUTF8StringParameter(1, bodyInfo);
    gBodyStorageStmt.execute();
  }
  
  // ensure that we think this header is indexed  
  msgHdr.markIndexed(true);
  // XXX don't we need to commit this change to the msf file?   
}

function UpdateAttachmentInfoForItem(msgHdr, folderID, attachmentInfo)
{
  if (MessageHeaderIsInvalid(msgHdr))
    return;

  // first, get the oid for the message
  // if the message is not in the index, GetOidFromMessageHdr() will return -1;
  var oid = FTIGetOidFromMessageHdr(msgHdr);
  if (oid == -1)
  {
    FTIDump("UpdateAttachmentInfoForItem: source message " + msgHdr.messageId + " not indexed, so queue it up\n");
    // XXX don't we need to commit this change to the msf file?    
    QueueMessageToGetIndexed(msgHdr);
    return;
  }
  
  var shouldCommitAttachmentInfo = true;
  // check if the attachments have changed, and if not, don't commit
  gAttachmentsLookupStmt.bindInt64Parameter(0, oid);
  if (gAttachmentsLookupStmt.executeStep())
  {
    shouldCommitAttachmentInfo = (gAttachmentsLookupStmt.getUTF8String(0) != attachmentInfo);
  }
  gAttachmentsLookupStmt.reset();

  if (shouldCommitAttachmentInfo)
  {
    var existingMimeTypesIds = FTIGetSortedMimeTypesForOid(oid);
    
    // ConvertIndexableAttachmentsStringToMimeTypes() comes from chrome://messenger/content/mailCore.js
    // EnsureSubScriptsLoaded() ensures that mailCore.js is loaded into scope.
    EnsureSubScriptsLoaded();
    var currentMimeTypesIds = GetSortedMimeTypeIdsForNames(ConvertIndexableAttachmentsStringToMimeTypes(attachmentInfo));

    BeginTransaction();
    UpdateMimeTypesForItem(oid, currentMimeTypesIds, existingMimeTypesIds, false /* batch */);
    gAttachmentsStorageStmt.bindInt64Parameter(0, oid);
    gAttachmentsStorageStmt.bindUTF8StringParameter(1, attachmentInfo);
    gAttachmentsStorageStmt.execute();
    EndTransaction();
  }
}
  
function UpdateFrecencyForItem(msgHdr, folderID)
{
  FTIDump("UpdateFrecencyForItem\n");
  if (MessageHeaderIsInvalid(msgHdr))
    return;

  FTIDump("set frecency for (" + msgHdr.messageId + "," + folderID + ")\n");
  // update the frecency for this message
  gFrecencyLookupStmt.bindUTF8StringParameter(0, encodeURIComponent(msgHdr.messageId));
  gFrecencyLookupStmt.bindInt64Parameter(1, folderID);

  // XXX sspitzer todo
  // safe to rely on the caller to end any current transaction?
  if (gFrecencyLookupStmt.executeStep())
  {
    var oldFrecency = gFrecencyLookupStmt.getString(0);
    var oid = gFrecencyLookupStmt.getInt64(1);
    try
    {  
      var newFrecency = CalculateFrecency(msgHdr, msgHdr.folder.flags, oldFrecency, true /* incrementVisitCount */, true /* preserveUniqueId */);
      FTIDump("oldFrecency = " + oldFrecency + " newFrecency = " + newFrecency + "\n");
      gFrecencyUpdateStmt.bindStringParameter(0, newFrecency);
      gFrecencyUpdateStmt.bindInt64Parameter(1, oid);
      // XXX TODO
      // we will want to batch up all the lazy commits into one, see bug #2160
      gFrecencyUpdateStmt.execute();
    } catch (ex) {PigLog(ex);}
    gFrecencyLookupStmt.reset();
  }
  else
  {
    // we don't know about this message, so queue it to be indexed.
    gFrecencyLookupStmt.reset();

    FTIDump("UpdateFrecencyForItem: source message " + msgHdr.messageId + " not indexed, so queue it up\n");
    msgHdr.markIndexed(false);
    // XXX don't we need to commit this change to the msf file?
    QueueMessageToGetIndexed(msgHdr); 
  }
}

function onDelayedTransaction()
{
  // we wait for gLazyUpdateTime seconds after the last delayed transaction to process it
  // but make sure we've been idle for that long, otherwise, post-pone it (and wait for the user to be idle)
  // also make sure we are not currently processing batched headers, see bug #2461 for details
  // XXX todo, also check if running a url (indexer's _runningUrls.length)
  PigLog("BEACHBALL: onDelayedTransaction " + gInProcessBatchedHeaders);
  if (!gInProcessBatchedHeaders && Cc["@mozilla.org/widget/idleservice;1"].getService(Ci.nsIIdleService).idleTime < (gLazyUpdateTime * 1000))
  {
    BeginDelayedTransaction();
    return;
  }

  ProcessBatchedHeadersAndLazyTables();
  
  gDelayedTransactionAlarm = null;
}

function ProcessBatchedHeadersAndLazyTables()
{
  // any code that does an executeStep must end any current transaction
  EndTransaction(); 

  // all batched headers need to be processed before we call addAttachmentInfo(), addBodyInfo(), addUrlInfo(), etc.
  // as those assume the header already exists in the message table
  ProcessBatchedHeaders();

  BeginTransaction();
  ProcessLazyTables();
  EndTransaction();
}

function BeginDelayedTransaction()
{
  // wait before committing, to prevent us from committing too frequently.
  gDelayedTransactionAlarm = restartTimer(gDelayedTransactionAlarm, onDelayedTransaction, gLazyUpdateTime);
}

function BeginTransaction()
{
  try {
    if (!gDBConn.transactionInProgress)
      gDBConn.beginTransaction();
  }
  catch (ex)
  {
    PigLog("unexpected failure during beginTransaction, see bug #908: " + ex);
  }
}

function EndTransaction()
{
  try {
    if (gDBConn.transactionInProgress)
      gDBConn.commitTransaction();
  }
  catch (ex)
  {
    PigLog("XXX unexpected failure during commitTransaction, see bug #908: " + ex);
  }
}

function IndexValues(aItem, aBatch)
{
  if (aItem.rowId == -1)
  {
    FTIAlert("skipping [" + aItem.msgHdr.messageId + "], invalid rowId");
    return;
  }
   
  gMsgTextTableUpdateStmt.bindInt64Parameter(0, aItem.rowId);
  gMsgTextTableUpdateStmt.bindUTF8StringParameter(1, aItem.plainBodyText);

  if (aItem.attachmentStr.length)
    gMsgTextTableUpdateStmt.bindUTF8StringParameter(2, aItem.attachmentStr);
  else
    gMsgTextTableUpdateStmt.bindNullParameter(2);

  // make sure to preserve null for urls
  if (aItem.urls)
  {
    gMsgTextTableUpdateStmt.bindUTF8StringParameter(3, aItem.urls);
    aItem.msgHdr.setStringProperty("urls", aItem.urls);
  }
  else
  {
    gMsgTextTableUpdateStmt.bindNullParameter(3);
  }

  gMsgTextTableUpdateStmt.execute();
  gMsgTextTableUpdateStmt.reset();

  UpdateMimeTypesForItem(aItem.rowId, aItem.currentMimeTypesIds, aItem.existingMimeTypesIds, aBatch);

  if (gPrefBranch.getBoolPref("mail.search.events.enabled"))
    FTIAddDatesFromText(aItem.msgHdr.dateInSeconds, aItem.msgHdr.mime2DecodedSubject + "\n" + aItem.plainBodyText, aItem.rowId, aBatch);

  aItem.msgHdr.markIndexed(true);
  
  // can we move these two lines out, to avoid committing every header?
  var msgDB = aItem.msgHdr.folder.getMsgDatabase(null);
  msgDB.Commit(MSG_DB_LARGE_COMMIT);
  
  PigLog("message successfully indexed: [" + aItem.msgHdr.messageId + "]\n");
}

function GetValuesToIndex(aItem)
{
  var rowId = aItem.rowId;
  var bodyText = aItem.bodyText;
  var msgHdr = aItem.msgHdr;
  
  var plainBodyText = "";
  try {
    gBodyLookupStmt.bindInt64Parameter(0, rowId);
    if (gBodyLookupStmt.executeStep())
    {
      plainBodyText = gBodyLookupStmt.getUTF8String(0);
      if (plainBodyText)
        FTIDump("already have plainBodyText = " + plainBodyText + "\n");
    }
  }
  catch (ex)
  {
    FTIAlert("AddHdrToFullTextIndex:  gBodyLookupStmt failed, ex = " + ex);
  }
  gBodyLookupStmt.reset();
  
  try
  {  
    if (!plainBodyText)
    {
      var stringStream = Cc["@mozilla.org/io/string-input-stream;1"].
        createInstance(Ci.nsIStringInputStream);
      stringStream.setData(bodyText, bodyText.length);    
      var maxOutputLen = gPrefBranch.getIntPref("mail.fulltextsearch.max_body_length");
      
      // if we have the first reference, compress the quotes out of the body before we index
      // so that we don't index the same content twice.
      // if we don't have the first reference, don't compress, so that we'll index the quoted content.
      // see bug #3005 for details
      var globalIndex = Components.classes['@mozilla.org/msg-global-index;1'].getService(Components.interfaces.nsIMsgGlobalIndex);
      var compressQuotes = globalIndex.hasFirstReference(msgHdr);      
      plainBodyText = msgHdr.folder.getMsgTextFromStream(stringStream, maxOutputLen, compressQuotes);
    }
        
    // even if plainBodyText is null, by using bindUTF8StringParameter we'll set body to "" (and not null)
    // we want as "" means the message has been indexed, but it has no body
    // where null means the message body has not been indexed
    aItem.plainBodyText = plainBodyText;
    
    // ConvertAttachmentsToIndexableString() comes from chrome://messenger/content/mailCore.js
    // EnsureSubScriptsLoaded() ensures that mailCore.js is loaded into scope.
    EnsureSubScriptsLoaded();
    aItem.attachmentStr = ConvertAttachmentsToIndexableString(GetAttachmentsFromHeader(msgHdr));

    var urls = msgHdr.getStringProperty("urls");
    if (urls.length)
    {
      aItem.urls = urls;
    }
    else if (plainBodyText.length)
    {
      var urls = [];
      var titles = [];

      // use something like mozTXTToHTMLConv::FindURLInPlaintext(), but make it do exactly what I want
      // see bug #1030 for details
      var scannedTxt = gUrlDetector.scanTXT(plainBodyText, Ci.mozITXTToHTMLConv.kURLs);
      // if the scannedText is the same, there were no urls
      if (scannedTxt != plainBodyText)
      {
        var pos = 0;
        while (pos < scannedTxt.length)
        {
          // see mozTXTToHTMLConv::CheckURLAndCreateHTML()
          var anchor = scannedTxt.indexOf('<a class="moz-txt-link-', pos);
          if (anchor == -1)
            break;
          
          pos = anchor + '<a class="moz-txt-link-'.length;
          var href = scannedTxt.indexOf('href="', pos);
          if (href == -1)
            continue;

          pos = href + 'href="'.length;
          var end = scannedTxt.indexOf('">', pos);
          if (end == -1)
            continue;

          urls.push(scannedTxt.substring(pos, end));
          // for now, we don't get the title out of the message before we view it
          // see bug #1030 for details
          titles.push('');
          pos = end + 1;
        }
      }
      aItem.urls = convertUrlsAndTitlesToIndexableString(urls, titles);
    }
    
    aItem.existingMimeTypesIds = FTIGetSortedMimeTypesForOid(rowId);
       
    // ConvertIndexableAttachmentsStringToMimeTypes() comes from chrome://messenger/content/mailCore.js
    // EnsureSubScriptsLoaded() ensures that mailCore.js is loaded into scope.
    EnsureSubScriptsLoaded();
    aItem.currentMimeTypesIds = GetSortedMimeTypeIdsForNames(ConvertIndexableAttachmentsStringToMimeTypes(aItem.attachmentStr));
  }
  catch (ex) 
  {
    FTIAlert("unexpected error during GetValuesToIndex(): " + ex);
    // We will skip anything with rowId == -1 in IndexValues()
    aItem.rowId = -1;
  }
}

var gInProcessBatchedHeaders = false;

function ProcessBatchedHeaders()
{
  var start = Date.now();
  var count = gBatchedHeadersToAdd.length;
  PigLog("ProcessBatchedHeaders, count = " + count);
  if (!count)
    return;

  gInProcessBatchedHeaders = true;
  var headersToIndex = [];
  var item = gBatchedHeadersToAdd.pop();
  while (item)
  {
    // if rowId is not -1, the header already existed in message, so don't add it
    // but if body text is not null, we should index it.
    if (item.rowId != -1)
    {
      if (item.bodyText)
        headersToIndex.push({msgHdr: item.msgHdr, bodyText: item.bodyText, rowId: item.rowId});
    }
    else
    {
      try
      {
        // something might have interrupted our transaction, so attempt to begin it again.
        BeginTransaction();
        // don't batch, because we are already in a transaction
        var rowId = AddHdrToMsgTable(item.msgHdr, gBatchedHeadersToAddFolderId, false /* batch */);
        if (rowId == -1)
          FTIAlert("AddHdrToMsgTable failed\n");
        else if (item.bodyText)
          headersToIndex.push({msgHdr: item.msgHdr, bodyText: item.bodyText, rowId: rowId});
      }
      catch (ex)
      {
        FTIAlert("AddHdrToMsgTable() failed, ex = " + ex);
      }
    }
    item = gBatchedHeadersToAdd.pop();
  }
  EndTransaction();
  PigLog("ProcessBatchedHeaders, AddHdrToMsgTable took " + (Date.now() - start) + "ms");
  
  PigLog("ProcessBatchedHeaders, headersToIndex.length = " + headersToIndex.length);
  for (var i = 0; i < headersToIndex.length; i++)
  {
    GetValuesToIndex(headersToIndex[i]);
  }
  PigLog("ProcessBatchedHeaders, GetValuesToIndex took " + (Date.now() - start) + "ms");

  for (var i = 0; i < headersToIndex.length; i++)
  {
    // something might have interrupted our transaction, so attempt to begin it again.
    BeginTransaction();
    IndexValues(headersToIndex[i], false /* aBatch */);
  }
  EndTransaction();
  PigLog("ProcessBatchedHeaders, IndexValues took " + (Date.now() - start) + "ms");
  
  PigLog("ProcessBatchedHeaders of (" + count + "," + headersToIndex.length + ") took " + (Date.now() - start) + "ms");
  gInProcessBatchedHeaders = false;
}

function BatchFindOrAddHdrToMsgTable(aMsgHdr, aBodyText)
{  
  if (MessageHeaderIsInvalid(aMsgHdr))
    return;

  // check if the message is already in the message table.
  var rowId = FTIGetOidFromMessageHdr(aMsgHdr);
  if (rowId != -1 && !aBodyText)
  {
    // if aBodyText is null, we aren't going to index the body
    // since rowId is not -1, the message is already in the message table, so we're done.
    return;
  }
  
  // do we have any bodies to process?
  var haveBodyTexts = aBodyText;
  var folderURI = aMsgHdr.folder.URI;
  var messageId = aMsgHdr.messageId;
  
  // don't batch the same header twice
  // see bug #1701 for details
  //
  // start at the end and work backwards.  since we append headers to gBatchedHeadersToAdd
  // the one we want is almost always at the end.
  for (var i = gBatchedHeadersToAdd.length - 1; i >= 0; i--)
  {
    if (gBatchedHeadersToAdd[i].bodyText)
      haveBodyTexts = true;
      
    if (gBatchedHeadersToAdd[i].msgHdr.messageId == messageId && gBatchedHeadersToAdd[i].msgHdr.folder.URI == folderURI)
    {
      // if this header is already batched, but the item doesn't have bodyText, use the current bodyText.
      // this happens on indexing, as we first call BatchFindOrAddHdrToMsgTable() from QueueMessageToGetIndexed() [from onTimer()], 
      // and then again from AddHdrToFullTextIndex() [from onStopRequest()] after streaming the message.
      if (aBodyText && !gBatchedHeadersToAdd[i].bodyText)
      {
        gBatchedHeadersToAdd[i].bodyText = aBodyText;
      } 
      return;
    }
  }

  var now = Date.now();
  var folderID = GetFolderIdForURI(folderURI);
  var shouldProcessBatchedHeaders = haveBodyTexts ? (gBatchedHeadersToAdd.length >= gBatchMaxWithBodies) : (gBatchedHeadersToAdd.length >= gBatchMaxWithoutBodies);  
  // if the new message is in a different folder that the other headers, or if it has been long enough, or it has been enough headers, process all the batched headers.
  // Do "time since the last header" and not "time since first header". see bug #2461 for details
  // XXX todo, also check if running a url (indexer's _runningUrls.length)  
  if (gBatchedHeadersToAdd.length && (shouldProcessBatchedHeaders || gBatchedHeadersToAddFolderId != folderID || ((now - gBatchedHeadersToAdd[gBatchedHeadersToAdd.length - 1].time) >= gBatchDelay)))
  {
    PigLog("BatchFindOrAddHdrToMsgTable: length = " + gBatchedHeadersToAdd.length + " delay = " + (now - gBatchedHeadersToAdd[gBatchedHeadersToAdd.length - 1].time) + " folderIds = " + gBatchedHeadersToAddFolderId + " vs " + folderID);
    ProcessBatchedHeaders();
  }

  gBatchedHeadersToAdd.push({msgHdr: aMsgHdr, time: now, bodyText: aBodyText, rowId: rowId});
  gBatchedHeadersToAddFolderId = folderID;  
}

function AddHdrToMsgTable(msgHdr, folderID, batch)
{
  gMsgTableStorageStmt.bindUTF8StringParameter(0, encodeURIComponent(msgHdr.messageId));
  gMsgTableStorageStmt.bindInt64Parameter(1, folderID);
  var frecency = CalculateFrecency(msgHdr, msgHdr.folder.flags, "0.0" /* oldFrecency */, true /* incrementVisitCount */, false /* preserveUniqueId */);
  gMsgTableStorageStmt.bindStringParameter(2, frecency);
  gMsgTableStorageStmt.bindStringParameter(3, GetNextUniqueStringForSorting(msgHdr.dateInSeconds));
  
  PigLog("BEACHBALL: inserting into message message id = " + msgHdr.messageId + " folderID = " + folderID + " date = " + msgHdr.dateInSeconds);
  gMsgTableStorageStmt.execute();
  var rowId = getDBConn().lastInsertRowID;
 
  var refs = [];
  for (var i = 0; i < msgHdr.numReferences; i++)
  {
    refs.push(encodeURIComponent(msgHdr.getStringReference(i)));
    PigLog("BEACHBALL: adding reference for message id = " + msgHdr.messageId + " rowId = " + rowId + " ref = " + msgHdr.getStringReference(i));
  }
  FTIAddRefsForOid(rowId, refs, batch);
  
  var tags = [];
  // tagKeysForMsgHdr() comes from chrome://messenger/content/mailCore.js
  // EnsureSubScriptsLoaded() ensures that mailCore.js is loaded into scope.
  EnsureSubScriptsLoaded();
  var labels = tagKeysForMsgHdr(msgHdr);
  var tagIds = labels ? GetTagIdsForLabels(labels.split(" ")) : [];
  for (var i = 0; i < tagIds.length; i++)
  {
    tags.push(tagIds[i]);
  }
  FTIAddTagsForOid(rowId, tags, batch);

  try
  {
    gMsgTextTablePartialStorageStmt.bindInt64Parameter(0, rowId);
    
    var utf8Subject = gUnicodeConverter.ConvertFromUnicode(msgHdr.mime2DecodedSubject);
    utf8Subject += gUnicodeConverter.Finish();
    gMsgTextTablePartialStorageStmt.bindUTF8StringParameter(1, utf8Subject);
  
    var author = msgHdr.mime2DecodedAuthor;
    var utf8Author = gUnicodeConverter.ConvertFromUnicode(author);
    utf8Author += gUnicodeConverter.Finish();
    gMsgTextTablePartialStorageStmt.bindUTF8StringParameter(2, utf8Author);
  
    var recipients = msgHdr.mime2DecodedRecipients;
    var utf8Recipients = gUnicodeConverter.ConvertFromUnicode(recipients);
    utf8Recipients += gUnicodeConverter.Finish();
    // also index the cc list, see bug #799 for details 
    var ccList = msgHdr.mime2DecodedCcList;
    if (utf8Recipients && utf8Recipients.length && ccList && ccList.length)
      utf8Recipients += ", ";
    utf8Recipients += gUnicodeConverter.ConvertFromUnicode(ccList);
    utf8Recipients += gUnicodeConverter.Finish();
    gMsgTextTablePartialStorageStmt.bindUTF8StringParameter(3, utf8Recipients);
    
    gMsgTextTablePartialStorageStmt.execute();
  }
  catch (ex) 
  {
    FTIAlert("FindOrAddHdrToMsgTable failed, ex = " + ex); 
    gMsgTextTablePartialStorageStmt.reset();
  }
  return rowId;
}

var gSubScriptsLoaded = false;

function EnsureSubScriptsLoaded()
{
  if (gSubScriptsLoaded)
    return;

  Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
            .getService(Components.interfaces.mozIJSSubScriptLoader)
            .loadSubScript("chrome://messenger/content/date.js", null);
            
  Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
            .getService(Components.interfaces.mozIJSSubScriptLoader)
            .loadSubScript("chrome://messenger/content/dateDetector.js", null);

  Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
            .getService(Components.interfaces.mozIJSSubScriptLoader)
            .loadSubScript("chrome://messenger/content/mailCore.js", null);

  gSubScriptsLoaded = true;
}

function FTIAddDatesFromText(aSeconds, aSubjectAndBody, aRowId, aBatch)
{
  // gDateDetector comes from chrome://messenger/content/dateDetector.js
  // EnsureSubScriptsLoaded() ensures that dateDetector.js is loaded into scope.
  EnsureSubScriptsLoaded();
  
  var results = gDateDetector.convert(aSeconds * 1000, aSubjectAndBody);
  var events = [];
  for (var i = 0; i < results.length; i++)
  {
    // getTime() is in milliseconds, we want seconds
    events.push({date: Math.floor(results[i].date.getTime() / 1000), text: results[i].text});
  }
  
  FTIAddEventsForOid(aRowId, events, aBatch);
}

function DoubleCheckBody(aMsgHdr, aBodyText)
{
  if (aBodyText.indexOf("This body part will be downloaded on demand.") != -1)
  {
    FTIAlert("hitting mime parts on demand bug #1648, please add details: messageId=" + aMsgHdr.messageId + " subject=" + aMsgHdr.mime2DecodedSubject);
  }
}

function AddHdrToFullTextIndex(msgHdr, aBodyText)
{
  DoubleCheckBody(msgHdr, aBodyText);
  BatchFindOrAddHdrToMsgTable(msgHdr, aBodyText);
}
      
function CreateTables()
{
  var dbConn = getDBConn();

  // PRAGMA page_size must be first.
  dbConn.executeSimpleSQL("PRAGMA page_size = 4096");
  
  // this breaks SQL Manager - we can't have that for now!
  // dbConn.executeSimpleSQL("PRAGMA locking_mode = EXCLUSIVE");
    
  var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(null);
  var cacheSize = prefs.getIntPref("mail.search.cache_size");
  dbConn.executeSimpleSQL("PRAGMA default_cache_size = " + cacheSize);
  
  var dbVersion = dbConn.schemaVersion;
  if (dbVersion != kCurDBVersion)
    dbConn.schemaVersion = kCurDBVersion;
    
  if (dbVersion != kCurDBVersion)
  {
    ResetGlobalIndex(dbConn, "" /* aFolderURI */);
  }

  if (!dbConn.tableExists("message"))
  {
    // message.frecency and message.date are TEXT not INTEGER because we are allowing the user to sort by message.frecency or message.date
    // and for sorting and pagination to work, message.frecency and message.date must be unique.
    // see bug #1321 for details.
    dbConn.executeSimpleSQL("CREATE TABLE message (message_id TEXT NOT NULL, folder_id INTEGER, frecency TEXT, date TEXT)");
  }
  
  if (!dbConn.indexExists("message_folder_id"))
    dbConn.executeSimpleSQL("CREATE INDEX message_folder_id ON message (folder_id)");
  if (!dbConn.indexExists("frecency_index"))
    dbConn.executeSimpleSQL("CREATE INDEX frecency_index ON message (frecency)");
  if (!dbConn.indexExists("message_date_index"))
    dbConn.executeSimpleSQL("CREATE INDEX message_date_index ON message (date)");

  // after pig.1 and before pig.2, I added a refs column to the message table, and an index, but they are gone now.
  // temporary code, remove it at some point
  if (dbConn.indexExists("message_refs_index"))
    dbConn.executeSimpleSQL("DROP INDEX message_refs_index");

  // drop these indexes.  see bug #2755 for details  
  if (dbConn.indexExists("mimetypes_fk_index"))
    dbConn.executeSimpleSQL("DROP INDEX mimetypes_fk_index");
  if (dbConn.indexExists("mimetypes_is_image_index"))
    dbConn.executeSimpleSQL("DROP INDEX mimetypes_is_image_index");
 
  if (!dbConn.tableExists("folder"))
    dbConn.executeSimpleSQL("CREATE TABLE folder (uri TEXT NOT NULL UNIQUE)");

  // use a multi-column index.  we have several queries that do "WHERE message_id = ?1 AND folder_id = ?2"
  // and this will speed those up.  see bug #838 for details
  dbConn.executeSimpleSQL("DROP INDEX IF EXISTS folder_id_index");
  dbConn.executeSimpleSQL("DROP INDEX IF EXISTS message_id_index");
  if (!dbConn.indexExists("message_id_folder_id_index"))
    dbConn.executeSimpleSQL("CREATE INDEX message_id_folder_id_index ON message (message_id, folder_id)");

  // if the order of the columns in message_text change, make sure to update executeFullTextQuery() / offsets() code
  if (!dbConn.tableExists("message_text")) 
  {
    // if you add a column, make sure to adjust matchCondition() in focusPane.js
    dbConn.executeSimpleSQL("CREATE VIRTUAL TABLE message_text USING fts3(subject, sender, recipients, body, attachments, urls)");
  }
  
  if (!dbConn.tableExists("event"))
  {
    // fk refers to a message.rowid
    // event.date is a TEXT not an INTEGER because we are allowing the user to sort by event.date
    // and for sorting and pagination to work, event.date must be unique.
    // see bug #1321 for details.
    dbConn.executeSimpleSQL("CREATE TABLE event (fk INTEGER NOT NULL, date TEXT, text TEXT)");
  }
  
  if (!dbConn.indexExists("event_fk_index"))
    dbConn.executeSimpleSQL("CREATE INDEX event_fk_index ON event (fk)");

  if (!dbConn.indexExists("event_date_index"))
    dbConn.executeSimpleSQL("CREATE INDEX event_date_index ON event (date)");

  if (!dbConn.tableExists("refs"))
  {
    // fk refers to a message.rowid, ref is a reference (message_id)
    // note, a message can have multiple references
    dbConn.executeSimpleSQL("CREATE TABLE refs (fk INTEGER NOT NULL, ref TEXT)");
  }

  if (!dbConn.indexExists("refs_fk_index"))
    dbConn.executeSimpleSQL("CREATE INDEX refs_fk_index ON refs (fk)");
    
  if (!dbConn.indexExists("refs_ref_index"))
    dbConn.executeSimpleSQL("CREATE INDEX refs_ref_index ON refs (ref)");

  if (!dbConn.tableExists("tags"))
  {
    // fk refers to a message.rowid, tag_id is a reference to a row in the tag_ids table
    // note, a message can have multiple tags
    dbConn.executeSimpleSQL("CREATE TABLE tags (fk INTEGER NOT NULL, tag_id INTEGER NOT NULL)");
  }

  if (!dbConn.tableExists("tag_ids"))
    dbConn.executeSimpleSQL("CREATE TABLE tag_ids (label TEXT NOT NULL UNIQUE)");

  if (!dbConn.indexExists("tags_fk_index"))
    dbConn.executeSimpleSQL("CREATE INDEX tags_fk_index ON tags (fk)");
    
  if (!dbConn.indexExists("tags_tag_id_index"))
    dbConn.executeSimpleSQL("CREATE INDEX tags_tag_id_index ON tags (tag_id)");

  if (!dbConn.tableExists("mimetypes"))
  {
    // fk refers to a message.rowid, mimetype_id is a reference to a row in the mimetype_ids table
    // note, a message can have multiple attachments of multiple types
    dbConn.executeSimpleSQL("CREATE TABLE mimetypes (fk INTEGER NOT NULL, is_image INTEGER DEFAULT 0, mimetype_id INTEGER NOT NULL)");
  }

  if (!dbConn.tableExists("mimetype_ids"))
    dbConn.executeSimpleSQL("CREATE TABLE mimetype_ids (name TEXT NOT NULL UNIQUE)");

  // see bug #2755 for details about this index
  if (!dbConn.indexExists("mimetypes_fk_is_image_index"))
    dbConn.executeSimpleSQL("CREATE INDEX mimetypes_fk_is_image_index ON mimetypes (fk, is_image)");
    
  if (!dbConn.indexExists("mimetypes_mimetype_id_index"))
    dbConn.executeSimpleSQL("CREATE INDEX mimetypes_mimetype_id_index ON mimetypes (mimetype_id)");

  gUniqueId = 0;
  var hasMoreStmt = getDBConn().createStatement("SELECT message.date, message.frecency FROM message ORDER BY rowid DESC LIMIT 1");
  if (hasMoreStmt.executeStep())
  {
    // turn "123.456" into 456
    gUniqueId = Math.floor(hasMoreStmt.getString(0).split(".")[1]);
    var maxFrecencyUniqueId = Math.floor(hasMoreStmt.getString(1).split(".")[1]);
    if (maxFrecencyUniqueId > gUniqueId)
      gUniqueId = maxFrecencyUniqueId;
  }
  hasMoreStmt.reset();

  hasMoreStmt = getDBConn().createStatement("SELECT event.date FROM event ORDER BY rowid DESC LIMIT 1");
  if (hasMoreStmt.executeStep())
  {
    // turn "123.456" into 456
    var maxEventDateUniqueId = Math.floor(hasMoreStmt.getString(0).split(".")[1]);
    if (maxEventDateUniqueId > gUniqueId)
      gUniqueId = maxEventDateUniqueId;
  }
  hasMoreStmt.reset();
}

function InitSQLStatements()
{
  var dbConn = getDBConn();
  gMsgTextTablePartialStorageStmt = dbConn.createStatement("INSERT INTO message_text (rowid, subject, sender, recipients) VALUES (?1, ?2, ?3, ?4)");
  gMsgTextTableStorageStmt = dbConn.createStatement("INSERT INTO message_text (rowid, subject, sender, recipients, body, attachments, urls) VALUES (?1, ?2, ?3, ?4, ?5, ?6, ?7)");
  gMsgTextTableUpdateStmt = dbConn.createStatement("UPDATE message_text SET body = ?2, attachments = ?3, urls = ?4 WHERE rowid = ?1");
  gMsgTableStorageStmt = dbConn.createStatement("INSERT INTO message (message_id, folder_id, frecency, date) VALUES (?1, ?2, ?3, ?4)");
  gFolderTableStorageStmt = dbConn.createStatement("INSERT INTO folder (uri) VALUES (?1)");
  gFolderURILookupStmt = dbConn.createStatement("SELECT rowid FROM folder WHERE uri = ?1");
  gMsgTxtLookupStmt = dbConn.createStatement("SELECT subject, sender, recipients, body, attachments, frecency, urls FROM message JOIN message_text ON message_text.rowid = message.rowid WHERE message.rowid = ?1");
  gFrecencyLookupStmt = dbConn.createStatement("SELECT frecency, oid FROM message WHERE message_id = ?1 AND folder_id = ?2");
  gMessageLookupStmt = dbConn.createStatement("SELECT oid FROM message WHERE message_id = ?1 AND folder_id = ?2");
  gFrecencyUpdateStmt = dbConn.createStatement("UPDATE message SET frecency = ?1 WHERE oid = ?2");
  gMsgMovedStmt = dbConn.createStatement("UPDATE message SET folder_id = ?2, frecency = ?3 WHERE oid = ?1");
  gUrlInfoStorageStmt = dbConn.createStatement("UPDATE message_text SET urls = ?2 WHERE rowid = ?1");
  gUrlInfoLookupStmt = dbConn.createStatement("SELECT urls FROM message_text WHERE rowid = ?1");
  gAttachmentsStorageStmt = dbConn.createStatement("UPDATE message_text SET attachments = ?2 WHERE rowid = ?1");
  gAttachmentsLookupStmt = dbConn.createStatement("SELECT attachments FROM message_text WHERE rowid = ?1");
  gBodyStorageStmt = dbConn.createStatement("UPDATE message_text SET body = ?2 WHERE rowid = ?1");
  gBodyLookupStmt = dbConn.createStatement("SELECT body FROM message_text WHERE rowid = ?1");
  gEventLookupStmt = dbConn.createStatement("SELECT date, text FROM event WHERE fk = ?1");
  gEventStorageStmt = dbConn.createStatement("INSERT INTO event (fk, date, text) VALUES (?1, ?2, ?3)");
  gTotalMessageCountForFolder = dbConn.createStatement("SELECT COUNT(*) FROM message WHERE message_id = ?1 AND folder_id = ?2");
  gTotalCountForFolder = dbConn.createStatement("SELECT COUNT(*) FROM message WHERE folder_id = ?1");
  // use > -1 instead of != -1, so that we use the message_folder_id index
  // see bug #2842 for details
  gTotalCount = dbConn.createStatement("SELECT COUNT(*) FROM message WHERE folder_id > -1");
  gTotalPendingDeleteCount = dbConn.createStatement("SELECT COUNT(*) FROM message WHERE folder_id = -1");
  gFolderLookupStmt = dbConn.createStatement("SELECT uri FROM message JOIN folder ON message.folder_id = folder.rowid WHERE message_id = ?1");
  gDistinctMessageIdAndFolderLookupStmt = dbConn.createStatement("SELECT DISTINCT message_id, uri FROM message JOIN folder ON message.folder_id = folder.rowid JOIN refs ON message.rowid = refs.fk WHERE refs.ref = ?1");
  gAllMessageIdsInFolder = dbConn.createStatement("SELECT message_id FROM message WHERE folder_id = ?1");
  
  gRefsStorageStmt = dbConn.createStatement("INSERT INTO refs (fk, ref) VALUES (?1, ?2)");
  gRefsLookupStmt = dbConn.createStatement("SELECT ref FROM refs WHERE fk = ?1");

  gTagsStorageStmt = dbConn.createStatement("INSERT INTO tags (fk, tag_id) VALUES (?1, ?2)");
  gTagsLookupStmt = dbConn.createStatement("SELECT tag_id FROM tags WHERE fk = ?1");
  gTagsRemoveStmt = dbConn.createStatement("DELETE FROM tags WHERE fk = ?1 AND tag_id = ?2");
  gTagLabelStorageStmt = dbConn.createStatement("INSERT INTO tag_ids (label) VALUES (?1)");
  gTagLabelLookupStmt = dbConn.createStatement("SELECT rowid FROM tag_ids WHERE label = ?1");

  gMimeTypesStorageStmt = dbConn.createStatement("INSERT INTO mimetypes (fk, is_image, mimetype_id) VALUES (?1, ?2, ?3)");
  gMimeTypesLookupStmt = dbConn.createStatement("SELECT mimetype_id FROM mimetypes WHERE fk = ?1");
  // we can have duplicate values, so delete should only delete one at a time.
  gMimeTypesRemoveStmt = dbConn.createStatement("DELETE FROM mimetypes WHERE rowid in (SELECT rowid FROM mimetypes WHERE fk = ?1 AND mimetype_id = ?2 LIMIT 1)");
  gMimeTypeNameStorageStmt = dbConn.createStatement("INSERT INTO mimetype_ids (name) VALUES (?1)");
  gMimeTypeNameLookupStmt = dbConn.createStatement("SELECT rowid FROM mimetype_ids WHERE name = ?1");
  
  // use > -1 instead of != -1, so that we use the message_folder_id index
  // see bug #2842 for details
  gHasMessage = dbConn.createStatement("SELECT COUNT(*) FROM message WHERE message_id = ?1 AND folder_id > -1 LIMIT 1");
}

function FTIAddEventsForOid(aOid, aEvents, aBatch)
{      
  if (aBatch && aEvents.length > 1)
    BeginTransaction();
    
  for (var i = 0; i < aEvents.length; i++)
  {
    gEventStorageStmt.bindInt64Parameter(0, aOid);
    var uniqueEventDate = GetNextUniqueStringForSorting(aEvents[i].date);
    gEventStorageStmt.bindStringParameter(1, uniqueEventDate);
    gEventStorageStmt.bindStringParameter(2, aEvents[i].text);
    gEventStorageStmt.execute();
  }
  
  if (aBatch && aEvents.length > 1)
    EndTransaction();
}

function FTIAddRefsForOid(aOid, aRefs, aBatch)
{      
  if (aBatch && aRefs.length > 1)
    BeginTransaction();
    
  for (var i = 0; i < aRefs.length; i++)
  {
    gRefsStorageStmt.bindInt64Parameter(0, aOid);
    gRefsStorageStmt.bindStringParameter(1, aRefs[i]);
    gRefsStorageStmt.execute();
  }
  
  if (aBatch && aRefs.length > 1)
    EndTransaction();
}

function FTIAddMimeTypesForOid(aOid, aMimeTypes, aBatch)
{      
  if (aBatch && aMimeTypes.length > 1)
    BeginTransaction();
    
  for (var i = 0; i < aMimeTypes.length; i++)
  {
    gMimeTypesStorageStmt.bindInt64Parameter(0, aOid);
    gMimeTypesStorageStmt.bindInt32Parameter(1, gMimeTypeIdIsImageCache[aMimeTypes[i]] ? 1 : 0);
    gMimeTypesStorageStmt.bindInt64Parameter(2, aMimeTypes[i]);
    gMimeTypesStorageStmt.execute();
  }
  
  if (aBatch && aMimeTypes.length > 1)
    EndTransaction();
}

function FTIRemoveMimeTypesForOid(aOid, aMimeTypes, aBatch)
{      
  if (aBatch && aMimeTypes.length > 1)
    BeginTransaction();
    
  for (var i = 0; i < aMimeTypes.length; i++)
  {
    // gMimeTypesRemoveStmt will only remove one at a time, since we could have duplicates
    gMimeTypesRemoveStmt.bindInt64Parameter(0, aOid);
    gMimeTypesRemoveStmt.bindInt64Parameter(1, aMimeTypes[i]);
    gMimeTypesRemoveStmt.execute();
  }
  
  if (aBatch && aMimeTypes.length > 1)
    EndTransaction();
}

function FTIAddTagsForOid(aOid, aTags, aBatch)
{      
  if (aBatch && aTags.length > 1)
    BeginTransaction();
    
  for (var i = 0; i < aTags.length; i++)
  {
    gTagsStorageStmt.bindInt64Parameter(0, aOid);
    gTagsStorageStmt.bindInt64Parameter(1, aTags[i]);
    gTagsStorageStmt.execute();
  }
  
  if (aBatch && aTags.length > 1)
    EndTransaction();
}

function FTIRemoveTagsForOid(aOid, aTags, aBatch)
{      
  if (aBatch && aTags.length > 1)
    BeginTransaction();
    
  for (var i = 0; i < aTags.length; i++)
  {
    gTagsRemoveStmt.bindInt64Parameter(0, aOid);
    gTagsRemoveStmt.bindInt64Parameter(1, aTags[i]);
    gTagsRemoveStmt.execute();
  }
  
  if (aBatch && aTags.length > 1)
    EndTransaction();
}

function FTIGetOidFromMessageHdr(aMsgHdr)
{
  var oid = -1;
  if (MessageHeaderIsInvalid(aMsgHdr))
    return oid;

  // any code that does an executeStep must end any current transaction
  EndTransaction();

  var folderID = GetFolderIdForURI(aMsgHdr.folder.URI);
  gMessageLookupStmt.bindUTF8StringParameter(0, encodeURIComponent(aMsgHdr.messageId));
  gMessageLookupStmt.bindInt64Parameter(1, folderID);
  try
  {
    if (gMessageLookupStmt.executeStep())
      oid = gMessageLookupStmt.getInt64(0);
  } catch (ex) {PigLog("FTIGetOidFromMessageHdr() failed: " + ex);}
  gMessageLookupStmt.reset();
  return oid;
}

function getFTIDB()
{
  var dirSvc = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties);
  var db = dirSvc.get("ProfD", Ci.nsIFile);
  db.append("msgfti.sqlite");
  return db;
}

var gMimeTypeIdCache = {};
var gMimeTypeIdIsImageCache = {};

// this will add the label to the mimetype_ids table if it doesn't
// exist, and return the new row id.
function GetMimeTypeIdForName(aName)
{
  if (gMimeTypeIdCache[aName])
    return gMimeTypeIdCache[aName];

  var retMimeTypeId;
  // any code that does an executeStep must end any current transaction
  EndTransaction();

  gMimeTypeNameLookupStmt.bindUTF8StringParameter(0, aName);
  try
  {
    if (gMimeTypeNameLookupStmt.executeStep())
    {
      retMimeTypeId = gMimeTypeNameLookupStmt.getInt64(0);
    }
    else
    {
      try
      {
        gMimeTypeNameStorageStmt.bindUTF8StringParameter(0, aName);
        gMimeTypeNameStorageStmt.execute();
        retMimeTypeId = getDBConn().lastInsertRowID;
      } catch (ex) {PigLog("GetMimeTypeIdForName() failed: " + ex);}
    }
  }
  catch (ex) {PigLog("GetMimeTypeIdForName() failed: " + ex);}
  
  gMimeTypeNameLookupStmt.reset();
  gMimeTypeIdCache[aName] = retMimeTypeId;
  // XXX todo stop duplicating this image/* logic
  gMimeTypeIdIsImageCache[retMimeTypeId] = (aName == 'image/jpeg' || aName == 'image/jpg' || aName == 'image/png' || aName == 'image/pjpeg' || aName == 'image/gif');
  return retMimeTypeId;
}
    
function GetSortedMimeTypeIdsForNames(aNames)
{
  var mimeTypeIds = [];
  for (var i=0; i < aNames.length; i++)
  {
    mimeTypeIds.push(GetMimeTypeIdForName(aNames[i]));
  }
  return mimeTypeIds.sort();
}

var gTagIdCache = {};

// this will add the label to the tag_ids table if it doesn't
// exist, and return the new row id.
function GetTagIdForLabel(aLabel)
{
  if (gTagIdCache[aLabel])
    return gTagIdCache[aLabel];

  var retTagId;
  // any code that does an executeStep must end any current transaction
  EndTransaction();

  gTagLabelLookupStmt.bindUTF8StringParameter(0, aLabel);
  try
  {
    if (gTagLabelLookupStmt.executeStep())
    {
      retTagId = gTagLabelLookupStmt.getInt64(0);
    }
    else
    {
      try
      {
        gTagLabelStorageStmt.bindUTF8StringParameter(0, aLabel);
        gTagLabelStorageStmt.execute();
        retTagId = getDBConn().lastInsertRowID;
      } catch (ex) {PigLog("GetTagIdForLabel() failed: " + ex);}
    }
  }
  catch (ex) {PigLog("GetTagIdForLabel() failed: " + ex);}
  
  gTagLabelLookupStmt.reset();
  gTagIdCache[aLabel] = retTagId;
  return retTagId;
}
    
function GetTagIdsForLabels(aLabels)
{
  var tagIds = [];
  for (var i=0; i < aLabels.length; i++)
  {
    tagIds.push(GetTagIdForLabel(aLabels[i]));
  }
  return tagIds;
}
 
var gFolderIdCache = {};

function GetFolderIdForURIInternal(uri, aCreate)
{
  if (gFolderIdCache[uri])
    return gFolderIdCache[uri];

  var retFolderID;
  // any code that does an executeStep must end any current transaction
  EndTransaction();

  gFolderURILookupStmt.bindUTF8StringParameter(0, uri);
  try
  {
    if (gFolderURILookupStmt.executeStep())
    {
      retFolderID = gFolderURILookupStmt.getInt64(0);
    }
    else if (aCreate)
    {
      // the uri doesn't exist in the folder table, so create it.
      try
      {
        gFolderTableStorageStmt.bindUTF8StringParameter(0, uri);
        gFolderTableStorageStmt.execute();
        retFolderID = getDBConn().lastInsertRowID;
      } catch (ex) {PigLog("GetFolderIdForURI() failed: " + ex);}
    }
    else 
    {
      gFolderURILookupStmt.reset();
      return -1;
    }
  }
  catch (ex) {PigLog("GetFolderIdForURI() failed: " + ex);}
  
  gFolderURILookupStmt.reset();
  gFolderIdCache[uri] = retFolderID;
  return retFolderID;
}

// this will add the uri to the folder table if it doesn't
// exist, and return the new row id.
function GetFolderIdForURI(uri)
{
  return GetFolderIdForURIInternal(uri, true /* aCreate */);
}

function GetFolderIdForURIIfExists(uri)
{
  return GetFolderIdForURIInternal(uri, false /* aCreate */);
}

function FTIGetEventsForOid(aOid)
{
  var events = [];
  // any code that does an executeStep must end any current transaction
  EndTransaction();

  gEventLookupStmt.bindInt64Parameter(0, aOid);
  while (gEventLookupStmt.executeStep())
  {
    // use Math.floor to drop the unique parts of the event.date string, and turn it into seconds
    events.push({date: Math.floor(gEventLookupStmt.getString(0)), text: gEventLookupStmt.getString(1)});
  }
  gEventLookupStmt.reset();

  return events;
}

function FTIGetRefsForOid(aOid)
{
  var refs = [];
  // any code that does an executeStep must end any current transaction
  EndTransaction();

  gRefsLookupStmt.bindInt64Parameter(0, aOid);
  while (gRefsLookupStmt.executeStep())
  {
    refs.push(gRefsLookupStmt.getString(0));
  }
  gRefsLookupStmt.reset();

  return refs;
}

function FTIGetTagsForOid(aOid)
{
  var tags = [];
  // any code that does an executeStep must end any current transaction
  EndTransaction();

  gTagsLookupStmt.bindInt64Parameter(0, aOid);
  while (gTagsLookupStmt.executeStep())
  {
    tags.push(gTagsLookupStmt.getInt64(0));
  }
  gTagsLookupStmt.reset();

  return tags;
}

function FTIGetSortedMimeTypesForOid(aOid)
{
  var mimeTypes = [];
  // any code that does an executeStep must end any current transaction
  EndTransaction();

  gMimeTypesLookupStmt.bindInt64Parameter(0, aOid);
  while (gMimeTypesLookupStmt.executeStep())
  {
    mimeTypes.push(gMimeTypesLookupStmt.getInt64(0));
  }
  gMimeTypesLookupStmt.reset();

  return mimeTypes.sort();
}

function FTIRemoveMsgsByMessageIDs(aFolderId, aEscapeAndQuotedMessageIDs)
{
  var start = Date.now();
  // remove message now means:
  //   1) set the folder_id to -1 for any messages in this folder matching these message ids
  //
  // on idle, we'll remove those rows from the message table with folder_id = -1
  // and any rows associated with these rows in other tables.
  // See FTIRemoveDeletedItems() for details.
  var dbConn = getDBConn();
  // if we have multiple messages in a folder with the same message id, this will remove all instances, and we'll fix it on re-index, see bug #1123 and bug #2789
  var updateMessageStmt = dbConn.createStatement("UPDATE message SET folder_id = -1 WHERE message_id IN (" + aEscapeAndQuotedMessageIDs + ") AND folder_id = " + aFolderId);
  updateMessageStmt.execute();
  PigLog("FTIRemoveMsgsByMessageIDs, " + aFolderId + ", " + aEscapeAndQuotedMessageIDs + ", UPDATE message SET folder_id = -1 WHERE... took " + (Date.now() - start) + "ms");
}

function FTIRemoveMsg(msgHdr)
{
  if (MessageHeaderIsInvalid(msgHdr))
    return;

  var folderId = GetFolderIdForURI(msgHdr.folder.URI);
  var escapeAndQuotedMessageIDs = encodeURIComponentAndQuote(msgHdr.messageId);

  // if we remove the message before it's indexed, remove it
  // from list of messages to index.
  gMsgHdrsToIndex.RemoveElement(msgHdr);

  // remove matching items from the lazy update tables, and don't bother updating the frecency or urlinfo
  RemoveLazyItemsAndProcessBatchedHeaders(msgHdr, folderId, false /* update */);
  
  FTIRemoveMsgsByMessageIDs(folderId, escapeAndQuotedMessageIDs);
}

function FTIRemoveMsgs(msgHdrArray)
{
  var folderId = "";
  var escapeAndQuotedMessageIDs = "";

  for (var i = 0; i < msgHdrArray.Count(); ++i)
  {
    var msgHdr = msgHdrArray.GetElementAt(i).QueryInterface(Components.interfaces.nsIMsgDBHdr);

    if (MessageHeaderIsInvalid(msgHdr))
      continue;

    if (!folderId)
      folderId = GetFolderIdForURI(msgHdr.folder.URI);
      
    if (escapeAndQuotedMessageIDs)
      escapeAndQuotedMessageIDs += ",";
    escapeAndQuotedMessageIDs += encodeURIComponentAndQuote(msgHdr.messageId);

    // if we remove the message before it's indexed, remove it
    // from list of messages to index.
    gMsgHdrsToIndex.RemoveElement(msgHdr);

    // remove matching items from the lazy update tables, and don't bother updating the lazy data
    RemoveLazyItemsAndProcessBatchedHeaders(msgHdr, folderId, false /* update */);
  }
  
  if (folderId && escapeAndQuotedMessageIDs)
    FTIRemoveMsgsByMessageIDs(folderId, escapeAndQuotedMessageIDs);
}

function encodeURIComponentAndQuote(aMessageID)
{
  // encodeURIComponent() will convert " to %22, but will not convert '
  // so use " to wrap all message ids.
  return '"' + encodeURIComponent(aMessageID) + '"';
}

function FTIRemoveDeletedItems()
{
  var start = Date.now();
  
  var dbConn = getDBConn();
  BeginTransaction();  

  var removeEventStmt = dbConn.createStatement("DELETE FROM event WHERE event.fk IN (SELECT rowid FROM message WHERE folder_id = -1)");
  removeEventStmt.execute();
  PigLog("FTIRemoveDeletedItems, DELETE FROM event took " + (Date.now() - start) + "ms");

  var removeRefsStmt = dbConn.createStatement("DELETE FROM refs WHERE refs.fk IN (SELECT rowid FROM message WHERE folder_id = -1)");
  removeRefsStmt.execute();
  PigLog("FTIRemoveDeletedItems, DELETE FROM refs took " + (Date.now() - start) + "ms");
  
  var removeTagsStmt = dbConn.createStatement("DELETE FROM tags WHERE tags.fk IN (SELECT rowid FROM message WHERE folder_id = -1)");
  removeTagsStmt.execute();
  PigLog("FTIRemoveDeletedItems, DELETE FROM tags took " + (Date.now() - start) + "ms");

  var removeMimeTypesStmt = dbConn.createStatement("DELETE FROM mimetypes WHERE mimetypes.fk IN (SELECT rowid FROM message WHERE folder_id = -1)");
  removeMimeTypesStmt.execute();
  PigLog("FTIRemoveDeletedItems, DELETE FROM mimetypes took " + (Date.now() - start) + "ms");

  var removeMsgTextStmt = dbConn.createStatement("DELETE FROM message_text WHERE message_text.docid IN (SELECT rowid FROM message WHERE folder_id = -1)");
  removeMsgTextStmt.execute();
  PigLog("FTIRemoveDeletedItems, DELETE FROM message_text took " + (Date.now() - start) + "ms");

  var removeMsgStmt = dbConn.createStatement("DELETE FROM message WHERE folder_id = -1");
  removeMsgStmt.execute();
  PigLog("FTIRemoveDeletedItems, DELETE FROM message took " + (Date.now() - start) + "ms");
  
  EndTransaction();
  PigLog("FTIRemoveDeletedItems, EndTransaction took " + (Date.now() - start) + "ms");
}  

function FTIRemoveFolder(aFolderURI)
{
  var start = Date.now();
  
  var deletedFolderID = GetFolderIdForURIIfExists(aFolderURI);
  PigLog("FTIRemoveFolder = " + deletedFolderID);
 
  // we don't know about this folder yet, so nothing to do.
  if (deletedFolderID == -1)
  {
    PigLog("FTIRemoveFolder, " + aFolderURI + " not known, so do nothing. took " + (Date.now() - start) + "ms");
    return;
  }

  // remove all matching items from the lazy update tables
  RemoveLazyItemsAndProcessBatchedHeaders(null, deletedFolderID, false /* update */);
  PigLog("FTIRemoveFolder, " + aFolderURI + ", RemoveLazyItemsAndProcessBatchedHeaders took " + (Date.now() - start) + "ms");
  
  var dbConn = getDBConn();
  BeginTransaction();   
  
  // remove folder now means:
  //   1) set the folder_id to -1 for any messages in this folder
  //   2) remove the folder from the folder table
  //   3) drop the folder cache
  //
  // on idle, we'll remove those rows from the message table with folder_id = -1
  // and any rows associated with these rows in other tables.
  // See FTIRemoveDeletedItems() for details.
  var updateMessageStmt = dbConn.createStatement("UPDATE message SET folder_id = -1 WHERE folder_id = " + deletedFolderID);
  updateMessageStmt.execute();
  FTIAlert("FTIRemoveFolder, " + aFolderURI + ", UPDATE message SET folder_id = -1 took " + (Date.now() - start) + "ms");

  var removeFolderStmt = dbConn.createStatement("DELETE FROM folder WHERE oid = " + deletedFolderID);
  removeFolderStmt.execute();
  PigLog("FTIRemoveFolder, " + aFolderURI + ", DELETE FROM folder took " + (Date.now() - start) + "ms");

  EndTransaction();
  // upon folder removal, clear the folder id cache.  see bug #996 for details
  gFolderIdCache = {};
  
  PigLog("FTIRemoveFolder, " + aFolderURI + ", EndTransaction took " + (Date.now() - start) + "ms");
}

function FTIFolderRenamed(aOrigFolder, aNewFolder)
{
  FTIDump("folder renamed origFolderPath = " + aOrigFolder.filePath.path + "newFolderPath = " + aNewFolder.filePath.path + "\n");

  var folderRenamedStmt = getDBConn().createStatement("UPDATE folder SET uri = ?1 WHERE uri = ?2");
  folderRenamedStmt.bindUTF8StringParameter(0, aNewFolder.URI);
  folderRenamedStmt.bindUTF8StringParameter(1, aOrigFolder.URI);
  folderRenamedStmt.execute();
  
  // upon rename, clear the folder id cache.  see bug #996 for details
  gFolderIdCache = {};
}

function FTIFolderMoved(origFolder, destParentFolder)
{
  var newFolder = destParentFolder.getChildNamed(origFolder.name).QueryInterface(Ci.nsIMsgFolder);
  FTIFolderRenamed(origFolder, newFolder);
}

function FTIFolderCopied(origFolder, destParentFolder)
{
  var newFolder = destParentFolder.getChildNamed(origFolder.name).QueryInterface(Ci.nsIMsgFolder);
}

function FTIMsgMoved(aSrcFolderID, aDstFolderID, aDstFolderFlags, aMsgHdr)
{
  var oldFrecency = 0;
  var oid = -1;

  gFrecencyLookupStmt.bindUTF8StringParameter(0, encodeURIComponent(aMsgHdr.messageId));
  gFrecencyLookupStmt.bindInt64Parameter(1, aSrcFolderID);
  if (gFrecencyLookupStmt.executeStep())
  {
    oldFrecency = gFrecencyLookupStmt.getString(0);
    oid = gFrecencyLookupStmt.getInt64(1);
  }
  gFrecencyLookupStmt.reset();

  if (oid == -1)
  { 
    FTIDump("FTIMsgMoved: source message " + aMsgHdr.messageId + " not indexed, so queue it up\n");
    aMsgHdr.markIndexed(false);
    // XXX don't we need to commit this change to the msf file?
    QueueMessageToGetIndexed(aMsgHdr); 
    return null;
  }
  
  var newFrecency = CalculateFrecency(aMsgHdr, aDstFolderFlags, oldFrecency, false /* incrementVisitCount */, true /* preserveUniqueId */);
  FTIDump("FTIMsgMoved: oldFrecency = " + oldFrecency + " newFrecency = " + newFrecency + "\n");  
  return ({oid: oid, folderId: aDstFolderID, frecency: newFrecency});
}

function formatDateHeader(aDate)
{
  var dateService = Components.classes["@mozilla.org/intl/scriptabledateformat;1"].
                      getService(Components.interfaces.nsIScriptableDateFormat);
  var msgDate = new Date(aDate);
  var currentDate = new Date();
  var dateFormat = dateService.dateFormatShort;
  // same day?
  if (currentDate.getFullYear() == msgDate.getFullYear() &&
      currentDate.getMonth() == msgDate.getMonth() &&
      currentDate.getDate() == msgDate.getDate())
    dateFormat = dateService.dateFormatNone;
  return dateService.FormatDateTime("", dateFormat,
                             dateService.timeFormatNoSeconds,
                             msgDate.getFullYear(), msgDate.getMonth()+1, msgDate.getDate(),
                             msgDate.getHours(),msgDate.getMinutes(), msgDate.getSeconds());
}

function FTIMsgCopied(aSrcFolderID, aDstFolderID, aDstFolderFlags, msgHdr)
{
  if (MessageHeaderIsInvalid(msgHdr))
    return;

  var dbConn = getDBConn();
  
  // any code that does an executeStep must end any current transaction
  EndTransaction(); 

  // update any matching pending lazy changes and then remove them
  RemoveLazyItemsAndProcessBatchedHeaders(msgHdr, aSrcFolderID, true /* update */);

  // XXX what if msgHdr (by the messageId) is already in aDstFolderID?
  // this can happen if the user copies the message into the same folder twice
  
  var sourceOid = FTIGetOidFromMessageHdr(msgHdr);
  if (sourceOid == -1)
  {
    FTIDump("FTIMsgCopied: source message " + msgHdr.messageId + " not indexed, so queue it up\n");
    msgHdr.markIndexed(false);
    // XXX don't we need to commit this change to the msf file?
    QueueMessageToGetIndexed(msgHdr);
    return;
  }

  // look up any existing events, refs, tags, and mime types
  var existingEvents = FTIGetEventsForOid(sourceOid);
  var existingRefs = FTIGetRefsForOid(sourceOid);
  var existingTags = FTIGetTagsForOid(sourceOid);
  var existingMimeTypes = FTIGetSortedMimeTypesForOid(sourceOid);

  BeginTransaction();

  FTIDump("copying into message table: messageId = " + msgHdr.messageId + " folderID = " + aDstFolderID + "\n");
  // now we want to copy the row for that message in the full text index table, and use the new rowid
  // from the previous insert into the message table as the rowid of the new row in the full text index table.
  try
  {
    gMsgTxtLookupStmt.bindUTF8StringParameter(0, sourceOid);
    gMsgTxtLookupStmt.executeStep();
    var subject = gMsgTxtLookupStmt.getUTF8String(0);
    var sender = gMsgTxtLookupStmt.getUTF8String(1);
    var recipients = gMsgTxtLookupStmt.getUTF8String(2);
    var body = gMsgTxtLookupStmt.getUTF8String(3);
    var attachments = gMsgTxtLookupStmt.getUTF8String(4);
    var oldFrecency = gMsgTxtLookupStmt.getInt64(5);
    var urls = gMsgTxtLookupStmt.getUTF8String(6);

    // compute the new frecency using the old frecency and the destination folder's flags
    var newFrecency = CalculateFrecency(msgHdr, aDstFolderFlags, oldFrecency, false /* incrementVisitCount */, false /* preserveUniqueId */);
    gMsgTableStorageStmt.bindUTF8StringParameter(0, encodeURIComponent(msgHdr.messageId));
    gMsgTableStorageStmt.bindInt64Parameter(1, aDstFolderID);
    gMsgTableStorageStmt.bindStringParameter(2, newFrecency);
    gMsgTableStorageStmt.bindStringParameter(3, GetNextUniqueStringForSorting(msgHdr.dateInSeconds));
    gMsgTableStorageStmt.execute();
  }
  catch (ex) {
    // this is very bad, as we are going to use lastInsertRowID() to copy over references and events.
    // in the case of references, if we mess up the reference table, conversation gathering and tag slamming (by conversation)
    // will be horked.  see bug #1675 for details.
    FTIAlert("hitting bug #1675, please add details to the bug: " + ex);
    gMsgTableStorageStmt.reset();
    gMsgTxtLookupStmt.reset();
    EndTransaction();
    return;
  }
  gMsgTxtLookupStmt.reset();
  
  var destRowId = getDBConn().lastInsertRowID;

  try
  {
    gMsgTextTableStorageStmt.bindInt64Parameter(0, destRowId);
    gMsgTextTableStorageStmt.bindUTF8StringParameter(1, subject);
    gMsgTextTableStorageStmt.bindUTF8StringParameter(2, sender);
    gMsgTextTableStorageStmt.bindUTF8StringParameter(3, recipients);
    // if the original message has null for the body, it has not been indexed
    // we want to preserve that, and not set "" (which means the message has been indexed, but it has no body)
    if (body == null)
      gMsgTextTableStorageStmt.bindNullParameter(4);
    else
      gMsgTextTableStorageStmt.bindUTF8StringParameter(4, body);

    gMsgTextTableStorageStmt.bindUTF8StringParameter(5, attachments);

    // make sure to preserve null for urls
    if (urls)
      gMsgTextTableStorageStmt.bindUTF8StringParameter(6, urls);
    else
      gMsgTextTableStorageStmt.bindNullParameter(6);

    gMsgTextTableStorageStmt.execute();
  }
  catch (ex) {PigLog("FTIMsgCopied() failed: " + ex); gMsgTextTableStorageStmt.reset();}
  
  // copy over any events
  // don't batch, because we are already in a transaction
  FTIAddEventsForOid(destRowId, existingEvents, false /* batch */);

  // copy over any refs
  // don't batch, because we are already in a transaction
  FTIAddRefsForOid(destRowId, existingRefs, false /* batch */);
  
  // copy over any tags
  // don't batch, because we are already in a transaction
  FTIAddTagsForOid(destRowId, existingTags, false /* batch */);

  // copy over any mime types
  // don't batch, because we are already in a transaction
  FTIAddMimeTypesForOid(destRowId, existingMimeTypes, false /* batch */);

  EndTransaction();
}

var gFTIDump = false;
var gFTIAlert = false;

var gDoneCreatingIndexerService = false;

function PigLog(str)
{
  if (!gPigLogToDisk)
    return;
    
  if (!gDoneCreatingIndexerService)
  {
    // in msgGlobalIndexer() we call addIdleObserver
    // we might be idle (on startup), and calling addIdleObserver() can call PigLog()
    // which tries to getService() to create the indexer, which is in the middle of being created.
    // see bug #2339 for details
    return;
  }

  try
  {
    Cc['@mozilla.org/msg-global-index;1'].getService(Ci.nsIMsgGlobalIndex).log(str);
  }
  catch (ex)
  {  
    dump("XXX unexpected failure to PigLog() = " + ex + "\n");
  }
}

function FTIDump(str)
{
  if (gFTIDump)
    dump("XXX FTIDump: " + str);
}

function FTIAlert(text)
{
  dump("XXX FTIAlert: " + text + "\n");

  gConsoleService.logStringMessage("FTIAlert: " + text);

  PigLog("FTIAlert: " + text);
  
  if (gFTIAlert)
  {
    var ww = Cc["@mozilla.org/embedcomp/window-watcher;1"].
             getService(Ci.nsIWindowWatcher);
    ww.getNewPrompter(null).alert("Indexing error", text);
  }  
}

/* copied from widgetglue.js, perhaps we should move it to mailCore.js? */
function GetMsgFolderFromUri(uri, checkFolderAttributes)
{
  if (!uri.length)
    return null;

  var msgfolder = null;
  try 
  {
    var resource = Cc['@mozilla.org/rdf/rdf-service;1'].getService(Ci.nsIRDFService).GetResource(uri);

    msgfolder = resource.QueryInterface(Ci.nsIMsgFolder);
    if (checkFolderAttributes && !(msgfolder && (msgfolder.parent || msgfolder.isServer)))
    {
      PigLog("XXX " + uri + " parent=" + msgfolder.parent + " isServer=" + msgfolder.isServer + "\n");
      msgfolder = null;
    }
  }
  catch (ex) 
  {
    PigLog("XXX GetMsgFolderFromUri " + ex + "\n");
  }
  return msgfolder;
}
