/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/***** BEGIN POSTBOX LICENSE BLOCK *****
 *
 * pbAnalytics.js
 *
 * Copyright (c) 2007-2009 Postbox, Inc.
 *
 ***** END POSTBOX LICENSE BLOCK *****/

const Cc = Components.classes;
const Ci = Components.interfaces;

// Module object
var PbAnalyticsMod = {
  firstTime: true, 
  cid: Components.ID("{dd12429d-21c7-473a-9fb6-2e4f735024da}"),
  progid: "@postbox-inc.com/analytics;1",
  className: "Postbox Analytics",

factory:  
  {
  createInstance: function (aOuter, aIID) 
    {
      if (aOuter != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
      
      if (!aIID.equals(Ci.nsISupports) &&
          !aIID.equals(Ci.nsIPbAnalytics))
        throw Components.results.NS_ERROR_INVALID_ARG;
      // return the singleton
      var singletonGlobalIndexer = new pbAnalytics;
      return singletonGlobalIndexer.QueryInterface(aIID);
    }       
  }, // factory
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Ci.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    
      if (aCID.equals(this.cid))
        return this.factory;
    
    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {        
    if (this.firstTime) 
    {
      this.firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Ci.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(this.cid, this.className, this.progid, aFileSpec, aLocation, aType);  
  },

  unregisterSelf: function(aCompMgr, aFileSpec, aLocation)
  {
    aCompMgr = aCompMgr.QueryInterface(Ci.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(this.cid, aFileSpec);  
  },

  canUnload: function(aCompMgr)
  {
    return true;
  }
};
                            
function NSGetModule(aCompMgr, aFileSpec)
{
  return PbAnalyticsMod;
}

// constructor for pbAnalytics
function pbAnalytics()
{
  this.observerService =
    Components.classes["@mozilla.org/observer-service;1"]
              .getService(Components.interfaces.nsIObserverService);

  this.observerService.addObserver(this, "pb:analytics", false);
  this.observerService.addObserver(this, "quit-application", false);
}

// sdbm hash from
// http://www.cse.yorku.ca/~oz/hash.html
function sdbm_hash(str) 
{
  var hash = 0; 
  for (var i = 0; i < str.length; i++) { 
    hash = str[i].charCodeAt(str[i]) + (hash << 6) + (hash << 16) - hash; 
  } 
  return hash;
}

var gAnalyticsPrefBranch = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch("pb.analytics.");
var gPrefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(null);

function Debug(aStr)
{
  if (gPrefs.getBoolPref("pb.analytics_service.debug"))
    dump("XXX ANALYTICS: " + aStr + "\n");
}

const START_PAGE_PREF = "pb.start_page.url";

function UsingDefaultStartPage()
{
  var startPage = gPrefs.getCharPref(START_PAGE_PREF);

  var defaultStartPage = Cc["@mozilla.org/preferences-service;1"].
                           getService(Ci.nsIPrefService).
                           getDefaultBranch(null).
                           getCharPref(START_PAGE_PREF);

  return (startPage == defaultStartPage ? "1" : "0");
}

pbAnalytics.prototype = {
  QueryInterface: function(aIID)
  {
    if (aIID.equals(Ci.nsISupports))
      return this;
    else if (aIID.equals(Ci.nsIPbAnalytics))
      return this;
    else if (aIID.equals(Ci.nsIObserver))
      return this;
    
    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
  },

  get data()
  {
    if (gPrefs.getBoolPref("pb.analytics_service.enabled") == false)
    {
      this.reset();
      return "";
    }

    // always report at least one statistic, so that we tell between people who don't report
    // and those who do report, but don't do anything that we collect
    // see bug #2594
    var queryString = sdbm_hash("analytics") + "=" + 1;
 
    // is the user using the default start page?
    // see bug #3298
    queryString += "&" + sdbm_hash("defaultstartpage") + "=" + UsingDefaultStartPage();
   
    var analyticsPrefList = gAnalyticsPrefBranch.getChildList("", {});
    for each (var analyticsPrefName in analyticsPrefList) 
    {
     queryString += "&" + analyticsPrefName + "=" + gAnalyticsPrefBranch.getIntPref(analyticsPrefName);
    }

    Debug("data=" + queryString);
    return queryString;
  },
  
  reset: function ()
  {
    Debug("reset");
    gAnalyticsPrefBranch.deleteBranch("");
  },
  
  observe: function (aSubject, aTopic, aData) 
  {
    switch (aTopic) 
    {
     case "pb:analytics":
        if (gPrefs.getBoolPref("pb.analytics_service.enabled") == false)
        {
          Debug("ignored, disabled: " + aTopic);
          return;
        }
  
        var oldValue = 0;
        var prefName = sdbm_hash(aData);
        try
        {
          oldValue = gAnalyticsPrefBranch.getIntPref(prefName);
        }
        catch (ex)
        {
          oldValue = 0;
        }
        gAnalyticsPrefBranch.setIntPref(prefName, oldValue + 1);
        Debug(aTopic + " data=" + aData + " hash=" + prefName + " count=" + (oldValue + 1));
        break;
     case "quit-application":
        this.observerService.removeObserver(this, "quit-application");
        this.observerService.removeObserver(this, "pb:analytics");
        break;
    }
  },
}
